@@ -66,7 +66,7 @@ do
 
     function ctxrunner.reflag(flags)
         local t = { }
-        for _, flag in pairs(flags) do
+        for _, flag in next, flags do
             local key, value = flag:match("^(.-)=(.+)$")
             if key and value then
                 t[key] = value
@@ -122,21 +122,24 @@ do
                 return
             end
         end
-        if table.is_empty(ctxdata.prepfiles) then
-            logs.simple("nothing prepared, no ctl file saved")
-            os.remove(ctlname)
-        else
+        local prepfiles = ctxdata.prepfiles
+        if prepfiles and next(prepfiles) then
             logs.simple("saving logdata in: %s",ctlname)
             f = io.open(ctlname,'w')
             if f then
                 f:write("<?xml version='1.0' standalone='yes'?>\n\n")
                 f:write(string.format("<ctx:preplist local='%s'>\n",yn(ctxdata.runlocal)))
-                for _, name in ipairs(table.sortedkeys(ctxdata.prepfiles)) do
-                    f:write(string.format("\t<ctx:prepfile done='%s'>%s</ctx:prepfile>\n",yn(ctxdata.prepfiles[name]),name))
+                local sorted = table.sortedkeys(prepfiles)
+                for i=1,#sorted do
+                    local name = sorted[i]
+                    f:write(string.format("\t<ctx:prepfile done='%s'>%s</ctx:prepfile>\n",yn(prepfiles[name]),name))
                 end
                 f:write("</ctx:preplist>\n")
                 f:close()
             end
+        else
+            logs.simple("nothing prepared, no ctl file saved")
+            os.remove(ctlname)
         end
     end
 
@@ -179,7 +182,7 @@ do
         local found    = lfs.isfile(usedname)
 
         if not found then
-            for _, path in pairs(ctxdata.locations) do
+            for _, path in next, ctxdata.locations do
                 local fullname = file.join(path,ctxdata.ctxname)
                 if lfs.isfile(fullname) then
                     usedname, found = fullname, true
@@ -220,8 +223,9 @@ do
 
         ctxdata.flags = ctxrunner.reflag(ctxdata.flags)
 
-        for _, message in ipairs(ctxdata.messages) do
-            logs.simple("ctx comment: %s", xml.tostring(message))
+        local messages = ctxdata.messages
+        for i=1,#messages do
+            logs.simple("ctx comment: %s", xml.tostring(messages[i]))
         end
 
         for r, d, k in xml.elements(ctxdata.xmldata,"ctx:value[@name='job']") do
@@ -233,8 +237,8 @@ do
             commands[e.at and e.at['name'] or "unknown"] = e
         end
 
-        local suffix   = xml.filter(ctxdata.xmldata,"/ctx:job/ctx:preprocess/attribute(suffix)") or ctxdata.suffix
-        local runlocal = xml.filter(ctxdata.xmldata,"/ctx:job/ctx:preprocess/ctx:processors/attribute(local)")
+        local suffix   = xml.filter(ctxdata.xmldata,"/ctx:job/ctx:preprocess/attribute('suffix')") or ctxdata.suffix
+        local runlocal = xml.filter(ctxdata.xmldata,"/ctx:job/ctx:preprocess/ctx:processors/attribute('local')")
 
         runlocal = toboolean(runlocal)
 
@@ -261,7 +265,9 @@ do
                     local pluspath = false
                     if #oldfiles == 0 then
                         -- message: no files match pattern
-                        for _, p in ipairs(ctxdata.paths) do
+                        local paths = ctxdata.paths
+                        for i=1,#paths do
+                            local p = paths[i]
                             local oldfiles = dir.glob(path.join(p,pattern))
                             if #oldfiles > 0 then
                                 pluspath = true
@@ -272,15 +278,18 @@ do
                     if #oldfiles == 0 then
                         -- message: no old files
                     else
-                        for _, oldfile in ipairs(oldfiles) do
-                            newfile = oldfile .. "." .. suffix -- addsuffix will add one only
+                        for i=1,#oldfiles do
+                            local oldfile = oldfiles[i]
+                            local newfile = oldfile .. "." .. suffix -- addsuffix will add one only
                             if ctxdata.runlocal then
                                 newfile = file.basename(newfile)
                             end
                             if oldfile ~= newfile and file.needsupdate(oldfile,newfile) then
                             --  message: oldfile needs preprocessing
                             --  os.remove(newfile)
-                                for _, pp in ipairs(preprocessor:split(',')) do
+                                local splitted = preprocessor:split(',')
+                                for i=1,#splitted do
+                                    local pp = splitted[i]
                                     local command = commands[pp]
                                     if command then
                                         command = xml.copy(command)
@@ -311,6 +320,7 @@ do
                                         command = ctxrunner.justtext(command)
                                         logs.simple("command: %s",command)
                                         local result = os.spawn(command) or 0
+                                        -- somehow we get the wrong return value
                                         if result > 0 then
                                             logs.simple("error, return code: %s",result)
                                         end
@@ -360,7 +370,9 @@ scripts.context.multipass = {
 
 function scripts.context.multipass.hashfiles(jobname)
     local hash = { }
-    for _, suffix in ipairs(scripts.context.multipass.suffixes) do
+    local suffixes = scripts.context.multipass.suffixes
+    for i=1,#suffixes do
+        local suffix = suffixes[i]
         local full = jobname .. suffix
         hash[full] = md5.hex(io.loaddata(full) or "unknown")
     end
@@ -368,7 +380,7 @@ function scripts.context.multipass.hashfiles(jobname)
 end
 
 function scripts.context.multipass.changed(oldhash, newhash)
-    for k,v in pairs(oldhash) do
+    for k,v in next, oldhash do
         if v ~= newhash[k] then
             return true
         end
@@ -406,7 +418,7 @@ function scripts.context.multipass.makeoptionfile(jobname,ctxdata,kindofrun,curr
         end
         local function setvalues(flag,format,plural)
             if type(flag) == "table"  then
-                for k, v in pairs(flag) do
+                for k, v in next, flag do
                     f:write(format:format(v),"\n")
                 end
             else
@@ -431,13 +443,6 @@ function scripts.context.multipass.makeoptionfile(jobname,ctxdata,kindofrun,curr
         --
         setalways("\\unprotect")
         --
-        setalways("%% special commands, mostly for the ctx development team")
-        --
-        if environment.argument("dumpdelta") then
-            setalways("\\tracersdumpdelta")
-        elseif environment.argument("dumphash") then
-            setalways("\\tracersdumphash")
-        end
         setalways("%% feedback and basic job control")
         if type(environment.argument("track")) == "string" then
             setvalue ("track"    , "\\enabletrackers[%s]")
@@ -450,7 +455,9 @@ function scripts.context.multipass.makeoptionfile(jobname,ctxdata,kindofrun,curr
         end
         setfixed ("timing"       , "\\usemodule[timing]")
         setfixed ("batchmode"    , "\\batchmode")
+        setfixed ("batch"        , "\\batchmode")
         setfixed ("nonstopmode"  , "\\nonstopmode")
+        setfixed ("nonstop"      , "\\nonstopmode")
         setfixed ("tracefiles"   , "\\tracefilestrue")
         setfixed ("nostats"      , "\\nomkivstatistics")
         setfixed ("paranoid"     , "\\def\\maxreadlevel{1}")
@@ -469,7 +476,7 @@ function scripts.context.multipass.makeoptionfile(jobname,ctxdata,kindofrun,curr
         setalways(                 "\\setupsystem[inputfile=%s]",environment.argument("input") or environment.files[1] or "\\jobname")
         setvalue ("result"       , "\\setupsystem[file=%s]")
         setalways(                 "\\setupsystem[\\c!n=%s,\\c!m=%s]", kindofrun or 0, currentrun or 0)
-    --  setalways(                 "\\setupsystem[\\c!type=%s]",os.platform)
+    --  setalways(                 "\\setupsystem[\\c!type=%s]",os.type) -- windows or unix
         setvalues("path"         , "\\usepath[%s]")
         setvalue ("setuppath"    , "\\setupsystem[\\c!directory={%s}]")
         setvalue ("randomseed"   , "\\setupsystem[\\c!random=%s]")
@@ -572,18 +579,15 @@ scripts.context.interfaces = {
     it = "cont-it",
     ro = "cont-ro",
     pe = "cont-pe",
-    -- for taco and me
- -- xp = "cont-xp",
 }
 
 scripts.context.defaultformats  = {
     "cont-en",
     "cont-nl",
---  "cont-xp",
-    "mptopdf",
---  "metatex",
-    "metafun",
-    "plain"
+--  "mptopdf", -- todo: mak emkiv variant
+--  "metatex", -- will show up soon
+--  "metafun", -- todo: mp formats
+--  "plain"
 }
 
 local function analyze(filename)
@@ -600,6 +604,9 @@ local function analyze(filename)
         elseif line:find("^<?xml ") then
             t.type = "xml"
         end
+        if t.nofruns then
+            scripts.context.multipass.nofruns = t.nofruns
+        end
         if not t.engine then
             t.engine = 'luatex'
         end
@@ -647,7 +654,7 @@ function scripts.context.run(ctxdata,filename)
     local files = (filename and { filename }) or environment.files
     if ctxdata then
         -- todo: interface
-        for k,v in pairs(ctxdata.flags) do
+        for k,v in next, ctxdata.flags do
             environment.setargument(k,v)
         end
     end
@@ -668,20 +675,28 @@ function scripts.context.run(ctxdata,filename)
         end
         --
         if formatfile and scriptfile then
-            for _, filename in ipairs(files) do
+            for i=1,#files do
+                local filename = files[i]
                 local basename, pathname = file.basename(filename), file.dirname(filename)
                 local jobname = file.removesuffix(basename)
-                if pathname == "" then
+                if pathname == "" and not environment.argument("global") then
                     filename = "./" .. filename
                 end
                 -- look at the first line
                 local a = analyze(filename)
                 if a and (a.engine == 'pdftex' or a.engine == 'xetex' or environment.argument("pdftex") or environment.argument("xetex")) then
-                    local texexec = resolvers.find_file("texexec.rb") or ""
-                    if texexec ~= "" then
-                        os.setenv("RUBYOPT","")
-                        local command = string.format("ruby %s %s",texexec,environment.reconstruct_commandline(environment.arguments_after))
-                        os.exec(command)
+                    if false then
+                        -- we need to write a top etc too and run mp etc so it's not worth the
+                        -- trouble, so it will take a while before the next is finished
+                        --
+                        -- require "mtx-texutil.lua"
+                    else
+                        local texexec = resolvers.find_file("texexec.rb") or ""
+                        if texexec ~= "" then
+                            os.setenv("RUBYOPT","")
+                            local command = string.format("ruby %s %s",texexec,environment.reconstruct_commandline(environment.arguments_after))
+                            os.exec(command)
+                        end
                     end
                 else
                     if a and a.interface and a.interface ~= interface then
@@ -725,14 +740,23 @@ function scripts.context.run(ctxdata,filename)
                             oldbase = file.removesuffix(jobname)
                             newbase = file.removesuffix(resultname)
                             if oldbase ~= newbase then
-                                for _, suffix in pairs(scripts.context.beforesuffixes) do
-                                    local oldname = file.addsuffix(oldbase,suffix)
-                                    local newname = file.addsuffix(newbase,suffix)
-                                    local tmpname = "keep-"..oldname
-                                    os.remove(tmpname)
-                                    os.rename(oldname,tmpname)
-                                    os.remove(oldname)
-                                    os.rename(newname,oldname)
+                                if environment.argument("purgeresult") then
+                                    for _, suffix in next, scripts.context.aftersuffixes do
+                                        local oldname = file.addsuffix(oldbase,suffix)
+                                        local newname = file.addsuffix(newbase,suffix)
+                                        os.remove(newname)
+                                        os.remove(oldname)
+                                    end
+                                else
+                                    for _, suffix in next, scripts.context.beforesuffixes do
+                                        local oldname = file.addsuffix(oldbase,suffix)
+                                        local newname = file.addsuffix(newbase,suffix)
+                                        local tmpname = "keep-"..oldname
+                                        os.remove(tmpname)
+                                        os.rename(oldname,tmpname)
+                                        os.remove(oldname)
+                                        os.rename(newname,oldname)
+                                    end
                                 end
                             else
                                 resultname = nil
@@ -741,7 +765,7 @@ function scripts.context.run(ctxdata,filename)
                             resultname = nil
                         end
                         --
-                        if environment.argument("autopdf") then
+                        if environment.argument("autopdf") or environment.argument("closepdf") then
                             scripts.context.closepdf(filename)
                             if resultname then
                                 scripts.context.closepdf(resultname)
@@ -756,7 +780,7 @@ function scripts.context.run(ctxdata,filename)
                         end
                         --
                         local flags = { }
-                        if environment.argument("batchmode") then
+                        if environment.argument("batchmode") or environment.argument("batch") then
                             flags[#flags+1] = "--interaction=batchmode"
                         end
                         if environment.argument("synctex") then
@@ -829,13 +853,24 @@ function scripts.context.run(ctxdata,filename)
                         os.remove(jobname..".top")
                         --
                         if resultname then
-                            for _, suffix in pairs(scripts.context.aftersuffixes) do
-                                local oldname = file.addsuffix(oldbase,suffix)
-                                local newname = file.addsuffix(newbase,suffix)
-                                local tmpname = "keep-"..oldname
-                                os.remove(newname)
-                                os.rename(oldname,newname)
-                                os.rename(tmpname,oldname)
+                            if environment.argument("purgeresult") then
+                                -- so, if there is no result then we don't get the old one, but
+                                -- related files (log etc) are still there for tracing purposes
+                                for _, suffix in next, scripts.context.aftersuffixes do
+                                    local oldname = file.addsuffix(oldbase,suffix)
+                                    local newname = file.addsuffix(newbase,suffix)
+                                    os.remove(newname) -- to be sure
+                                    os.rename(oldname,newname)
+                                end
+                            else
+                                for _, suffix in next, scripts.context.aftersuffixes do
+                                    local oldname = file.addsuffix(oldbase,suffix)
+                                    local newname = file.addsuffix(newbase,suffix)
+                                    local tmpname = "keep-"..oldname
+                                    os.remove(newname)
+                                    os.rename(oldname,newname)
+                                    os.rename(tmpname,oldname)
+                                end
                             end
                             logs.simple("result renamed to: %s",newbase)
                         end
@@ -931,30 +966,33 @@ function scripts.context.pipe()
     end
 end
 
+local make_mkiv_format = environment.make_format
+
+local function make_mkii_format(name,engine)
+    if environment.argument(engine) then
+        local command = string.format("mtxrun texexec.rb --make --%s %s",name,engine)
+        logs.simple("running command: %s",command)
+        os.spawn(command)
+    end
+end
+
 function scripts.context.make(name)
-    local runners = {
-        "luatools --make --compile ",
-        (environment.argument("pdftex") and "mtxrun texexec.rb --make --pdftex ") or false,
-        (environment.argument("xetex")  and "mtxrun texexec.rb --make --xetex " ) or false,
-    }
     local list = (name and { name }) or (environment.files[1] and environment.files) or scripts.context.defaultformats
-    for _, name in ipairs(list) do
-        name = scripts.context.interfaces[name] or name
-        for _, runner in ipairs(runners) do
-            if runner then
-                local command = runner .. name
-                logs.simple("running command: %s",command)
-                os.spawn(command)
-            end
+    for i=1,#list do
+        local name = list[i]
+        name = scripts.context.interfaces[name] or name or ""
+        if name ~= "" then
+            make_mkiv_format(name)
+            make_mkii_format(name,"pdftex")
+            make_mkii_format(name,"xetex")
         end
     end
 end
 
 function scripts.context.generate()
-    -- hack, should also be a shared function
-    local command = "luatools --generate "
-    logs.simple("running command: %s",command)
-    os.spawn(command)
+    resolvers.instance.renewcache = true
+    trackers.enable("resolvers.locating")
+    resolvers.load()
 end
 
 function scripts.context.ctx()
@@ -1086,15 +1124,15 @@ function scripts.context.purge_job(jobname,all)
         jobname = file.basename(jobname)
         local filebase = file.removesuffix(jobname)
         local deleted = { }
-        for _, suffix in ipairs(obsolete_results) do
-            deleted[#deleted+1] = purge_file(filebase.."."..suffix,filebase..".pdf")
+        for i=1,#obsolete_results do
+            deleted[#deleted+1] = purge_file(filebase.."."..obsolete_results[i],filebase..".pdf")
         end
-        for _, suffix in ipairs(temporary_runfiles) do
-            deleted[#deleted+1] = purge_file(filebase.."."..suffix)
+        for i=1,#temporary_runfiles do
+            deleted[#deleted+1] = purge_file(filebase.."."..temporary_runfiles[i])
         end
         if all then
-            for _, suffix in ipairs(persistent_runfiles) do
-                deleted[#deleted+1] = purge_file(filebase.."."..suffix)
+            for i=1,#persistent_runfiles do
+                deleted[#deleted+1] = purge_file(filebase.."."..persistent_runfiles[i])
             end
         end
         if #deleted > 0 then
@@ -1112,7 +1150,8 @@ function scripts.context.purge(all)
     local persistent = table.tohash(persistent_runfiles)
     local generic = table.tohash(generic_files)
     local deleted = { }
-    for _, name in ipairs(files) do
+    for i=1,#files do
+        local name = files[i]
         local suffix = file.extname(name)
         local basename = file.basename(name)
         if obsolete[suffix] or temporary[suffix] or persistent[suffix] or generic[basename] then
@@ -1181,7 +1220,8 @@ function scripts.context.extras(pattern)
         else
             logs.extendbanner(extra)
         end
-        for k,v in ipairs(list) do
+        for i=1,#list do
+            local v = list[i]
             local data = io.loaddata(v) or ""
             data = string.match(data,"begin help(.-)end help")
             if data then
@@ -1228,14 +1268,14 @@ end
 -- todo: we need to do a dummy run
 
 function scripts.context.trackers()
-    environment.files = { "m-trackers" }
+    environment.files = { resolvers.find_file("m-trackers.tex") }
     scripts.context.multipass.nofruns = 1
     scripts.context.run()
     -- maybe filter from log
 end
 
 function scripts.context.directives()
-    environment.files = { "m-directives" }
+    environment.files = { resolvers.find_file("m-directives.tex") }
     scripts.context.multipass.nofruns = 1
     scripts.context.run()
     -- maybe filter from log
@@ -1245,9 +1285,10 @@ function scripts.context.timed(action)
     statistics.timed(action)
 end
 
-local zipname    = "cont-tmf.zip"
-local mainzip    = "http://www.pragma-ade.com/context/latest/" .. zipname
-local validtrees = { "texmf-local", "texmf-context" }
+local zipname     = "cont-tmf.zip"
+local mainzip     = "http://www.pragma-ade.com/context/latest/" .. zipname
+local validtrees  = { "texmf-local", "texmf-context" }
+local selfscripts = { "mtxrun.lua" } -- was: { "luatools.lua", "mtxrun.lua" }
 
 function zip.loaddata(zipfile,filename) -- should be in zip lib
     local f = zipfile:open(filename)
@@ -1354,7 +1395,7 @@ function scripts.context.update()
                 end
             end
         end
-        for _, scriptname in next, { "luatools.lua", "mtxrun.lua" } do
+        for _, scriptname in next, selfscripts do
             local oldscript = resolvers.find_file(scriptname) or ""
             if oldscript ~= "" and is_okay(oldscript) then
                 local newscript = "./scripts/context/lua/" .. scriptname
@@ -1370,8 +1411,10 @@ function scripts.context.update()
             end
         end
         if force then
-            os.execute("context --generate")
-            os.execute("context --make")
+         -- os.execute("context --generate")
+         -- os.execute("context --make")
+            scripts.context.generate()
+            scripts.context.make()
         end
     end
     if force then
@@ -1381,32 +1424,42 @@ function scripts.context.update()
     end
 end
 
-logs.extendbanner("ConTeXt Tools 0.51",true)
+logs.extendbanner("ConTeXt Process Management 0.51")
 
 messages.help = [[
 --run                 process (one or more) files (default action)
 --make                create context formats
---generate            generate file database etc.
---ctx=name            use ctx file
---version             report installed context version
+
+--ctx=name            use ctx file (process management specification)
+--interface           use specified user interface (default: en)
+
+--autopdf             close pdf file in viewer and start pdf viewer afterwards
+--purge(all)          purge files either or not after a run (--pattern=...)
+
+--usemodule=list      load the given module or style, normally part o fthe distribution
+--environment=list    load the given environment file first (document styles)
+--mode=list           enable given the modes (conditional processing in styles)
+--path=list           also consult the given paths when files are looked for
+--arguments=list      set variables that can be consulted during a run (key/value pairs)
+--randomseed=number   set the randomseed
+--result=name         rename the resulting output to the given name
+--trackers=list       show/set tracker variables
+--directives=list     show/set directive variables
+--purgeresult         purge result file before run
+
 --forcexml            force xml stub (optional flag: --mkii)
 --forcecld            force cld (context lua document) stub
---autopdf             close pdf file in viewer and start pdf viewer afterwards
---once                only one run
---purge(all)          purge files (--pattern=...)
---result=name         rename result to given name
---arrange             run extra arrange pass
---noarrange           ignore arrange commands in the file
+
+--arrange             run extra imposition pass, given that the style sets up imposition
+--noarrange           ignore imposition specifications in the style
+
+--once                only run once (no multipass data file is produced)
 --batchmode           run without stopping and don't show messages on the console
 --nonstopmode         run without stopping
---usemodule=list      load the given module (or style)
---environment=list    load the given file first
---mode=list           enable given the mode(s)
---path=list           also consult the given paths when files are looked for
+
+--generate            generate file database etc. (as luatools does)
 --paranoid            don't descend to .. and ../..
---randomseed=number   set the randomseed
---arguments=list      set variables that can be consulted during a run (key/value pairs)
---interface           use specified user interface
+--version             report installed context version
 
 --expert              expert options
 ]]
@@ -1426,23 +1479,13 @@ expert options:
 --nostats             omit runtime statistics at the end of the run
 --update              update context from website (not to be confused with contextgarden)
 --profile             profile job (use: mtxrun --script profile --analyse)
---trackers            show/set tracker variables
---directives          show/set directive variables
 --timing              generate timing and statistics overview
 --tracefiles          show some extra info when locating files (at the tex end)
---randomseed
 
 --extra=name          process extra (mtx-context-<name> in distribution)
 --extras              show extras
 ]]
 
-messages.private = [[
-private options:
-
---dumphash            dump hash table afterwards
---dumpdelta           dump hash table afterwards (only new entries)
-]]
-
 messages.special = [[
 special options:
 
@@ -1454,6 +1497,8 @@ special options:
 
 if environment.argument("once") then
     scripts.context.multipass.nofruns = 1
+elseif environment.argument("runs") then
+    scripts.context.multipass.nofruns = tonumber(environment.argument("runs")) or nil
 end
 
 if environment.argument("profile") then
@@ -1483,7 +1528,7 @@ elseif environment.argument("touch") then
 elseif environment.argument("update") then
     scripts.context.update()
 elseif environment.argument("expert") then
-    logs.help(table.join({ messages.expert, messages.private, messages.special },"\n"))
+    logs.help(table.join({ messages.expert, messages.special },"\n"))
 elseif environment.argument("extras") then
     scripts.context.extras()
 elseif environment.argument("extra") then
