@@ -11,7 +11,7 @@ if not modules then modules = { } end modules ['mtx-convert'] = {
 graphics            = graphics            or { }
 graphics.converters = graphics.converters or { }
 
-local gsprogram = (os.platform == "windows" and "gswin32c") or "gs"
+local gsprogram = (os.type == "windows" and "gswin32c") or "gs"
 local gstemplate = "%s -q -sDEVICE=pdfwrite -dEPSCrop -dNOPAUSE -dNOCACHE -dBATCH -dAutoRotatePages=/None -dProcessColorModel=/DeviceCMYK -sOutputFile=%s %s -c quit"
 
 function graphics.converters.eps(oldname,newname)
@@ -112,13 +112,14 @@ function scripts.convert.convertall()
 end
 
 function scripts.convert.convertgiven()
-    for _, name in ipairs(environment.files) do
-        graphics.converters.convertfile(name)
+    local files = environment.files
+    for i=1,#files do
+        graphics.converters.convertfile(files[i])
     end
 end
 
 
-logs.extendbanner("Graphic Conversion Tools 0.10",true)
+logs.extendbanner("ConTeXT Graphic Conversion Helpers 0.10")
 
 messages.help = [[
 --convertall          convert all graphics on path
