@@ -114,31 +114,39 @@ scripts.update.goodies = {
 }
 
 scripts.update.platforms = {
-    ["mswin"]         = "mswin",
-    ["windows"]       = "mswin",
-    ["win32"]         = "mswin",
-    ["win"]           = "mswin",
-    ["linux"]         = "linux",
-    ["freebsd"]       = "freebsd",
-    ["freebsd-amd64"] = "freebsd-amd64",
-    ["linux-32"]      = "linux",
-    ["linux-64"]      = "linux-64",
-    ["linux32"]       = "linux",
-    ["linux64"]       = "linux-64",
-    ["linux-ppc"]     = "linux-ppc",
-    ["ppc"]           = "linux-ppc",
-    ["osx"]           = "osx-intel",
-    ["macosx"]        = "osx-intel",
-    ["osx-intel"]     = "osx-intel",
-    ["osx-ppc"]       = "osx-ppc",
-    ["osx-powerpc"]   = "osx-ppc",
-    ["osx-64"]        = "osx-64",
-    ["osxintel"]      = "osx-intel",
-    ["osxppc"]        = "osx-ppc",
-    ["osxpowerpc"]    = "osx-ppc",
-    ["solaris-intel"] = "solaris-intel",
-    ["solaris-sparc"] = "solaris-sparc",
-    ["solaris"]       = "solaris-sparc",
+    ["mswin"]          = "mswin",
+    ["windows"]        = "mswin",
+    ["win32"]          = "mswin",
+    ["win"]            = "mswin",
+    ["linux"]          = "linux",
+    ["freebsd"]        = "freebsd",
+    ["freebsd-amd64"]  = "freebsd-amd64",
+    ["kfreebsd"]       = "kfreebsd-i386",
+    ["kfreebsd-i386"]  = "kfreebsd-i386",
+    ["kfreebsd-amd64"] = "kfreebsd-amd64",
+    ["linux-32"]       = "linux",
+    ["linux-64"]       = "linux-64",
+    ["linux32"]        = "linux",
+    ["linux64"]        = "linux-64",
+    ["linux-ppc"]      = "linux-ppc",
+    ["ppc"]            = "linux-ppc",
+    ["osx"]            = "osx-intel",
+    ["macosx"]         = "osx-intel",
+    ["osx-intel"]      = "osx-intel",
+    ["osx-ppc"]        = "osx-ppc",
+    ["osx-powerpc"]    = "osx-ppc",
+    ["osx-64"]         = "osx-64",
+    ["osxintel"]       = "osx-intel",
+    ["osxppc"]         = "osx-ppc",
+    ["osxpowerpc"]     = "osx-ppc",
+    ["solaris-intel"]  = "solaris-intel",
+    ["solaris-sparc"]  = "solaris-sparc",
+    ["solaris"]        = "solaris-sparc",
+}
+
+scripts.update.selfscripts = {
+    "mtxrun",
+ -- "luatools",
 }
 
 -- the list is filled up later (when we know what modules to download)
@@ -146,13 +154,15 @@ scripts.update.platforms = {
 scripts.update.modules = {
 }
 
-function scripts.update.run(str,force)
-    logs.report("run", str)
-    if force or environment.argument("force") then
-        -- important, otherwise formats fly to a weird place
-        -- (texlua sets luatex as the engine, we need to reset that or to fix texexec :)
-        os.setenv("engine",nil)
+function scripts.update.run(str)
+    -- important, otherwise formats fly to a weird place
+    -- (texlua sets luatex as the engine, we need to reset that or to fix texexec :)
+    os.setenv("engine",nil)
+    if environment.argument("force") then
+        logs.report("run", str)
         os.execute(str)
+    else
+        logs.report("dry run", str)
     end
 end
 
@@ -168,16 +178,16 @@ function scripts.update.synchronize()
 
     logs.report("update","start")
 
-    local texroot = scripts.update.fullpath(states.get("paths.root"))
-    local engines = states.get('engines') or { }
-    local platforms = states.get('platforms') or { }
-    local repositories = states.get('repositories') -- minimals
-    local bin = states.get("rsync.program")         -- rsync
-    local url = states.get("rsync.server")          -- contextgarden.net
-    local version = states.get("context.version")   -- current (or beta)
-    local extras = states.get("extras")             -- extras (like modules)
-    local goodies = states.get("goodies")           -- goodies (like editors)
-    local force = environment.argument("force")
+    local texroot      = scripts.update.fullpath(states.get("paths.root"))
+    local engines      = states.get('engines') or { }
+    local platforms    = states.get('platforms') or { }
+    local repositories = states.get('repositories')      -- minimals
+    local bin          = states.get("rsync.program")     -- rsync
+    local url          = states.get("rsync.server")      -- contextgarden.net
+    local version      = states.get("context.version")   -- current (or beta)
+    local extras       = states.get("extras")            -- extras (like modules)
+    local goodies      = states.get("goodies")           -- goodies (like editors)
+    local force        = environment.argument("force")
 
     bin = string.gsub(bin,"\\","/")
 
@@ -191,18 +201,20 @@ function scripts.update.synchronize()
     if force then
         dir.mkdirs(format("%s/%s", texroot, "texmf-cache"))
         dir.mkdirs(format("%s/%s", texroot, "texmf-local"))
+        dir.mkdirs(format("%s/%s", texroot, "texmf-project"))
+        dir.mkdirs(format("%s/%s", texroot, "texmf-fonts"))
     end
 
     if ok or not force then
 
-        local fetched, individual, osplatform = { }, { }, os.currentplatform()
+        local fetched, individual, osplatform = { }, { }, os.platform
 
         -- takes a collection as argument and returns a list of folders
 
         local function collection_to_list_of_folders(collection, platform)
             local archives = {}
-            for _, c in ipairs(collection) do
-                local archive = c[1]
+            for i=1,#collection do
+                local archive = collection[i][1]
                 archive = archive:gsub("<platform>", platform)
                 archive = archive:gsub("<version>", version)
                 archives[#archives+1] = archive
@@ -223,7 +235,7 @@ function scripts.update.synchronize()
             return prefix .. concat(list_of_folders, format(" %s", prefix))
         end
 
-        -- example of usage: print(list_of_folders_to_rsync_string(collection_to_list_of_folders(scripts.update.base, os.currentplatform)))
+        -- example of usage: print(list_of_folders_to_rsync_string(collection_to_list_of_folders(scripts.update.base, os.platform)))
 
         -- rename function and add some more functionality:
         --   * recursive/non-recursive (default: non-recursive)
@@ -258,7 +270,8 @@ function scripts.update.synchronize()
             local available_modules = get_list_of_files_from_rsync({"modules/"})
             -- hash of requested modules
             -- local h = table.tohash(extras:split(","))
-            for _, s in ipairs(available_modules) do
+            for i=1,#available_modules do
+                local s = available_modules[i]
             --  if extras == "all" or h[s] then
                 if extras.all or extras[s] then
                     scripts.update.modules[#scripts.update.modules+1] = { format("modules/%s/",s), "texmf-context" }
@@ -272,7 +285,8 @@ function scripts.update.synchronize()
             if collection and platform then
                 platform = scripts.update.platforms[platform]
                 if platform then
-                    for _, c in ipairs(collection) do
+                    for i=1,#collection do
+                        local c = collection[i]
                         local archive = c[1]:gsub("<platform>", platform)
                         local destination = format("%s/%s", texroot, c[2]:gsub("<platform>", platform))
                         destination = destination:gsub("\\","/")
@@ -286,30 +300,32 @@ function scripts.update.synchronize()
             end
         end
 
-        for platform, _ in pairs(platforms) do
+        for platform, _ in next, platforms do
             add_collection(scripts.update.base,platform)
         end
-        for platform, _ in pairs(platforms) do
+        for platform, _ in next, platforms do
             add_collection(scripts.update.modules,platform)
         end
-        for engine, _ in pairs(engines) do
-            for platform, _ in pairs(platforms) do
+        for engine, _ in next, engines do
+            for platform, _ in next, platforms do
                 add_collection(scripts.update.engines[engine],platform)
             end
         end
 
         if goodies and type(goodies) == "table" then
-            for goodie, _ in pairs(goodies) do
-                for platform, _ in pairs(platforms) do
+            for goodie, _ in next, goodies do
+                for platform, _ in next, platforms do
                     add_collection(scripts.update.goodies[goodie],platform)
                 end
             end
         end
 
         local combined = { }
-        for _, repository in ipairs(scripts.update.repositories) do
+        local update_repositories = scripts.update.repositories
+        for i=1,#update_repositories do
+            local repository = update_repositories[i]
             if repositories[repository] then
-                for _, v in pairs(individual) do
+                for _, v in next, individual do
                     local archive, destination = v[1], v[2]
                     local cd = combined[destination]
                     if not cd then
@@ -320,15 +336,13 @@ function scripts.update.synchronize()
                 end
             end
         end
-        if logs.verbose then
-            for k, v in pairs(combined) do
-                logs.report("update", k)
-                for k,v in ipairs(v) do
-                    logs.report("update", "  <= " .. v)
-                end
-            end
-        end
-        for destination, archive in pairs(combined) do
+        --~ for k, v in next, combined do
+        --~     logs.report("update", k)
+        --~     for i=1,#v do
+        --~         logs.report("update", "  <= " .. v[i])
+        --~     end
+        --~ end
+        for destination, archive in next, combined do
             local archives, command = concat(archive," "), ""
         --  local normalflags, deleteflags = states.get("rsync.flags.normal"), states.get("rsync.flags.delete")
         --    if environment.argument("keep") or destination:find("%.$") then
@@ -337,6 +351,9 @@ function scripts.update.synchronize()
         --        command = format("%s %s %s %s'%s' '%s'", bin, normalflags, deleteflags, url, archives, destination)
         --    end
             local normalflags, deleteflags = states.get("rsync.flags.normal"), ""
+if os.name == "windows" then
+    normalflags = normalflags .. " -L" -- no symlinks
+end
             local dryrunflags = ""
             if not environment.argument("force") then
                 dryrunflags = "--dry-run"
@@ -370,9 +387,10 @@ function scripts.update.synchronize()
             end
         end
 
-        for platform, _ in pairs(platforms) do
-            update_script('luatools',platform)
-            update_script('mtxrun',platform)
+        for platform, _ in next, platforms do
+            for i=1, #scripts.update.selfscripts do
+                update_script(scripts.update.selfscripts[i],platform)
+            end
         end
 
     else
@@ -381,12 +399,20 @@ function scripts.update.synchronize()
     if not force then
         logs.report("update", "use --force to really update files")
     end
+
+    resolvers.load_tree(texroot) -- else we operate in the wrong tree
+
+    -- update filename database for pdftex/xetex
+    scripts.update.run("mktexlsr")
+    -- update filename database for luatex
+    scripts.update.run(format('mtxrun --tree="%s" --generate',texroot))
+
     logs.report("update","done")
 end
 
 function table.fromhash(t)
     local h = { }
-    for k, v in pairs(t) do -- no ipairs here
+    for k, v in next, t do -- not indexed
         if v then h[#h+1] = k end
     end
     return h
@@ -397,36 +423,36 @@ function scripts.update.make()
 
     logs.report("make","start")
 
-    local force = environment.argument("force")
-    local texroot = scripts.update.fullpath(states.get("paths.root"))
-    local engines = states.get('engines')
-    local goodies = states.get('goodies')
+    local force     = environment.argument("force")
+    local texroot   = scripts.update.fullpath(states.get("paths.root"))
+    local engines   = states.get('engines')
+    local goodies   = states.get('goodies')
     local platforms = states.get('platforms')
-    local formats = states.get('formats')
+    local formats   = states.get('formats')
 
     resolvers.load_tree(texroot)
-    -- update filename database for pdftex/xetex
+
     scripts.update.run("mktexlsr")
-    -- update filename database for luatex
-    scripts.update.run("luatools --generate")
+    scripts.update.run(format('mtxrun --tree="%s" --generate',texroot))
+
     local askedformats = formats
     local texformats = table.tohash(scripts.update.texformats)
     local mpformats = table.tohash(scripts.update.mpformats)
-    for k,v in pairs(texformats) do
+    for k,v in next, texformats do
         if not askedformats[k] then
             texformats[k] = nil
         end
     end
-    for k,v in pairs(mpformats) do
+    for k,v in next, mpformats do
         if not askedformats[k] then
             mpformats[k] = nil
         end
     end
     local formatlist = concat(table.fromhash(texformats), " ")
     if formatlist ~= "" then
-        for engine in pairs(engines) do
+        for engine in next, engines do
             if engine == "luatex" then
-                scripts.update.run(format("context --make")) -- maybe also formatlist
+                scripts.update.run(format('mtxrun --tree="%s" --script context --autogenerate --make',texroot))
             else
                 -- todo: just handle make here or in mtxrun --script context --make
                 scripts.update.run(format("texexec --make --all --fast --%s %s",engine,formatlist))
@@ -441,18 +467,18 @@ function scripts.update.make()
         logs.report("make", "use --force to really make formats")
     end
     scripts.update.run("mktexlsr")
-    scripts.update.run("luatools --generate")
+    scripts.update.run(format('mtxrun --tree="%s" --generate',texroot))
     logs.report("make","done")
 end
 
-logs.extendbanner("Download Tools 0.20",true)
+logs.extendbanner("ConTeXt Minimals Updater 0.21")
 
 messages.help = [[
 --platform=string     platform (windows, linux, linux-64, osx-intel, osx-ppc, linux-ppc)
 --server=string       repository url (rsync://contextgarden.net)
 --module=string       repository url (minimals)
 --repository=string   specify version (current, experimental)
---context=string      specify version (current, latest, yyyy.mm.dd)
+--context=string      specify version (current, latest, beta, yyyy.mm.dd)
 --rsync=string        rsync binary (rsync)
 --texroot=string      installation directory (not guessed for the moment)
 --engine=string       tex engine (luatex, pdftex, xetex)
@@ -497,7 +523,7 @@ if scripts.savestate then
     local valid = scripts.update.engines
     for r in gmatch(environment.argument("engine") or "all","([^, ]+)") do
         if r == "all" then
-            for k, v in pairs(valid) do
+            for k, v in next, valid do
                 if k ~= "all" then
                     states.set("engines." .. k, true)
                 end
@@ -507,7 +533,7 @@ if scripts.savestate then
         end
     end
     local valid = scripts.update.platforms
-    for r in gmatch(environment.argument("platform") or os.currentplatform(),"([^, ]+)") do
+    for r in gmatch(environment.argument("platform") or os.platform,"([^, ]+)") do
         if valid[r] then states.set("platforms." .. r, true) end
     end
 
