@@ -198,7 +198,7 @@ class TEX
         'filters', 'usemodules', 'environments', 'separation', 'setuppath',
         'arguments', 'input', 'output', 'randomseed', 'modes', 'mode', 'filename',
         'ctxfile', 'printformat', 'paperformat', 'paperoffset',
-        'timeout', 'passon'
+        'timeout', 'passon', 'pdftitle'
     ]
     @@mainstandardvars = [
         'mainlanguage', 'bodyfont', 'language'
@@ -761,7 +761,8 @@ class TEX
         # utilities
         report('start of analysis')
         results = Array.new
-        ['texexec','texutil','ctxtools'].each do |program|
+        # ['texexec','texutil','ctxtools'].each do |program|
+        ['texexec'].each do |program|
             result = `texmfstart #{program} --help`
             result.sub!(/.*?(#{program}[^\n]+)\n.*/mi) do $1 end
             results.push("#{result}")
@@ -1723,10 +1724,10 @@ end
 
     def fixbackendvars(backend)
         if backend then
-            report("fixing backend map path for #{backend}") if getvariable('verbose')
             ENV['backend']     = backend ;
             ENV['progname']    = backend unless validtexengine(backend)
-            ENV['TEXFONTMAPS'] = ['.',"\$TEXMF/fonts/map/{#{backend},pdftex,dvips,}//",'./fonts//'].join_path
+            ENV['TEXFONTMAPS'] = ['.',"\$TEXMF/fonts/{data,map}/{#{backend},pdftex,dvips,}//",'./fonts//'].join_path
+            report("fixing backend map path for #{backend}: #{ENV['TEXFONTMAPS']}") if getvariable('verbose')
         else
             report("unable to fix backend map path") if getvariable('verbose')
         end
@@ -2052,7 +2053,10 @@ end
 
                     Kpse.runscript('ctxtools',rawbase,'--purge')       if getvariable('purge')
                     Kpse.runscript('ctxtools',rawbase,'--purge --all') if getvariable('purgeall')
-# till here
+
+                    # runcommand('mtxrun','--script','ctxtools',rawbase,'--purge')       if getvariable('purge')
+                    # runcommand('mtxrun','--script','ctxtools',rawbase,'--purge --all') if getvariable('purgeall')
+
                 when 'latex' then
 
                     ok = runtex(rawname)
