@@ -13,10 +13,13 @@ if not modules then modules = { } end modules ['chem-str'] = {
 local trace_structure = false  trackers.register("chemistry.structure",  function(v) trace_structure = v end)
 local trace_textstack = false  trackers.register("chemistry.textstack",  function(v) trace_textstack = v end)
 
+local report_chemistry = logs.new("chemistry")
+
 local format, gmatch, match, lower, gsub = string.format, string.gmatch, string.match, string.lower, string.gsub
 local concat, insert, remove = table.concat, table.insert, table.remove
 local apply = structure.processors.apply
 local texsprint, ctxcatcodes = tex.sprint, tex.ctxcatcodes
+local lpegmatch = lpeg.match
 
 local variables = interfaces.variables
 
@@ -59,7 +62,7 @@ local one_keys = {
     sb = "line", db = "line", tb = "line",
     ep = "line", es = "line", ed = "line", et = "line",
     sd = "line", ldd = "line", rdd = "line",
-    hb = "line", bb = "line", oe = "line",
+    hb = "line", bb = "line", oe = "line", bd = "line", bw = "line",
     z = "text", cz = "text", zt = "text", zn = "number",
     zbt = "text", zbn = "number", ztt = "text", ztn = "number",
     mov = "transform", sub = "transform", dir = "transform", off = "transform",
@@ -141,10 +144,11 @@ local syntax = {
 local definitions = { }
 
 function chemicals.undefine(name)
-    definitions[name] = nil
+    definitions[lower(name)] = nil
 end
 
 function chemicals.define(name,spec,text)
+    name = lower(name)
     local dn = definitions[name]
     if not dn then dn = { } definitions[name] = dn end
     dn[#dn+1] = {
@@ -154,7 +158,7 @@ function chemicals.define(name,spec,text)
 end
 
 local metacode, kind, keys, bonds, max, txt, textsize, rot, pstack
-local molecule = chemicals.molecule -- or use chemicals.moleculeparser:match(...)
+local molecule = chemicals.molecule -- or use lpegmatch(chemicals.moleculeparser,...)
 
 local function fetch(txt)
     local st = stack[txt]
@@ -168,7 +172,7 @@ local function fetch(txt)
     end
     if t then
         if trace_textstack then
-            logs.report("chemical", "fetching from stack %s slot %s: %s",txt,st.n,t)
+            report_chemistry("fetching from stack %s slot %s: %s",txt,st.n,t)
         end
 st.n = st.n + 1
     end
@@ -197,12 +201,12 @@ local pattern   =
         lpeg.Cc(false) * lpeg.Cc(false) * lpeg.Cc(false) * text
     )
 
---~ local n, operation, index, upto, set, text = pattern:match("RZ1357")
+--~ local n, operation, index, upto, set, text = lpegmatch(pattern,"RZ1357")
 
---~ print(pattern:match("RZ=x"))        1 RZ false false false  x
---~ print(pattern:match("RZ1=x"))       1 RZ 1     false false	x
---~ print(pattern:match("RZ1..3=x"))    1 RZ 1     3     false	x
---~ print(pattern:match("RZ13=x"))      1 RZ false false table	x
+--~ print(lpegmatch(pattern,"RZ=x"))        1 RZ false false false  x
+--~ print(lpegmatch(pattern,"RZ1=x"))       1 RZ 1     false false	x
+--~ print(lpegmatch(pattern,"RZ1..3=x"))    1 RZ 1     3     false	x
+--~ print(lpegmatch(pattern,"RZ13=x"))      1 RZ false false table	x
 
 local function process(spec,text,n,rulethickness,rulecolor,offset)
     insert(stack,{ spec=spec, text=text, n=n })
@@ -216,7 +220,7 @@ local function process(spec,text,n,rulethickness,rulecolor,offset)
                 process(di.spec,di.text,1,rulethickness,rulecolor)
             end
         else
-            local rep, operation, special, index, upto, set, text = pattern:match(s)
+            local rep, operation, special, index, upto, set, text = lpegmatch(pattern,s)
             if operation == "pb" then
                 insert(pstack,kind)
                 metacode[#metacode+1] = syntax.pb.direct
@@ -439,9 +443,9 @@ function chemicals.stop()
     metacode[#metacode+1] = "chem_stop_structure ;"
     local mpcode = concat(metacode,"\n")
     if trace_structure then
-        logs.report("chemical", "metapost code:\n%s", mpcode)
+        report_chemistry("metapost code:\n%s", mpcode)
     end
-    metapost.graphic(chemicals.instance,chemicals.format,mpcode,"")
+    metapost.graphic(chemicals.instance,chemicals.format,mpcode)
     metacode = nil
 end
 
