@@ -21,7 +21,11 @@ local trace_features = false  trackers.register("afm.features", function(v) trac
 local trace_indexing = false  trackers.register("afm.indexing", function(v) trace_indexing = v end)
 local trace_loading  = false  trackers.register("afm.loading",  function(v) trace_loading  = v end)
 
-local format, match, gmatch, lower = string.format, string.match, string.gmatch, string.lower
+local report_afm = logs.new("load afm")
+
+local format, match, gmatch, lower, gsub = string.format, string.match, string.gmatch, string.lower, string.gsub
+local lpegmatch = lpeg.match
+local abs = math.abs
 
 fonts      = fonts     or { }
 fonts.afm  = fonts.afm or { }
@@ -29,7 +33,7 @@ fonts.afm  = fonts.afm or { }
 local afm = fonts.afm
 local tfm = fonts.tfm
 
-afm.version          = 1.400 -- incrementing this number one up will force a re-cache
+afm.version          = 1.402 -- incrementing this number one up will force a re-cache
 afm.syncspace        = true  -- when true, nicer stretch values
 afm.enhance_data     = true  -- best leave this set to true
 afm.features         = { }
@@ -88,7 +92,7 @@ local pattern = ( c * s^1 * (
 
 local function scan_comment(str)
     fd = { }
-    pattern:match(str)
+    lpegmatch(pattern,str)
     return fd
 end
 
@@ -161,50 +165,41 @@ local function get_variables(data,fontmetrics)
     end
 end
 
-local function get_indexes(data,filename)
-    local pfbfile = file.replacesuffix(filename,"pfb")
-    local pfbname = resolvers.find_file(pfbfile,"pfb") or ""
-    if pfbname == "" then
-        pfbname = resolvers.find_file(file.basename(pfbfile),"pfb") or ""
-    end
-    if pfbname ~= "" then
-        data.luatex.filename = pfbname
-        local pfbblob = fontloader.open(pfbname)
-        if pfbblob then
-            local characters = data.characters
-            local pfbdata = fontloader.to_table(pfbblob)
-        --~ print(table.serialize(pfbdata))
-            if pfbdata then
-                local glyphs = pfbdata.glyphs
-                if glyphs then
-                    if trace_loading then
-                        logs.report("load afm","getting index data from %s",pfbname)
-                    end
-                    -- local offset = (glyphs[0] and glyphs[0] != .notdef) or 0
-                    for index, glyph in next, glyphs do
-                        local name = glyph.name
-                        if name then
-                            local char = characters[name]
-                            if char then
-                                if trace_indexing then
-                                    logs.report("load afm","glyph %s has index %s",name,index)
-                                end
-                                char.index = index
+local function get_indexes(data,pfbname)
+    data.luatex.filename = pfbname
+    local pfbblob = fontloader.open(pfbname)
+    if pfbblob then
+        local characters = data.characters
+        local pfbdata = fontloader.to_table(pfbblob)
+    --~ print(table.serialize(pfbdata))
+        if pfbdata then
+            local glyphs = pfbdata.glyphs
+            if glyphs then
+                if trace_loading then
+                    report_afm("getting index data from %s",pfbname)
+                end
+                -- local offset = (glyphs[0] and glyphs[0] != .notdef) or 0
+                for index, glyph in next, glyphs do
+                    local name = glyph.name
+                    if name then
+                        local char = characters[name]
+                        if char then
+                            if trace_indexing then
+                                report_afm("glyph %s has index %s",name,index)
                             end
+                            char.index = index
                         end
                     end
-                elseif trace_loading then
-                    logs.report("load afm","no glyph data in pfb file %s",pfbname)
                 end
             elseif trace_loading then
-                logs.report("load afm","no data in pfb file %s",pfbname)
+                report_afm("no glyph data in pfb file %s",pfbname)
             end
-            fontloader.close(pfbblob)
         elseif trace_loading then
-            logs.report("load afm","invalid pfb file %s",pfbname)
+            report_afm("no data in pfb file %s",pfbname)
         end
+        fontloader.close(pfbblob)
     elseif trace_loading then
-        logs.report("load afm","no pfb file for %s",filename)
+        report_afm("invalid pfb file %s",pfbname)
     end
 end
 
@@ -219,23 +214,23 @@ function afm.read_afm(filename)
                 filename = file.removesuffix(file.basename(filename))
             }
         }
-        afmblob = afmblob:gsub("StartCharMetrics(.-)EndCharMetrics", function(charmetrics)
+        afmblob = gsub(afmblob,"StartCharMetrics(.-)EndCharMetrics", function(charmetrics)
             if trace_loading then
-                logs.report("load afm","loading char metrics")
+                report_afm("loading char metrics")
             end
             get_charmetrics(data,charmetrics,vector)
             return ""
         end)
-        afmblob = afmblob:gsub("StartKernPairs(.-)EndKernPairs", function(kernpairs)
+        afmblob = gsub(afmblob,"StartKernPairs(.-)EndKernPairs", function(kernpairs)
             if trace_loading then
-                logs.report("load afm","loading kern pairs")
+                report_afm("loading kern pairs")
             end
             get_kernpairs(data,kernpairs)
             return ""
         end)
-        afmblob = afmblob:gsub("StartFontMetrics%s+([%d%.]+)(.-)EndFontMetrics", function(version,fontmetrics)
+        afmblob = gsub(afmblob,"StartFontMetrics%s+([%d%.]+)(.-)EndFontMetrics", function(version,fontmetrics)
             if trace_loading then
-                logs.report("load afm","loading variables")
+                report_afm("loading variables")
             end
             data.afmversion = version
             get_variables(data,fontmetrics)
@@ -243,11 +238,10 @@ function afm.read_afm(filename)
             return ""
         end)
         data.luatex = { }
-        get_indexes(data,filename)
         return data
     else
         if trace_loading then
-            logs.report("load afm","no valid afm file %s",filename)
+            report_afm("no valid afm file %s",filename)
         end
         return nil
     end
@@ -264,28 +258,51 @@ function afm.load(filename)
     filename = resolvers.find_file(filename,'afm') or ""
     if filename ~= "" then
         local name = file.removesuffix(file.basename(filename))
-        local data = containers.read(afm.cache(),name)
-        local size = lfs.attributes(filename,"size") or 0
-        if not data or data.verbose ~= fonts.verbose or data.size ~= size then
-            logs.report("load afm", "reading %s",filename)
+        local data = containers.read(afm.cache,name)
+        local attr = lfs.attributes(filename)
+        local size, time = attr.size or 0, attr.modification or 0
+        --
+        local pfbfile = file.replacesuffix(name,"pfb")
+        local pfbname = resolvers.find_file(pfbfile,"pfb") or ""
+        if pfbname == "" then
+            pfbname = resolvers.find_file(file.basename(pfbfile),"pfb") or ""
+        end
+        local pfbsize, pfbtime = 0, 0
+        if pfbname ~= "" then
+            local attr = lfs.attributes(pfbname)
+            pfbsize, pfbtime = attr.size or 0, attr.modification or 0
+        end
+        if not data or data.verbose ~= fonts.verbose
+                or data.size ~= size or data.time ~= time or data.pfbsize ~= pfbsize or data.pfbtime ~= pfbtime then
+            report_afm( "reading %s",filename)
             data = afm.read_afm(filename)
             if data then
             --  data.luatex = data.luatex or { }
-                logs.report("load afm", "unifying %s",filename)
+                if pfbname ~= "" then
+                    get_indexes(data,pfbname)
+                elseif trace_loading then
+                    report_afm("no pfb file for %s",filename)
+                end
+                report_afm( "unifying %s",filename)
                 afm.unify(data,filename)
                 if afm.enhance_data then
-                    logs.report("load afm", "add ligatures")
+                    report_afm( "add ligatures")
                     afm.add_ligatures(data,'ligatures') -- easier this way
-                    logs.report("load afm", "add tex-ligatures")
+                    report_afm( "add tex-ligatures")
                     afm.add_ligatures(data,'texligatures') -- easier this way
-                    logs.report("load afm", "add extra kerns")
+                    report_afm( "add extra kerns")
                     afm.add_kerns(data) -- faster this way
                 end
+                report_afm( "add tounicode data")
+                fonts.map.add_to_unicode(data,filename)
                 data.size = size
+                data.time = time
+                data.pfbsize = pfbsize
+                data.pfbtime = pfbtime
                 data.verbose = fonts.verbose
-                logs.report("load afm","saving: %s in cache",name)
-                data = containers.write(afm.cache(), name, data)
-                data = containers.read(afm.cache(),name)
+                report_afm("saving: %s in cache",name)
+                data = containers.write(afm.cache, name, data)
+                data = containers.read(afm.cache,name)
             end
         end
         return data
@@ -306,7 +323,7 @@ function afm.unify(data, filename)
             if not code then
                 code = private
                 private = private + 1
-                logs.report("afm glyph", "assigning private slot U+%04X for unknown glyph name %s", code, name)
+                report_afm("assigning private slot U+%04X for unknown glyph name %s", code, name)
             end
         end
         local index = blob.index
@@ -453,6 +470,9 @@ function afm.add_dimensions(data) -- we need to normalize afm to otf i.e. indexe
     end
 end
 
+fonts.formats.afm = "type1"
+fonts.formats.pfb = "type1"
+
 function afm.copy_to_tfm(data)
     if data then
         local glyphs = data.glyphs
@@ -460,74 +480,48 @@ function afm.copy_to_tfm(data)
             local metadata, luatex = data.metadata, data.luatex
             local unicodes, indices = luatex.unicodes, luatex.indices
             local characters, parameters, descriptions = { }, { }, { }
-            local tfm = {
-                characters = characters,
-                parameters = parameters,
-                descriptions = descriptions,
-                indices = indices,
-                unicodes = unicodes,
-                luatex = luatex,
-            }
+            -- todo : merge into tfm
             for u, i in next, indices do
                 local d = glyphs[i]
                 characters[u] = { }
                 descriptions[u] = d
             end
-            tfm.encodingbytes      = metadata.encodingbytes or 2
-            tfm.fullname           = metadata.fullname
-            tfm.filename           = metadata.filename
-            tfm.name               = tfm.fullname
-            tfm.psname             = tfm.fullname
-            tfm.type               = "real"
-            tfm.units              = 1000
-            tfm.stretch            = stretch -- nil
-            tfm.slant              = slant -- nil
-            tfm.direction          = 0
-            tfm.boundarychar_label = 0
-            tfm.boundarychar       = 65536
-        --~ tfm.false_boundarychar = 65536 -- produces invalid tfm in luatex
-            tfm.designsize         = (metadata.designsize or 10)*65536
-            local spaceunits = 500
-            tfm.spacer = "500 units"
+            local filename = fonts.tfm.checked_filename(luatex) -- was metadata.filename
+            local fontname = metadata.fontname or metadata.fullname
+            local fullname = metadata.fullname or metadata.fontname
+            local endash, emdash, spacer, spaceunits = unicodes['space'], unicodes['emdash'], "space", 500
             -- same as otf
-            local endash, emdash = unicodes['space'], unicodes['emdash']
             if metadata.isfixedpitch then
                 if descriptions[endash] then
-                    spaceunits, tfm.spacer = descriptions[endash].width, "space"
+                    spaceunits, spacer = descriptions[endash].width, "space"
                 end
                 if not spaceunits and descriptions[emdash] then
-                    spaceunits, tfm.spacer = descriptions[emdash].width, "emdash"
+                    spaceunits, spacer = descriptions[emdash].width, "emdash"
                 end
                 if not spaceunits and metadata.charwidth then
-                    spaceunits, tfm.spacer = metadata.charwidth, "charwidth"
+                    spaceunits, spacer = metadata.charwidth, "charwidth"
                 end
             else
                 if descriptions[endash] then
-                    spaceunits, tfm.spacer = descriptions[endash].width, "space"
+                    spaceunits, spacer = descriptions[endash].width, "space"
                 end
-            --    if not spaceunits and descriptions[emdash] then
-            --        spaceunits, tfm.spacer = descriptions[emdash].width/2, "emdash/2"
-            --    end
                 if not spaceunits and metadata.charwidth then
-                    spaceunits, tfm.spacer = metadata.charwidth, "charwidth"
+                    spaceunits, spacer = metadata.charwidth, "charwidth"
                 end
             end
-            --
             spaceunits = tonumber(spaceunits)
+            if spaceunits < 200 then
+                -- todo: warning
+            end
+            --
             parameters.slant         = 0
             parameters.space         = spaceunits
             parameters.space_stretch = 500
             parameters.space_shrink  = 333
             parameters.x_height      = 400
             parameters.quad          = 1000
-            if spaceunits < 200 then
-                -- todo: warning
-            end
-            tfm.ascender    = math.abs(metadata.ascender  or 0)
-            tfm.descender   = math.abs(metadata.descender or 0)
             local italicangle = data.metadata.italicangle
             if italicangle then
-                tfm.italicangle = italicangle
                 parameters.slant = parameters.slant - math.round(math.tan(italicangle*math.pi/180))
             end
             if metadata.isfixedpitch then
@@ -557,8 +551,34 @@ function afm.copy_to_tfm(data)
                     parameters[k] = v
                 end
             end
+            --
             if next(characters) then
-                return tfm
+                return {
+                    characters         = characters,
+                    parameters         = parameters,
+                    descriptions       = descriptions,
+                    indices            = indices,
+                    unicodes           = unicodes,
+                    luatex             = luatex,
+                    encodingbytes      = 2,
+                    filename           = filename,
+                    fontname           = fontname,
+                    fullname           = fullname,
+                    psname             = fullname, -- in otf: tfm.fontname or tfm.fullname
+                    name               = filename or fullname or fontname,
+                    format             = fonts.fontformat(filename,"type1"),
+                    type               = 'real',
+                    units              = 1000,
+                    direction          = 0,
+                    boundarychar_label = 0,
+                    boundarychar       = 65536,
+                --~ false_boundarychar = 65536, -- produces invalid tfm in luatex
+                    designsize         = (metadata.designsize or 10)*65536,
+                    spacer             = spacer,
+                    ascender           = abs(metadata.ascender  or 0),
+                    descender          = abs(metadata.descender or 0),
+                    italicangle        = italicangle,
+                }
             end
         end
     end
@@ -581,7 +601,7 @@ function afm.set_features(tfmdata)
     local shared = tfmdata.shared
     local afmdata = shared.afmdata
     local features = shared.features
-    if not table.is_empty(features) then
+    if features and next(features) then
         local mode = tfmdata.mode or fonts.mode
         local initializers = fonts.initializers
         local fi = initializers[mode]
@@ -600,7 +620,7 @@ function afm.set_features(tfmdata)
                         local value = features[f]
                         if value and fiafm[f] then -- brr
                             if trace_features then
-                                logs.report("define afm","initializing feature %s to %s for mode %s for font %s",f,tostring(value),mode or 'unknown',tfmdata.name or 'unknown')
+                                report_afm("initializing feature %s to %s for mode %s for font %s",f,tostring(value),mode or 'unknown',tfmdata.name or 'unknown')
                             end
                             fiafm[f](tfmdata,value)
                             mode = tfmdata.mode or fonts.mode
@@ -612,7 +632,7 @@ function afm.set_features(tfmdata)
         end
         local fm = fonts.methods[mode]
         local fmafm = fm and fm.afm
-        if fmfm then
+        if fmafm then
             local lists = {
                 afm.features.list,
             }
@@ -649,13 +669,13 @@ function afm.afm_to_tfm(specification)
     local afmname = specification.filename or specification.name
     if specification.forced == "afm" or specification.format == "afm" then -- move this one up
         if trace_loading then
-            logs.report("load afm","forcing afm format for %s",afmname)
+            report_afm("forcing afm format for %s",afmname)
         end
     else
         local tfmname = resolvers.findbinfile(afmname,"ofm") or ""
         if tfmname ~= "" then
             if trace_loading then
-                logs.report("load afm","fallback from afm to tfm for %s",afmname)
+                report_afm("fallback from afm to tfm for %s",afmname)
             end
             afmname = ""
         end
@@ -667,13 +687,13 @@ function afm.afm_to_tfm(specification)
         specification = fonts.define.resolve(specification) -- new, was forgotten
         local features = specification.features.normal
         local cache_id = specification.hash
-        local tfmdata  = containers.read(tfm.cache(), cache_id) -- cache with features applied
+        local tfmdata  = containers.read(tfm.cache, cache_id) -- cache with features applied
         if not tfmdata then
             local afmdata = afm.load(afmname)
-            if not table.is_empty(afmdata) then
+            if afmdata and next(afmdata) then
                 afm.add_dimensions(afmdata)
                 tfmdata = afm.copy_to_tfm(afmdata)
-                if not table.is_empty(tfmdata) then
+                if tfmdata and next(tfmdata) then
                     tfmdata.shared = tfmdata.shared or { }
                     tfmdata.unique = tfmdata.unique or { }
                     tfmdata.shared.afmdata  = afmdata
@@ -681,9 +701,9 @@ function afm.afm_to_tfm(specification)
                     afm.set_features(tfmdata)
                 end
             elseif trace_loading then
-                logs.report("load afm","no (valid) afm file found with name %s",afmname)
+                report_afm("no (valid) afm file found with name %s",afmname)
             end
-            tfmdata = containers.write(tfm.cache(),cache_id,tfmdata)
+            tfmdata = containers.write(tfm.cache,cache_id,tfmdata)
         end
         return tfmdata
     end
@@ -711,23 +731,8 @@ function tfm.read_from_afm(specification)
     local tfmtable = afm.afm_to_tfm(specification)
     if tfmtable then
         tfmtable.name = specification.name
-        tfmtable = tfm.scale(tfmtable, specification.size)
+        tfmtable = tfm.scale(tfmtable, specification.size, specification.relativeid)
         local afmdata = tfmtable.shared.afmdata
-        local filename = afmdata and afmdata.luatex and afmdata.luatex.filename
-        if not filename then
-            -- try to locate anyway and set afmdata.luatex.filename
-        end
-        if filename then
-            tfmtable.encodingbytes = 2
-            tfmtable.filename = resolvers.findbinfile(filename,"") or filename
-            tfmtable.fontname = afmdata.metadata.fontname or afmdata.metadata.fullname
-            tfmtable.fullname = afmdata.metadata.fullname or afmdata.metadata.fontname
-            tfmtable.format   = 'type1'
-            tfmtable.name     = afmdata.luatex.filename or tfmtable.fullname
-        end
-        if fonts.dontembed[filename] then
-            tfmtable.file = nil
-        end
         fonts.logger.save(tfmtable,'afm',specification)
     end
     return tfmtable
