@@ -8,17 +8,22 @@ if not modules then modules = { } end modules ['font-ctx'] = {
 
 -- needs a cleanup: merge of replace, lang/script etc
 
-local texsprint, count = tex.sprint, tex.count
-local format, concat, gmatch, match, find, lower, gsub = string.format, table.concat, string.gmatch, string.match, string.find, string.lower, string.gsub
+local texsprint, count, texsetcount = tex.sprint, tex.count, tex.setcount
+local format, concat, gmatch, match, find, lower, gsub, byte = string.format, table.concat, string.gmatch, string.match, string.find, string.lower, string.gsub, string.byte
+
 local tostring, next, type = tostring, next, type
+local lpegmatch = lpeg.match
+local round = math.round
 
 local ctxcatcodes = tex.ctxcatcodes
 
 local trace_defining = false  trackers.register("fonts.defining", function(v) trace_defining = v end)
 
+local report_define = logs.new("define fonts")
+
 local tfm      = fonts.tfm
 local define   = fonts.define
-local fontdata = fonts.ids
+local fontdata = fonts.identifiers
 local specify  = define.specify
 
 specify.context_setups  = specify.context_setups  or { }
@@ -48,7 +53,7 @@ name*context specification
 function specify.predefined(specification)
     local detail = specification.detail
     if detail ~= "" then
-    --  detail = detail:gsub("["..define.splitsymbols.."].*$","") -- get rid of *whatever specs and such
+    --  detail = gsub(detail,"["..define.splitsymbols.."].*$","") -- get rid of *whatever specs and such
         if define.methods[detail] then                            -- since these may be appended at the
             specification.features.vtf = { preset = detail }      -- tex end by default
         end
@@ -276,12 +281,13 @@ local spaces       = space^0
 local leftparent   = (P"(")
 local rightparent  = (P")")
 local value        = C((leftparent * (1-rightparent)^0 * rightparent + (1-space))^1)
+local dimension    = C((space/"" + P(1))^1)
 local rest         = C(P(1)^0)
 local scale_none   =               Cc(0)
-local scale_at     = P("at")     * Cc(1) * spaces * value
-local scale_sa     = P("sa")     * Cc(2) * spaces * value
-local scale_mo     = P("mo")     * Cc(3) * spaces * value
-local scale_scaled = P("scaled") * Cc(4) * spaces * value
+local scale_at     = P("at")     * Cc(1) * spaces * dimension -- value
+local scale_sa     = P("sa")     * Cc(2) * spaces * dimension -- value
+local scale_mo     = P("mo")     * Cc(3) * spaces * dimension -- value
+local scale_scaled = P("scaled") * Cc(4) * spaces * dimension -- value
 
 local sizepattern  = spaces * (scale_at + scale_sa + scale_mo + scale_scaled + scale_none)
 local splitpattern = spaces * value * spaces * rest
@@ -295,10 +301,10 @@ local get_specification = define.get_specification
 
 function define.command_1(str)
     statistics.starttiming(fonts)
-    local fullname, size = splitpattern:match(str)
+    local fullname, size = lpegmatch(splitpattern,str)
     local lookup, name, sub, method, detail = get_specification(fullname)
     if not name then
-        logs.report("define font","strange definition '%s'",str)
+        report_define("strange definition '%s'",str)
         texsprint(ctxcatcodes,"\\fcglet\\somefontname\\defaultfontfile")
     elseif name == "unknown" then
         texsprint(ctxcatcodes,"\\fcglet\\somefontname\\defaultfontfile")
@@ -307,7 +313,7 @@ function define.command_1(str)
     end
     -- we can also use a count for the size
     if size and size ~= "" then
-        local mode, size = sizepattern:match(size)
+        local mode, size = lpegmatch(sizepattern,size)
         if size and mode then
             count.scaledfontmode = mode
             texsprint(ctxcatcodes,"\\def\\somefontsize{",size,"}")
@@ -328,17 +334,29 @@ end
 
 local n = 0
 
-function define.command_2(global,cs,str,size,classfeatures,fontfeatures,classfallbacks,fontfallbacks,mathsize,textsize)
+-- we can also move rscale to here (more consistent)
+
+function define.command_2(global,cs,str,size,classfeatures,fontfeatures,classfallbacks,fontfallbacks,
+        mathsize,textsize,relativeid,classgoodies,goodies)
+    if trace_defining then
+        report_define("memory usage before: %s",statistics.memused())
+    end
     -- name is now resolved and size is scaled cf sa/mo
     local lookup, name, sub, method, detail = get_specification(str or "")
     -- asome settings can be overloaded
-    if lookup and lookup ~= "" then specification.lookup = lookup end
+    if lookup and lookup ~= "" then
+        specification.lookup = lookup
+    end
+    if relativeid and relativeid ~= "" then -- experimental hook
+        local id = tonumber(relativeid) or 0
+        specification.relativeid = id > 0 and id
+    end
     specification.name = name
     specification.size = size
---~     specification.sub = sub
     specification.sub = (sub and sub ~= "" and sub) or specification.sub
     specification.mathsize = mathsize
     specification.textsize = textsize
+    specification.goodies = goodies
     if detail and detail ~= "" then
         specification.method, specification.detail = method or "*", detail
     elseif specification.detail and specification.detail ~= "" then
@@ -348,9 +366,6 @@ function define.command_2(global,cs,str,size,classfeatures,fontfeatures,classfal
     elseif classfeatures and classfeatures ~= "" then
         specification.method, specification.detail = "*", classfeatures
     end
-    if trace_defining then
-        logs.report("define font","memory usage before: %s",statistics.memused())
-    end
     if fontfallbacks and fontfallbacks ~= "" then
         specification.fallbacks = fontfallbacks
     elseif classfallbacks and classfallbacks ~= "" then
@@ -358,14 +373,17 @@ function define.command_2(global,cs,str,size,classfeatures,fontfeatures,classfal
     end
     local tfmdata = define.read(specification,size) -- id not yet known
     if not tfmdata then
-        logs.report("define font","unable to define %s as \\%s",name,cs)
+        report_define("unable to define %s as \\%s",name,cs)
+        texsetcount("global","lastfontid",-1)
     elseif type(tfmdata) == "number" then
         if trace_defining then
-            logs.report("define font","reusing %s with id %s as \\%s (features: %s/%s, fallbacks: %s/%s)",name,tfmdata,cs,classfeatures,fontfeatures,classfallbacks,fontfallbacks)
+            report_define("reusing %s with id %s as \\%s (features: %s/%s, fallbacks: %s/%s, goodies: %s/%s)",
+                name,tfmdata,cs,classfeatures,fontfeatures,classfallbacks,fontfallbacks,classgoodies,goodies)
         end
         tex.definefont(global,cs,tfmdata)
         -- resolved (when designsize is used):
         texsprint(ctxcatcodes,format("\\def\\somefontsize{%isp}",fontdata[tfmdata].size))
+        texsetcount("global","lastfontid",tfmdata)
     else
     --  local t = os.clock(t)
         local id = font.define(tfmdata)
@@ -375,20 +393,49 @@ function define.command_2(global,cs,str,size,classfeatures,fontfeatures,classfal
         tex.definefont(global,cs,id)
         tfm.cleanup_table(tfmdata)
         if trace_defining then
-            logs.report("define font","defining %s with id %s as \\%s (features: %s/%s, fallbacks: %s/%s)",name,id,cs,classfeatures,fontfeatures,classfallbacks,fontfallbacks)
+            report_define("defining %s with id %s as \\%s (features: %s/%s, fallbacks: %s/%s)",name,id,cs,classfeatures,fontfeatures,classfallbacks,fontfallbacks)
         end
         -- resolved (when designsize is used):
         texsprint(ctxcatcodes,format("\\def\\somefontsize{%isp}",tfmdata.size))
     --~ if specification.fallbacks then
     --~     fonts.collections.prepare(specification.fallbacks)
     --~ end
+        texsetcount("global","lastfontid",id)
     end
     if trace_defining then
-        logs.report("define font","memory usage after: %s",statistics.memused())
+        report_define("memory usage after: %s",statistics.memused())
     end
     statistics.stoptiming(fonts)
 end
 
+local enable_auto_r_scale = false
+
+experiments.register("fonts.autorscale", function(v)
+    enable_auto_r_scale = v
+end)
+
+local calculate_scale = fonts.tfm.calculate_scale
+
+-- Not ok, we can best use a database for this. The problem is that we
+-- have delayed definitions and so we never know what style is taken
+-- as start.
+
+function fonts.tfm.calculate_scale(tfmtable, scaledpoints, relativeid)
+    local scaledpoints, delta, units = calculate_scale(tfmtable,scaledpoints)
+--~     if enable_auto_r_scale and relativeid then -- for the moment this is rather context specific
+--~         local relativedata = fontdata[relativeid]
+--~         local rfmtable = relativedata and relativedata.unscaled and relativedata.unscaled
+--~         local id_x_height = rfmtable and rfmtable.parameters and rfmtable.parameters.x_height
+--~         local tf_x_height = tfmtable and tfmtable.parameters and tfmtable.parameters.x_height
+--~         if id_x_height and tf_x_height then
+--~             local rscale = id_x_height/tf_x_height
+--~             delta = rscale * delta
+--~             scaledpoints = rscale * scaledpoints
+--~         end
+--~     end
+    return scaledpoints, delta, units
+end
+
 --~ table.insert(readers.sequence,1,'vtf')
 
 --~ function readers.vtf(specification)
@@ -413,6 +460,154 @@ function fonts.dimenfactor(unit,tfmdata)
     end
 end
 
-function fonts.cleanname(name) -- mapped onto macro, see fonts.names.cleanname
-    texsprint(ctxcatcodes,(gsub(lower(name),"[^%a%d]","")))
+function fonts.cleanname(name)
+    texsprint(ctxcatcodes,fonts.names.cleanname(name))
 end
+
+local p, f = 1, "%0.1fpt" -- normally this value is changed only once
+
+local stripper = lpeg.patterns.strip_zeros
+
+function fonts.nbfs(amount,precision)
+    if precision ~= p then
+        p = precision
+        f = "%0." .. p .. "fpt"
+    end
+    texsprint(ctxcatcodes,lpegmatch(stripper,format(f,amount/65536)))
+end
+
+-- for the moment here, this will become a chain of extras that is
+-- hooked into the ctx registration (or scaler or ...)
+
+function fonts.set_digit_width(font) -- max(quad/2,wd(0..9))
+    local tfmtable = fontdata[font]
+    local parameters = tfmtable.parameters
+    local width = parameters.digitwidth
+    if not width then
+        width = round(parameters.quad/2) -- maybe tex.scale
+        local characters = tfmtable.characters
+        for i=48,57 do
+            local wd = round(characters[i].width)
+            if wd > width then
+                width = wd
+            end
+        end
+        parameters.digitwidth = width
+    end
+    return width
+end
+
+fonts.get_digit_width = fonts.set_digit_width
+
+-- soon to be obsolete:
+
+local loaded = { -- prevent loading (happens in cont-sys files)
+    ["original-base.map"     ] = true,
+    ["original-ams-base.map" ] = true,
+    ["original-ams-euler.map"] = true,
+    ["original-public-lm.map"] = true,
+}
+
+function fonts.map.loadfile(name)
+    name = file.addsuffix(name,"map")
+    if not loaded[name] then
+        pdf.mapfile(name)
+        loaded[name] = true
+    end
+end
+
+local loaded = { -- prevent double loading
+}
+
+function fonts.map.loadline(how,line)
+    if line then
+        how = how .. " " .. line
+    elseif how == "" then
+        how = "= " .. line
+    end
+    if not loaded[how] then
+        pdf.mapline(how)
+        loaded[how] = true
+    end
+end
+
+function fonts.map.reset()
+    pdf.mapfile("")
+end
+
+fonts.map.reset() -- resets the default file
+
+-- we need an 'do after the banner hook'
+
+-- pdf.mapfile("mkiv-base.map") -- loads the default file
+
+local nounicode = byte("?")
+
+local function name_to_slot(name) -- maybe some day rawdata
+    local tfmdata = fonts.ids[font.current()]
+    local shared = tfmdata and tfmdata.shared
+    local fntdata = shared and shared.otfdata or shared.afmdata
+    if fntdata then
+        local unicode = fntdata.luatex.unicodes[name]
+        if not unicode then
+            return nounicode
+        elseif type(unicode) == "number" then
+            return unicode
+        else -- multiple unicodes
+            return unicode[1]
+        end
+    end
+    return nounicode
+end
+
+fonts.name_to_slot = name_to_slot
+
+function fonts.char(n) -- todo: afm en tfm
+    if type(n) == "string" then
+        n = name_to_slot(n)
+    end
+    if type(n) == "number" then
+        texsprint(ctxcatcodes,format("\\char%s ",n))
+    end
+end
+
+-- this will become obsolete:
+
+fonts.otf.name_to_slot = name_to_slot
+fonts.afm.name_to_slot = name_to_slot
+
+fonts.otf.char = fonts.char
+fonts.afm.char = fonts.char
+
+-- this will change ...
+
+function fonts.show_char_data(n)
+    local tfmdata = fonts.ids[font.current()]
+    if tfmdata then
+        if type(n) == "string" then
+            n = utf.byte(n)
+        end
+        local chr = tfmdata.characters[n]
+        if chr then
+            write_nl(format("%s @ %s => U%04X => %s => ",tfmdata.fullname,tfmdata.size,n,utf.char(n)) .. serialize(chr,false))
+        end
+    end
+end
+
+function fonts.show_font_parameters()
+    local tfmdata = fonts.ids[font.current()]
+    if tfmdata then
+        local parameters, mathconstants = tfmdata.parameters, tfmdata.MathConstants
+        local hasparameters, hasmathconstants = parameters and next(parameters), mathconstants and next(mathconstants)
+        if hasparameters then
+            write_nl(format("%s @ %s => parameters => ",tfmdata.fullname,tfmdata.size) .. serialize(parameters,false))
+        end
+        if hasmathconstants then
+            write_nl(format("%s @ %s => math constants => ",tfmdata.fullname,tfmdata.size) .. serialize(mathconstants,false))
+        end
+        if not hasparameters and not hasmathconstants then
+            write_nl(format("%s @ %s => no parameters and/or mathconstants",tfmdata.fullname,tfmdata.size))
+        end
+    end
+end
+
