@@ -13,6 +13,9 @@ if not modules then modules = { } end modules ['font-ini'] = {
 local utf = unicode.utf8
 local format, serialize = string.format, table.serialize
 local write_nl = texio.write_nl
+local lower = string.lower
+
+local report_define = logs.new("define fonts")
 
 if not fontloader then fontloader = fontforge end
 
@@ -22,7 +25,13 @@ fontloader.totable = fontloader.to_table
 -- fix comes last
 
 fonts     = fonts     or { }
-fonts.ids = fonts.ids or { } -- aka fontdata
+
+-- we will also have des and fam hashes
+
+fonts.ids = fonts.ids or { } fonts.identifiers = fonts.ids -- aka fontdata
+fonts.chr = fonts.chr or { } fonts.characters  = fonts.chr -- aka chardata
+fonts.qua = fonts.qua or { } fonts.quads       = fonts.qua -- aka quaddata
+
 fonts.tfm = fonts.tfm or { }
 
 fonts.mode    = 'base'
@@ -35,6 +44,8 @@ fonts.ids[0] = { -- nullfont
     name         = "nullfont",
 }
 
+fonts.chr[0] = { }
+
 fonts.methods = fonts.methods or {
     base = { tfm = { }, afm = { }, otf = { }, vtf = { }, fix = { } },
     node = { tfm = { }, afm = { }, otf = { }, vtf = { }, fix = { }  },
@@ -64,54 +75,26 @@ fonts.define.specify.synonyms = fonts.define.specify.synonyms or { }
 
 -- tracing
 
-fonts.color = fonts.color or { }
+if not fonts.color then
 
-local attribute = attributes.private('color')
-local mapping   = (attributes and attributes.list[attribute])  or { }
+    fonts.color = {
+        set   = function() end,
+        reset = function() end,
+    }
 
-local set_attribute   = node.set_attribute
-local unset_attribute = node.unset_attribute
+end
 
-function fonts.color.set(n,c)
-    local mc = mapping[c]
-    if not mc then
-        unset_attribute(n,attribute)
+-- format identification
+
+fonts.formats = { }
+
+function fonts.fontformat(filename,default)
+    local extname = lower(file.extname(filename))
+    local format = fonts.formats[extname]
+    if format then
+        return format
     else
-        set_attribute(n,attribute,mc)
-    end
-end
-function fonts.color.reset(n)
-    unset_attribute(n,attribute)
-end
-
--- this will change ...
-
-function fonts.show_char_data(n)
-    local tfmdata = fonts.ids[font.current()]
-    if tfmdata then
-        if type(n) == "string" then
-            n = utf.byte(n)
-        end
-        local chr = tfmdata.characters[n]
-        if chr then
-            write_nl(format("%s @ %s => U%04X => %s => ",tfmdata.fullname,tfmdata.size,n,utf.char(n)) .. serialize(chr,false))
-        end
-    end
-end
-
-function fonts.show_font_parameters()
-    local tfmdata = fonts.ids[font.current()]
-    if tfmdata then
-        local parameters, mathconstants = tfmdata.parameters, tfmdata.MathConstants
-        local hasparameters, hasmathconstants = parameters and next(parameters), mathconstants and next(mathconstants)
-        if hasparameters then
-            write_nl(format("%s @ %s => parameters => ",tfmdata.fullname,tfmdata.size) .. serialize(parameters,false))
-        end
-        if hasmathconstants then
-            write_nl(format("%s @ %s => math constants => ",tfmdata.fullname,tfmdata.size) .. serialize(mathconstants,false))
-        end
-        if not hasparameters and not hasmathconstants then
-            write_nl(format("%s @ %s => no parameters and/or mathconstants",tfmdata.fullname,tfmdata.size))
-        end
+        report_define("unable to determine font format for '%s'",filename)
+        return default
     end
 end
