@@ -12,6 +12,8 @@ if not modules then modules = { } end modules ['grph-inc'] = {
 -- dimensions
 -- consult rlx
 
+-- figures.boxnumber can go as we now can use names
+
 --[[
 The ConTeXt figure inclusion mechanisms are among the oldest code
 in ConTeXt and evolve dinto a complex whole. One reason is that we
@@ -33,13 +35,24 @@ The TeX-Lua mix is suboptimal. This has to do with the fact that we cannot
 run TeX code from within Lua. Some more functionality will move to Lua.
 ]]--
 
-local texsprint, format, lower, find, match = tex.sprint, string.format, string.lower, string.find, string.match
+-- commands.writestatus -> report
+
+local format, lower, find, match, gsub, gmatch = string.format, string.lower, string.find, string.match, string.gsub, string.gmatch
+local texsprint, texbox = tex.sprint, tex.box
+local contains = table.contains
+local concat = table.concat
+local todimen = string.todimen
 
 local ctxcatcodes = tex.ctxcatcodes
 local variables = interfaces.variables
 
-local trace_figures = false  trackers.register("figures.locating",function(v) trace_figures = v end)
-local trace_bases   = false  trackers.register("figures.bases",   function(v) trace_bases   = v end)
+local trace_figures    = false  trackers.register("figures.locating",   function(v) trace_figures    = v end)
+local trace_bases      = false  trackers.register("figures.bases",      function(v) trace_bases      = v end)
+local trace_programs   = false  trackers.register("figures.programs",   function(v) trace_programs   = v end)
+local trace_conversion = false  trackers.register("figures.conversion", function(v) trace_conversion = v end)
+local trace_inclusion  = false  trackers.register("figures.inclusion",  function(v) trace_inclusion  = v end)
+
+local report_graphics = logs.new("graphics")
 
 --- some extra img functions ---
 
@@ -70,7 +83,7 @@ local validtypes = table.tohash(img.types())
 
 function img.check_size(size)
     if size then
-        size = size:gsub("box","")
+        size = gsub(size,"box","")
         return (validsizes[size] and size) or "crop"
     else
         return "crop"
@@ -79,13 +92,13 @@ end
 
 ---
 
-figures                = figures          or { }
-figures.loaded         = figures.loaded   or { }
-figures.used           = figures.used     or { }
-figures.found          = figures.found    or { }
-figures.suffixes       = figures.suffixes or { }
-figures.patterns       = figures.patterns or { }
-figures.boxnumber      = figures.boxid    or 0
+figures                = figures           or { }
+figures.loaded         = figures.loaded    or { }
+figures.used           = figures.used      or { }
+figures.found          = figures.found     or { }
+figures.suffixes       = figures.suffixes  or { }
+figures.patterns       = figures.patterns  or { }
+figures.boxnumber      = figures.boxnumber or 0
 figures.defaultsearch  = true
 figures.defaultwidth   = 0
 figures.defaultheight  = 0
@@ -124,19 +137,21 @@ figures.formats = {
 
 function figures.setlookups()
     figures.suffixes, figures.patterns = { }, { }
-    for _, format in pairs(figures.order) do
+    for _, format in next, figures.order do
         local data = figures.formats[format]
         local fs, fp = figures.suffixes, figures.patterns
-        if data.list then
-            for _, s in ipairs(data.list) do
-                fs[s] = format -- hash
+        local list = data.list
+        if list then
+            for i=1,#list do
+                fs[list[i]] = format -- hash
             end
         else
             fs[format] = format
         end
-        if data.patterns then
-            for _, s in ipairs(data.patterns) do
-                fp[#fp+1] = { s, format } -- array
+        local patterns = data.patterns
+        if patterns then
+            for i=1,#patterns do
+                fp[#fp+1] = { patterns[i], format } -- array
             end
         end
     end
@@ -151,12 +166,12 @@ local function register(tag,target,what)
         figures.formats[target] = data
     end
     local d = data[tag] -- list or pattern
-    if d and not table.contains(d,what) then
+    if d and not contains(d,what) then
         d[#d+1] = what -- suffix or patternspec
     else
         data[tag] = { what }
     end
-    if not table.contains(figures.order,target) then
+    if not contains(figures.order,target) then
         figures.order[#figures.order+1] = target
     end
     figures.setlookups()
@@ -184,8 +199,11 @@ function figures.setpaths(locationset,pathlist)
         last_locationset = locationset
     end
     if h[iv["global"]] then
-        for s in pathlist:gmatch("([^, ]+)") do
-            if not table.contains(t,s) then
+     -- for s in gmatch(pathlist,",* *([^,]+)") do
+        local list = aux.settings_to_array(pathlist)
+        for i=1,#list do
+            local s = list[i]
+            if not contains(t,s) then
                 t[#t+1] = s
             end
         end
@@ -193,26 +211,12 @@ function figures.setpaths(locationset,pathlist)
     figures.paths, last_pathlist = t, pathlist
     if trace_figures then
         commands.writestatus("figures","locations: %s",last_locationset)
-        commands.writestatus("figures","path list: %s",table.concat(figures.paths, " "))
+        commands.writestatus("figures","path list: %s",concat(figures.paths, " "))
     end
 end
 
 -- check conversions and handle it here
 
---~ local keys = img.keys()
-
---~ function figures.hash(data)
---~     local i = data.status.private
---~     local t = { }
---~     for _, v in ipairs(keys) do
---~         local iv = i[v]
---~         if iv then
---~             t[#t+1] = v .. '=' .. iv
---~         end
---~     end
---~     return table.concat(t,"+")
---~ end
-
 function figures.hash(data)
     return data.status.hash or tostring(data.status.private) -- the <img object>
 --  return data.status.fullname .. "+".. (data.status.page or data.request.page or 1) -- img is still not perfect
@@ -225,46 +229,54 @@ do
     local figuredata = { }
     local callstack  = { }
 
-    function figures.new()
+    function figures.new() -- we could use metatables status -> used -> request but it needs testing
+        local request = {
+            name       = false,
+            label      = false,
+            format     = false,
+            page       = false,
+            width      = false,
+            height     = false,
+            preview    = false,
+            ["repeat"] = false,
+            controls   = false,
+            display    = false,
+            conversion = false,
+            cache      = false,
+            prefix     = false,
+            size       = false,
+        }
+        local used = {
+            fullname   = false,
+            format     = false,
+            name       = false,
+            path       = false,
+            suffix     = false,
+            width      = false,
+            height     = false,
+        }
+        local status = {
+            status     = 0,
+            converted  = false,
+            cached     = false,
+            fullname   = false,
+            format     = false,
+        }
+     -- setmetatable(status, { __index = used })
+     -- setmetatable(used,   { __index = request })
         figuredata = {
-            request = {
-                name = false,
-                label = false,
-                format = false,
-                page = false,
-                width = false,
-                height = false,
-                preview = false,
-                ["repeat"] = false,
-                controls = false,
-                display = false,
-                conversion = false,
-                cache = false,
-                prefix = false,
-                size = false,
-            },
-            used = {
-                fullname = false,
-                format = false,
-                name = false,
-                path = false,
-                suffix = false,
-                width = false,
-                height = false,
-            },
-            status = {
-                status = 0,
-                converted = false,
-                cached = false,
-                fullname = false,
-                format = false,
-            },
+            request = request,
+            used    = used,
+            status  = status,
         }
         return figuredata
     end
 
     function figures.push(request)
-        statistics.starttiming(figures)
+        local ncs = #callstack + 1
+        if ncs == 1 then
+            statistics.starttiming(figures)
+        end
         local figuredata = figures.new()
         if request then
         local iv = interfaces.variables
@@ -283,13 +295,16 @@ do
         --  request.height    = (h and h > 0) or false
             table.merge(figuredata.request,request)
         end
-        callstack[#callstack+1] = figuredata
+        callstack[ncs] = figuredata
         return figuredata
     end
     function figures.pop()
-        figuredata = callstack[#callstack]
-        callstack[#callstack] = nil
-        statistics.stoptiming(figures)
+        local ncs = #callstack
+        figuredata = callstack[ncs]
+        callstack[ncs] = nil
+        if ncs == 1 then
+            statistics.stoptiming(figures)
+        end
     end
     -- maybe move texsprint to tex
     function figures.get(category,tag,default)
@@ -310,19 +325,48 @@ do
 
 end
 
+local defaultformat = "pdf"
+local defaultprefix = "m_k_v_i_"
+
 local function register(askedname,specification)
     if specification then
         local format = specification.format
         if format then
-            local converter = figures.converters[format]
+            local conversion = specification.conversion
+            if conversion == "" then
+                conversion = nil
+            end
+            local newformat = conversion
+            if not newformat or newformat == "" then
+                newformat = defaultformat
+            end
+            local converter = (newformat ~= format) and figures.converters[format]
+            if trace_conversion then
+                report_graphics("checking conversion of '%s': old format '%s', new format '%s', conversion '%s'",
+                    askedname,format,newformat,conversion or "default")
+            end
+            if converter then
+                if converter[newformat] then
+                    converter = converter[newformat]
+                else
+                    newformat = defaultformat
+                    if converter[newformat] then
+                        converter = converter[newformat]
+                    else
+                        newformat = defaultformat
+                    end
+                end
+            end
             if converter then
                 local oldname = specification.fullname
-                local newformat = "pdf" -- todo, other target than pdf
                 local newpath = file.dirname(oldname)
-                local newbase = file.replacesuffix(file.basename(oldname),newformat)
+                local oldbase = file.basename(oldname)
+                local newbase = file.replacesuffix(oldbase,newformat)
                 local fc = specification.cache or figures.cachepaths.path
                 if fc and fc ~= "" and fc ~= "." then
                     newpath = fc
+                else
+                    newbase = defaultprefix .. newbase
                 end
                 local subpath = specification.subpath or figures.cachepaths.subpath
                 if subpath and subpath ~= "" and subpath ~= "."  then
@@ -334,10 +378,19 @@ local function register(askedname,specification)
                 end
                 local newname = file.join(newpath,newbase)
                 dir.makedirs(newpath)
+                oldname = file.collapse_path(oldname)
+                newname = file.collapse_path(newname)
                 local oldtime = lfs.attributes(oldname,'modification') or 0
                 local newtime = lfs.attributes(newname,'modification') or 0
                 if oldtime > newtime then
+                    if trace_conversion then
+                        report_graphics("converting '%s' from '%s' to '%s'",askedname,format,newformat)
+                    end
                     converter(oldname,newname)
+                else
+                    if trace_conversion then
+                        report_graphics("no need to convert '%s' from '%s' to '%s'",askedname,format,newformat)
+                    end
                 end
                 if io.exists(newname) then
                     specification.foundname = oldname
@@ -372,19 +425,39 @@ local function register(askedname,specification)
         specification = { }
     end
     specification.foundname = specification.foundname or specification.fullname
-    figures.found[askedname] = specification
+    figures.found[askedname .. "->" .. (specification.conversion or "default")] = specification
     return specification
 end
 
+local resolve_too = true -- urls
+
 local function locate(request) -- name, format, cache
     local askedname = resolvers.clean_path(request.name)
-    if figures.found[askedname] then
-        return figures.found[askedname]
+    local foundname = figures.found[askedname .. "->" .. (request.conversion or "default")]
+    if foundname then
+        return foundname
     end
+    -- protocol check
+    local hashed = url.hashed(askedname)
+    if hashed then
+        if hashed.scheme == "file" then
+            local path = hashed.path
+            if path and path ~= "" then
+                askedname = path
+            end
+        else
+            local foundname = resolvers.findbinfile(askedname)
+            if foundname then
+                askedname = foundname
+            end
+        end
+    end
+    -- we could use the hashed data instead
     local askedpath= file.is_rootbased_path(askedname)
     local askedbase = file.basename(askedname)
     local askedformat = (request.format ~= "" and request.format ~= "unknown" and request.format) or file.extname(askedname) or ""
     local askedcache = request.cache
+    local askedconversion = request.conversion
     if askedformat ~= "" then
         if trace_figures then
             commands.writestatus("figures","strategy: forced format")
@@ -392,7 +465,9 @@ local function locate(request) -- name, format, cache
         askedformat = lower(askedformat)
         local format = figures.suffixes[askedformat]
         if not format then
-            for _, pattern in ipairs(figures.patterns) do
+            local figurepatterns = figures.patterns
+            for i=1,#figurepatterns do
+                local pattern = figurepatterns[i]
                 if find(askedformat,pattern[1]) then
                     format = pattern[2]
                     break
@@ -400,7 +475,7 @@ local function locate(request) -- name, format, cache
             end
         end
         if format then
-            local foundname = figures.exists(askedname,format) -- not askedformat
+            local foundname = figures.exists(askedname,format,resolve_too) -- not askedformat
             if foundname then
                 return register(askedname, {
                     askedname = askedname,
@@ -408,29 +483,36 @@ local function locate(request) -- name, format, cache
                     format = format,
                     cache = askedcache,
                     foundname = foundname,
+                    conversion = askedconversion,
                 })
             end
         end
         if askedpath then
             -- path and type given, todo: strip pieces of path
-            if figures.exists(askedname,askedformat) then
+            if figures.exists(askedname,askedformat,resolve_too) then
                 return register(askedname, {
                     askedname = askedname,
                     fullname = askedname,
                     format = askedformat,
                     cache = askedcache,
+                    conversion = askedconversion,
                 })
             end
         else
             -- type given
-            for _, path in ipairs(figures.paths) do
+            local figurepaths = figures.paths
+            for i=1,#figurepaths do
+                local path = figurepaths[i]
                 local check = path .. "/" .. askedname
-                if figures.exists(check,askedformat) then
+             -- we pass 'true' as it can be an url as well, as the type
+             -- is given we don't waste much time
+                if figures.exists(check,askedformat,resolve_too) then
                     return register(check, {
                         askedname = askedname,
                         fullname = check,
                         format = askedformat,
                         cache = askedcache,
+                        conversion = askedconversion,
                     })
                 end
             end
@@ -442,6 +524,7 @@ local function locate(request) -- name, format, cache
                         fullname = check,
                         format = askedformat,
                         cache = askedcache,
+                        conversion = askedconversion,
                     })
                 end
             end
@@ -450,16 +533,20 @@ local function locate(request) -- name, format, cache
         if trace_figures then
             commands.writestatus("figures","strategy: rootbased path")
         end
-        for _, format in ipairs(figures.order) do
+        local figureorder = figures.order
+        for i=1,#figureorder do
+            local format = figureorder[i]
             local list = figures.formats[format].list or { format }
-            for _, suffix in ipairs(list) do
+            for j=1,#list do
+                local suffix = list[j]
                 local check = file.addsuffix(askedname,suffix)
-                if figures.exists(check,format) then
+                if figures.exists(check,format,resolve_too) then
                     return register(askedname, {
                         askedname = askedname,
                         fullname = check,
                         format = format,
                         cache = askedcache,
+                        conversion = askedconversion,
                     })
                 end
             end
@@ -469,19 +556,30 @@ local function locate(request) -- name, format, cache
             if trace_figures then
                 commands.writestatus("figures","strategy: unknown format, prefer quality")
             end
-            for _, format in ipairs(figures.order) do
+            local figurepaths = figures.paths
+            local figureorder = figures.order
+            for j=1,#figureorder do
+                local format = figureorder[j]
                 local list = figures.formats[format].list or { format }
-                for _, suffix in ipairs(list) do
---~                     local name = file.replacesuffix(askedbase,suffix)
+                for k=1,#list do
+                    local suffix = list[k]
+                 -- local name = file.replacesuffix(askedbase,suffix)
                     local name = file.replacesuffix(askedname,suffix)
-                    for _, path in ipairs(figures.paths) do
+                    for i=1,#figurepaths do
+                        local path = figurepaths[i]
                         local check = path .. "/" .. name
-                        if figures.exists(check,format) then
+                        local isfile = url.hashed(check).scheme == "file"
+                        if not isfile then
+                            if trace_figures then
+                                commands.writestatus("figures","warning: skipping path %s",path)
+                            end
+                        elseif figures.exists(check,format,true) then
                             return register(askedname, {
                                 askedname = askedname,
                                 fullname = check,
                                 format = format,
                                 cache = askedcache,
+                                conversion = askedconversion,
                             })
                         end
                     end
@@ -491,17 +589,23 @@ local function locate(request) -- name, format, cache
             if trace_figures then
                 commands.writestatus("figures","strategy: unknown format, prefer path")
             end
-            for _, path in ipairs(figures.paths) do
-                for _, format in ipairs(figures.order) do
+            local figurepaths = figures.paths
+            local figureorder = figures.order
+            for i=1,#figurepaths do
+                local path = figurepaths[i]
+                for j=1,#figureorder do
+                    local format = figureorder[j]
                     local list = figures.formats[format].list or { format }
-                    for _, suffix in ipairs(list) do
+                    for k=1,#list do
+                        local suffix = list[k]
                         local check = path .. "/" .. file.replacesuffix(askedbase,suffix)
-                        if figures.exists(check,format) then
+                        if figures.exists(check,format,resolve_too) then
                             return register(askedname, {
                                 askedname = askedname,
                                 fullname = check,
                                 format = format,
                                 cache = askedcache,
+                                conversion = askedconversion,
                             })
                         end
                     end
@@ -512,9 +616,12 @@ local function locate(request) -- name, format, cache
             if trace_figures then
                 commands.writestatus("figures","strategy: default tex path")
             end
-            for _, format in ipairs(figures.order) do
+            local figureorder = figures.order
+            for j=1,#figureorder do
+                local format = figureorder[j]
                 local list = figures.formats[format].list or { format }
-                for _, suffix in ipairs(list) do
+                for k=1,#list do
+                    local suffix = list[k]
                     local check = resolvers.find_file(file.replacesuffix(askedname,suffix))
                     if check and check ~= "" then
                         return register(askedname, {
@@ -522,6 +629,7 @@ local function locate(request) -- name, format, cache
                             fullname = check,
                             format = format,
                             cache = askedcache,
+                            conversion = askedconversion,
                         })
                     end
                 end
@@ -538,6 +646,7 @@ figures.checkers    = figures.checkers    or { }
 figures.includers   = figures.includers   or { }
 figures.converters  = figures.converters  or { }
 figures.identifiers = figures.identifiers or { }
+figures.programs    = figures.programs    or { }
 
 figures.identifiers.list = {
     figures.identifiers.default
@@ -547,20 +656,22 @@ function figures.identifiers.default(data)
     local dr, du, ds = data.request, data.used, data.status
     local l = locate(dr)
     local foundname = l.foundname
-    local fullname = l.fullname or foundname
+    local fullname  = l.fullname or foundname
     if fullname then
-        du.format = l.format or false
+        du.format   = l.format or false
         du.fullname = fullname -- can be cached
         ds.fullname = foundname -- original
-        ds.format = l.format
-        ds.status = (l.found and 10) or 0
+        ds.format   = l.format
+        ds.status   = (l.found and 10) or 0
     end
     return data
 end
 
 function figures.identify(data)
     data = data or figures.current()
-    for _, identifier in ipairs(figures.identifiers.list) do
+    local list = figures.identifiers.list
+    for i=1,#list do
+        local identifier = list[i]
         data = identifier(data)
         if data.status.status > 0 then
             break
@@ -568,8 +679,8 @@ function figures.identify(data)
     end
     return data
 end
-function figures.exists(askedname,format)
-    return (figures.existers[format] or figures.existers.generic)(askedname)
+function figures.exists(askedname,format,resolve)
+    return (figures.existers[format] or figures.existers.generic)(askedname,resolve)
 end
 function figures.check(data)
     data = data or figures.current()
@@ -588,23 +699,30 @@ end
 function figures.done(data)
     figures.n = figures.n + 1
     data = data or figures.current()
-    local dr, du, ds = data.request, data.used, data.status
-    ds.width = tex.wd[figures.boxnumber]
-    ds.height = tex.ht[figures.boxnumber]
-    ds.xscale = ds.width/(du.width or 1)
+--~ print(table.serialize(figures.current()))
+    local dr, du, ds, nr = data.request, data.used, data.status, figures.boxnumber
+    local box = texbox[nr]
+    ds.width  = box.width
+    ds.height = box.height
+    ds.xscale = ds.width /(du.width  or 1)
     ds.yscale = ds.height/(du.height or 1)
+    ds.page   = ds.page or du.page or dr.page -- sort of redundant but can be limited
+--~ print(table.serialize(figures.current()))
     return data
 end
 
 function figures.dummy(data)
---~         data = data or figures.current()
---~         local dr, du, ds = data.request, data.used, data.status
---~         local r = node.new("rule")
---~         r.width  = du.width  or figures.defaultwidth
---~         r.height = du.height or figures.defaultheight
---~         r.depth  = du.depth  or figures.defaultdepth
---~         tex.box[figures.boxnumber] = node.hpack(node.write(r))
-    texsprint(ctxcatcodes,"\\emptyfoundexternalfigure")
+    data = data or figures.current()
+    local dr, du, ds, nr = data.request, data.used, data.status, figures.boxnumber
+    local box = node.hpack(node.new("hlist")) -- we need to set the dir (luatex 0.60 buglet)
+    du.width  = du.width  or figures.defaultwidth
+    du.height = du.height or figures.defaultheight
+    du.depth  = du.depth  or figures.defaultdepth
+ -- box.dir = "TLT"
+    box.width  = du.width
+    box.height = du.height
+    box.depth  = du.depth
+    texbox[nr] = box -- hm, should be global (to be checked for consistency)
 end
 
 -- -- -- generic -- -- --
@@ -630,7 +748,11 @@ end
 function figures.checkers.generic(data)
     local dr, du, ds = data.request, data.used, data.status
     local name, page, size, color = du.fullname or "unknown generic", du.page or dr.page, dr.size or "crop", dr.color or "natural"
-    local hash = name .. "->" .. page .. "->" .. size .. "->" .. color
+    local conversion = dr.conversion
+    if not conversion or conversion == "" then
+        conversion = "unknown"
+    end
+    local hash = name .. "->" .. page .. "->" .. size .. "->" .. color .. "->" .. conversion
     local figure = figures.loaded[hash]
     if figure == nil then
         figure = img.new { filename = name, page = page, pagebox = dr.size }
@@ -639,6 +761,13 @@ function figures.checkers.generic(data)
         local f, d = backends.codeinjections.setfigurealternative(data,figure)
         figure, data = f or figure, d or data
         figures.loaded[hash] = figure
+        if trace_conversion then
+            report_graphics("new graphic, hash: %s",hash)
+        end
+    else
+        if trace_conversion then
+            report_graphics("existing graphic, hash: %s",hash)
+        end
     end
     if figure then
         du.width = figure.width
@@ -665,10 +794,11 @@ function figures.includers.generic(data)
         figures.used[hash] = figure
     end
     if figure then
-        local n = figures.boxnumber
-        tex.box[n] = node.hpack(img.node(figure))
-     -- tex.box[n] = img.node(figure) -- img.write(figure) -- assigning img.node directly no longer valid
-        tex.wd[n], tex.ht[n], tex.dp[n] = figure.width, figure.height, 0 -- new, hm, tricky, we need to do that in tex (yet)
+        local nr = figures.boxnumber
+        -- it looks like we have a leak in attributes here .. todo
+        local box = node.hpack(img.node(figure)) -- img.node(figure) not longer valid
+        box.width, box.height, box.depth = figure.width, figure.height, 0 -- new, hm, tricky, we need to do that in tex (yet)
+        texbox[nr] = box
         ds.objectnumber = figure.objnum
         texsprint(ctxcatcodes,"\\relocateexternalfigure")
     end
@@ -701,22 +831,29 @@ end
 
 function figures.checkers.mov(data)
     local dr, du, ds = data.request, data.used, data.status
-    dr.width = (dr.width or figures.defaultwidth):todimen()
-    dr.height = (dr.height or figures.defaultheight):todimen()
-    du.width = dr.width
-    du.height = dr.height
-    du.foundname = du.fullname
-    local code = backends.codeinjections.insertmovie {
-        width      = du.width or dr.width,
-        height     = du.height or dr.height,
-        factor     = number.dimenfactors.bp,
-        ["repeat"] = dr["repeat"],
-        controls   = dr.controls,
-        preview    = dr.preview,
-        label      = dr.label,
-        foundname  = du.foundname,
-    }
-    texsprint(ctxcatcodes,format("\\startfoundexternalfigure{%ssp}{%ssp}%s\\stopfoundexternalfigure",du.width,du.height,code))
+    local width = todimen(dr.width or figures.defaultwidth)
+    local height = todimen(dr.height or figures.defaultheight)
+    local foundname = du.fullname
+    dr.width, dr.height = width, height
+    du.width, du.height, du.foundname = width, height, foundname
+    if trace_inclusion then
+        report_graphics("including movie '%s': width %s, height %s",foundname,width,height)
+    end
+    -- we need to push the node.write in between ... we could make a shared helper for this
+    context.startfoundexternalfigure(width .. "sp",height .. "sp")
+    context(function()
+        backends.codeinjections.insertmovie {
+            width      = width,
+            height     = height,
+            factor     = number.dimenfactors.bp,
+            ["repeat"] = dr["repeat"],
+            controls   = dr.controls,
+            preview    = dr.preview,
+            label      = dr.label,
+            foundname  = foundname,
+        }
+    end)
+    context.stopfoundexternalfigure()
     return data
 end
 
@@ -773,35 +910,100 @@ function figures.checkers.tex(data)
 end
 figures.includers.tex = figures.includers.nongeneric
 
+-- -- -- converters -- -- --
+
+local function makeoptions(program)
+    local to = type(options)
+    return (to == "table" and concat(options," ")) or (to == "string" and options) or ""
+end
+
+local function runprogram(...)
+    local command = format(...)
+    if trace_conversion or trace_programs then
+        report_graphics("running %s",command)
+    end
+    os.spawn(command)
+end
+
 -- -- -- eps -- -- --
 
-function figures.converters.eps(oldname,newname)
-    -- hack, we need a lua based converter script, or better, we should use
-    -- rlx as alternative
-    local outputpath = file.dirname(newname)
-    local outputbase = file.basename(newname)
-    if outputpath == "" then outputpath = "." end
-    local command = format("mtxrun bin:pstopdf --outputpath=%s %s",outputpath,oldname)
-    os.spawn(command)
+local epsconverter     = { }
+figures.converters.eps = epsconverter
+
+figures.programs.gs = {
+    options = {
+        "-dAutoRotatePages=/None",
+        "-dPDFSETTINGS=/prepress",
+        "-dEPSCrop",
+    },
+    command = (os.type == "windows" and "gswin32") or "gs"
+}
+
+function epsconverter.pdf(oldname,newname)
+    local gs = figures.programs.gs
+    runprogram (
+        '%s -q -sDEVICE=pdfwrite -dNOPAUSE -dNOCACHE -dBATCH %s -sOutputFile="%s" "%s" -c quit',
+        gs.command, makeoptions(gs.options), newname, oldname
+    )
 end
 
-figures.converters.svg = figures.converters.eps
+epsconverter.default = epsconverter.pdf
 
-function figures.converters.gif(oldname,newname)
-    -- hack, we need a lua based converter script, or better, we should use
-    -- rlx as alternative
-    local command = format("mtxrun bin:convert %s %s",oldname,newname)
-    os.spawn(command)
+-- -- -- svg -- -- --
+
+local svgconverter      = { }
+figures.converters.svg  = svgconverter
+figures.converters.svgz = svgconverter
+
+-- inkscape on windows only works with complete paths
+
+figures.programs.inkscape = {
+    options = {
+        "--export-dpi=600"
+    },
+    command = "inkscape"
+}
+
+function svgconverter.pdf(oldname,newname)
+    local inkscape = figures.programs.inkscape
+    runprogram (
+        '%s "%s" --export-pdf="%s" %s',
+        inkscape.command, oldname, newname, makeoptions(inkscape.options)
+    )
 end
 
--- -- -- lowres -- -- --
+function svgconverter.png(oldname,newname)
+    local inkscape = figures.programs.inkscape
+    runprogram (
+        '%s "%s" --export-png="%s" %s',
+        inkscape.command, oldname, newname, makeoptions(inkscape.options)
+    )
+end
 
---~ function figures.converters.pdf(oldname,newname)
---~     local outputpath = file.dirname(newname)
---~     local outputbase = file.basename(newname)
---~     local command = format("mtxrun bin:pstopdf --method=4 --outputpath=%s %s",outputpath,oldname)
---~     os.spawn(command)
---~ end
+svgconverter.default = svgconverter.pdf
+
+-- -- -- gif -- -- --
+
+local gifconverter     = { }
+figures.converters.gif = gifconverter
+
+figures.programs.convert = {
+    command = "convert" -- imagemagick
+}
+
+function gifconverter.pdf(oldname,newname)
+    local convert = figures.programs.convert
+    runprogram (
+        "%s %s %s %s",
+        convert.command, makeoptions(convert.options), oldname, newname
+    )
+end
+
+gifconverter.default = gifconverter.pdf
+
+-- todo: lowres
+
+-- -- -- bases -- -- --
 
 figures.bases         = { }
 figures.bases.list    = { } -- index      => { basename, fullname, xmlroot }
@@ -842,7 +1044,9 @@ function bases.find(basename,askedlabel)
         local page = 0
         if base[2] == nil then
             -- no yet located
-            for _, path in ipairs(figures.paths) do
+            local figurepaths = figures.paths
+            for i=1,#figurepaths do
+                local path = figurepaths[i]
                 local xmlfile = path .. "/" .. basename
                 if io.exists(xmlfile) then
                     base[2] = xmlfile
@@ -883,7 +1087,9 @@ end
 -- we can access sequential or by name
 
 function bases.locate(askedlabel)
-    for _, entry in ipairs(bases.list) do
+    local list = bases.list
+    for i=1,#list do
+        local entry = list[i]
         local t = bases.find(entry[1],askedlabel)
         if t then
             return t
