@@ -6,10 +6,22 @@ if not modules then modules = { } end modules ['l-aux'] = {
     license   = "see context related readme files"
 }
 
+-- for inline, no store split : for s in string.gmatch(str,",* *([^,]+)") do .. end
+
 aux = aux or { }
 
 local concat, format, gmatch = table.concat, string.format, string.gmatch
 local tostring, type = tostring, type
+local lpegmatch = lpeg.match
+
+local P, R, V = lpeg.P, lpeg.R, lpeg.V
+
+local escape, left, right = P("\\"), P('{'), P('}')
+
+lpeg.patterns.balanced = P {
+    [1] = ((escape * (left+right)) + (1 - (left+right)) + V(2))^0,
+    [2] = left * V(1) * right
+}
 
 local space     = lpeg.P(' ')
 local equal     = lpeg.P("=")
@@ -17,7 +29,7 @@ local comma     = lpeg.P(",")
 local lbrace    = lpeg.P("{")
 local rbrace    = lpeg.P("}")
 local nobrace   = 1 - (lbrace+rbrace)
-local nested    = lpeg.P{ lbrace * (nobrace + lpeg.V(1))^0 * rbrace }
+local nested    = lpeg.P { lbrace * (nobrace + lpeg.V(1))^0 * rbrace }
 local spaces    = space^0
 
 local value     = lpeg.P(lbrace * lpeg.C((nobrace + nested)^0) * rbrace) + lpeg.C((nested + (1-comma))^0)
@@ -58,11 +70,7 @@ end
 function aux.settings_to_hash(str,existing)
     if str and str ~= "" then
         hash = existing or { }
-        if moretolerant then
-            pattern_b_s:match(str)
-        else
-            pattern_a_s:match(str)
-        end
+        lpegmatch(pattern_a_s,str)
         return hash
     else
         return { }
@@ -72,7 +80,7 @@ end
 function aux.settings_to_hash_tolerant(str,existing)
     if str and str ~= "" then
         hash = existing or { }
-        pattern_b_s:match(str)
+        lpegmatch(pattern_b_s,str)
         return hash
     else
         return { }
@@ -82,26 +90,28 @@ end
 function aux.settings_to_hash_strict(str,existing)
     if str and str ~= "" then
         hash = existing or { }
-        pattern_c_s:match(str)
+        lpegmatch(pattern_c_s,str)
         return next(hash) and hash
     else
         return nil
     end
 end
 
-local seperator = comma * space^0
+local separator = comma * space^0
 local value     = lpeg.P(lbrace * lpeg.C((nobrace + nested)^0) * rbrace) + lpeg.C((nested + (1-comma))^0)
-local pattern   = lpeg.Ct(value*(seperator*value)^0)
+local pattern   = lpeg.Ct(value*(separator*value)^0)
 
 -- "aap, {noot}, mies" : outer {} removes, leading spaces ignored
 
 aux.settings_to_array_pattern = pattern
 
+-- we could use a weak table as cache
+
 function aux.settings_to_array(str)
     if not str or str == "" then
         return { }
     else
-        return pattern:match(str)
+        return lpegmatch(pattern,str)
     end
 end
 
@@ -110,10 +120,10 @@ local function set(t,v)
 end
 
 local value   = lpeg.P(lpeg.Carg(1)*value) / set
-local pattern = value*(seperator*value)^0 * lpeg.Carg(1)
+local pattern = value*(separator*value)^0 * lpeg.Carg(1)
 
 function aux.add_settings_to_array(t,str)
-    return pattern:match(str, nil, t)
+    return lpegmatch(pattern,str,nil,t)
 end
 
 function aux.hash_to_string(h,separator,yes,no,strict,omit)
@@ -161,6 +171,13 @@ function aux.settings_to_set(str,t)
     return t
 end
 
+local value     = lbrace * lpeg.C((nobrace + nested)^0) * rbrace
+local pattern   = lpeg.Ct((space + value)^0)
+
+function aux.arguments_to_table(str)
+    return lpegmatch(pattern,str)
+end
+
 -- temporary here
 
 function aux.getparameters(self,class,parentclass,settings)
@@ -174,31 +191,26 @@ end
 
 -- temporary here
 
-local digit    = lpeg.R("09")
-local period   = lpeg.P(".")
-local zero     = lpeg.P("0")
-
---~ local finish   = lpeg.P(-1)
---~ local nodigit  = (1-digit) + finish
---~ local case_1   = (period * zero^1 * #nodigit)/"" -- .000
---~ local case_2   = (period * (1-(zero^0/"") * #nodigit)^1 * (zero^0/"") * nodigit) -- .010 .10 .100100
-
+local digit         = lpeg.R("09")
+local period        = lpeg.P(".")
+local zero          = lpeg.P("0")
 local trailingzeros = zero^0 * -digit -- suggested by Roberto R
-local case_1 = period * trailingzeros / ""
-local case_2 = period * (digit - trailingzeros)^1 * (trailingzeros / "")
-
-local number   = digit^1 * (case_1 + case_2)
-local stripper = lpeg.Cs((number + 1)^0)
+local case_1        = period * trailingzeros / ""
+local case_2        = period * (digit - trailingzeros)^1 * (trailingzeros / "")
+local number        = digit^1 * (case_1 + case_2)
+local stripper      = lpeg.Cs((number + 1)^0)
 
 --~ local sample = "bla 11.00 bla 11 bla 0.1100 bla 1.00100 bla 0.00 bla 0.001 bla 1.1100 bla 0.100100100 bla 0.00100100100"
 --~ collectgarbage("collect")
 --~ str = string.rep(sample,10000)
 --~ local ts = os.clock()
---~ stripper:match(str)
---~ print(#str, os.clock()-ts, stripper:match(sample))
+--~ lpegmatch(stripper,str)
+--~ print(#str, os.clock()-ts, lpegmatch(stripper,sample))
+
+lpeg.patterns.strip_zeros = stripper
 
 function aux.strip_zeros(str)
-    return stripper:match(str)
+    return lpegmatch(stripper,str)
 end
 
 function aux.definetable(target) -- defines undefined tables
@@ -222,6 +234,10 @@ function aux.accesstable(target)
     return t
 end
 
+--~ function string.commaseparated(str)
+--~     return gmatch(str,"([^,%s]+)")
+--~ end
+
 -- as we use this a lot ...
 
 --~ function aux.cachefunction(action,weak)
