@@ -1,11 +1,14 @@
--- filename : l-boolean.lua
--- comment  : split off from luat-lib
--- author   : Hans Hagen, PRAGMA-ADE, Hasselt NL
--- copyright: PRAGMA ADE / ConTeXt Development Team
--- license  : see context related readme files
+if not modules then modules = { } end modules ['l-boolean'] = {
+    version   = 1.001,
+    comment   = "companion to luat-lib.mkiv",
+    author    = "Hans Hagen, PRAGMA-ADE, Hasselt NL",
+    copyright = "PRAGMA ADE / ConTeXt Development Team",
+    license   = "see context related readme files"
+}
 
-if not versions then versions = { } end versions['l-boolean'] = 1.001
-if not boolean  then boolean  = { } end
+boolean = boolean or { }
+
+local type, tonumber = type, tonumber
 
 function boolean.tonumber(b)
     if b then return 1 else return 0 end
@@ -15,7 +18,7 @@ function toboolean(str,tolerant)
     if tolerant then
         local tstr = type(str)
         if tstr == "string" then
-            return str == "true" or str == "yes" or str == "on" or str == "1"
+            return str == "true" or str == "yes" or str == "on" or str == "1" or str == "t"
         elseif tstr == "number" then
             return tonumber(str) ~= 0
         elseif tstr == "nil" then
@@ -32,15 +35,15 @@ function toboolean(str,tolerant)
     end
 end
 
-function string.is_boolean(str)
+function string.is_boolean(str,default)
     if type(str) == "string" then
-        if str == "true" or str == "yes" or str == "on" then
+        if str == "true" or str == "yes" or str == "on" or str == "t" then
             return true
-        elseif str == "false" or str == "no" or str == "off" then
+        elseif str == "false" or str == "no" or str == "off" or str == "f" then
             return false
         end
     end
-    return nil
+    return default
 end
 
 function boolean.alwaystrue()
