@@ -13,6 +13,10 @@ local trace_references   = false  trackers.register("references.references",   f
 local trace_destinations = false  trackers.register("references.destinations", function(v) trace_destinations = v end)
 local trace_bookmarks    = false  trackers.register("references.bookmarks",    function(v) trace_bookmarks    = v end)
 
+local report_references   = logs.new("references")
+local report_destinations = logs.new("destinations")
+local report_bookmarks    = logs.new("bookmarks")
+
 local variables = interfaces.variables
 local constants = interfaces.constants
 
@@ -31,18 +35,17 @@ local specials  = jobreferences.specials
 local handlers  = jobreferences.handlers
 local executers = jobreferences.executers
 
-local pdfdictionary = lpdf.dictionary
-local pdfarray      = lpdf.array
-local pdfreference  = lpdf.reference
-local pdfunicode    = lpdf.unicode
-local pdfconstant   = lpdf.constant
+local pdfdictionary       = lpdf.dictionary
+local pdfarray            = lpdf.array
+local pdfreference        = lpdf.reference
+local pdfunicode          = lpdf.unicode
+local pdfconstant         = lpdf.constant
+local pdfflushobject      = lpdf.flushobject
+local pdfreserveobject    = lpdf.reserveobject
+local pdfpagereference    = lpdf.pagereference
 
-local pdfreserveobj   = pdf.reserveobj
-local pdfimmediateobj = pdf.immediateobj
-local pdfpageref      = tex.pdfpageref
-
-local pdfannot  = nodes.pdfannot
-local pdfdest   = nodes.pdfdest
+local pdfannotation_node  = nodes.pdfannotation
+local pdfdestination_node = nodes.pdfdestination
 
 local pdf_uri        = pdfconstant("URI")
 local pdf_gotor      = pdfconstant("GoToR")
@@ -57,20 +60,24 @@ local pdf_border     = pdfarray { 0, 0, 0 }
 local cache = { }
 
 local function pagedest(n)
-    local pd = cache[n]
-    if not pd then
-        local a = pdfarray {
-            pdfreference(pdfpageref(n)),
-            pdfconstant("Fit")
-        }
-        pd = pdfreference(pdfimmediateobj(tostring(a)))
-        cache[n] = pd
+    if n > 0 then
+        local pd = cache[n]
+        if not pd then
+            local a = pdfarray {
+                pdfreference(pdfpagereference(n)),
+                pdfconstant("Fit")
+            }
+            pd = pdfreference(pdfflushobject(a))
+            cache[n] = pd
+        end
+        return pd
     end
-    return pd
 end
 
 lpdf.pagedest = pagedest
 
+local defaultdestination = pdfarray { 0, pdfconstant("Fit") }
+
 local function link(url,filename,destination,page,actions)
     if filename and filename ~= "" then
         if file.basename(filename) == tex.jobname then
@@ -92,10 +99,17 @@ local function link(url,filename,destination,page,actions)
             URI = url,
         }
     elseif filename and filename ~= "" then
+        -- no page ?
+        if destination == "" then
+            destination = nil
+        end
+        if not destination and page then
+            destination = pdfarray { page - 1, pdfconstant("Fit") }
+        end
         return pdfdictionary {
-            S = pdf_gotor,
+            S = pdf_gotor, -- can also be pdf_launch
             F = filename,
-            D = destination and destination ~= "" and destination,
+            D = destination or defaultdestination, -- D is mandate
             NewWindow = (actions.newwindow and true) or nil,
         }
     elseif destination and destination ~= "" then
@@ -190,7 +204,7 @@ local function pdfaction(actions)
     end
 end
 
-lpdf.pdfaction = pdfaction
+lpdf.action = pdfaction
 
 function codeinjections.prerollreference(actions)
     local main = actions and pdfaction(actions)
@@ -200,6 +214,7 @@ function codeinjections.prerollreference(actions)
             Border  = pdf_border,
             H       = (not actions.highlight and pdf_n) or nil,
             A       = main,
+            F       = 4, -- print (mandate in pdf/a)
         --  does not work at all in spite of specification
         --  OC      = (actions.layer and lpdf.layerreferences[actions.layer]) or nil,
         --  OC      = backends.pdf.layerreference(actions.layer),
@@ -221,7 +236,7 @@ end
 --         }
 --         local cm = cache[main]
 --         if not cm then
---             cm = "/A ".. tostring(pdfreference(pdfimmediateobj(tostring(main))))
+--             cm = "/A ".. tostring(pdfreference(pdfflushobject(main))
 --             cache[main] = cm
 --         end
 --         return cm
@@ -230,24 +245,18 @@ end
 
 function nodeinjections.reference(width,height,depth,prerolled)
     if prerolled then
-        if swapdir then
-            width = - width
-        end
         if trace_references then
-            logs.report("references","w=%s, h=%s, d=%s, a=%s",width,height,depth,prerolled)
+            report_references("w=%s, h=%s, d=%s, a=%s",width,height,depth,prerolled)
         end
-        return pdfannot(width,height,depth,prerolled)
+        return pdfannotation_node(width,height,depth,prerolled)
     end
 end
 
 function nodeinjections.destination(width,height,depth,name,view)
-    if swapdir then
-        width = - width
-    end
     if trace_destinations then
-        logs.report("destinations","w=%s, h=%s, d=%s, n=%s, v=%s",width,height,depth,name,view or "no view")
+        report_destinations("w=%s, h=%s, d=%s, n=%s, v=%s",width,height,depth,name,view or "no view")
     end
-    return pdfdest(width,height,depth,name,view)
+    return pdfdestination_node(width,height,depth,name,view)
 end
 
 -- runners and specials
@@ -266,17 +275,18 @@ runners["inner"] = function(var,actions)
 end
 
 runners["inner with arguments"] = function(var,actions)
-    logs.report("references","todo: inner with arguments")
+    report_references("todo: inner with arguments")
     return false
 end
 
 runners["outer"] = function(var,actions)
-    return link(nil,var.f,nil,nil,actions) -- var.o ?
+    local file, url = jobreferences.checkedfileorurl(var.outer,var.outer)
+    return link(url,file,var.arguments,nil,actions)
 end
 
 runners["outer with inner"] = function(var,actions)
-    -- todo: resolve url/file name
-    return link(nil,var.f,var.inner,var.r,actions)
+    local file = jobreferences.checkedfile(var.f)
+    return link(nil,file,var.inner,var.r,actions)
 end
 
 runners["special outer with operation"] = function(var,actions)
@@ -285,7 +295,7 @@ runners["special outer with operation"] = function(var,actions)
 end
 
 runners["special outer"] = function(var,actions)
-    logs.report("references","todo: special outer")
+    report_references("todo: special outer")
     return false
 end
 
@@ -295,22 +305,22 @@ runners["special"] = function(var,actions)
 end
 
 runners["outer with inner with arguments"] = function(var,actions)
-    logs.report("references","todo: outer with inner with arguments")
+    report_references("todo: outer with inner with arguments")
     return false
 end
 
 runners["outer with special and operation and arguments"] = function(var,actions)
-    logs.report("references","todo: outer with special and operation and arguments")
+    report_references("todo: outer with special and operation and arguments")
     return false
 end
 
 runners["outer with special"] = function(var,actions)
-    logs.report("references","todo: outer with special")
+    report_references("todo: outer with special")
     return false
 end
 
 runners["outer with special and operation"] = function(var,actions)
-    logs.report("references","todo: outer with special and operation")
+    report_references("todo: outer with special and operation")
     return false
 end
 
@@ -336,11 +346,8 @@ specials.i = specials.internal
 function specials.page(var,actions) -- better resolve in strc-ref
     local file = var.f
     if file then
-        local f = jobreferences.files.data[file]
-        if f then
-            file = f[1] or file
-        end
-        return link(nil,file,nil,p or var.operation,actions)
+        file = jobreferences.checkedfile(file)
+        return link(nil,file,nil,var.operation,actions)
     else
         local p = jobreferences.pages[var.operation]
         if type(p) == "function" then
@@ -350,62 +357,42 @@ function specials.page(var,actions) -- better resolve in strc-ref
     end
 end
 
-function specials.url(var,actions) -- better resolve in strc-ref
-    local url = var.operation
-    if url then
-        local u = jobreferences.urls.data[url]
-        if u then
-            local u, f = u[1], u[2]
-            if f and f ~= "" then
-                url = u .. "/" .. f
-            else
-                url = u
-            end
+-- todo, do this in references namespace ordered instead (this is an experiment)
+
+local splitter = lpeg.splitat(":")
+
+function specials.order(var,actions) -- jobreferences.specials !
+    local operation = var.operation
+    if operation then
+        local kind, name, n = lpegmatch(splitter,operation)
+        local order = lists.ordered[kind]
+        order = order and order[name]
+        local v = order[tonumber(n)]
+        local r = v and v.references.realpage
+        if r then
+            var.operation = r -- brrr, but test anyway
+            return specials.page(var,actions)
         end
     end
+end
+
+function specials.url(var,actions)
+    local url = jobreferences.checkedurl(var.operation)
     return link(url,nil,var.arguments,nil,actions)
 end
 
-function specials.file(var,actions) -- better resolve in strc-ref
-    local file = var.operation
-    if file then
-        local f = jobreferences.files.data[file]
-        if f then
-            file = f[1] or file
-        end
-    end
+function specials.file(var,actions)
+    local file = jobreferences.checkedfile(var.operation)
     return link(nil,file,var.arguments,nil,actions)
 end
 
-function specials.fileorurl(var,actions) -- better resolve in strc-ref
-    local whatever, url, file = var.operation, nil, nil
-    if whatever then
-        local w = jobreferences.files.data[whatever]
-        if w then
-            file = w[1]
-        else
-            w = jobreferences.urls.data[whatever]
-            if w then
-                local u, f = w[1], w[2]
-                if f and f ~= "" then
-                    url = u .. "/" .. f
-                else
-                    url = u
-                end
-            end
-        end
-    end
+function specials.fileorurl(var,actions)
+    local file, url = jobreferences.checkedfileorurl(var.operation,var.operation)
     return link(url,file,var.arguments,nil,actions)
 end
 
-function specials.program(var,content) -- better resolve in strc-ref
-    local program = var.operation
-    if program then
-        local p = jobreferences.programs[program]
-        if p then
-            program = p[1]
-        end
-    end
+function specials.program(var,content)
+    local program = jobreferences.checkedprogram(var.operation)
     return lpdf.launch(program,var.arguments)
 end
 
@@ -549,13 +536,13 @@ local function build(levels,start,parent,method)
         local level, title, reference, open = li[1], li[2], li[3], li[4]
         if level == startlevel then
             if trace_bookmarks then
-                logs.report("bookmark","%3i %s%s %s",realpage,rep("  ",level-1),(open and "+") or "-",title)
+                report_bookmarks("%3i %s%s %s",reference.realpage,rep("  ",level-1),(open and "+") or "-",title)
             end
             local prev = child
-            child = pdfreserveobj()
+            child = pdfreserveobject()
             if entry then
                 entry.Next = child and pdfreference(child)
-                pdfimmediateobj(prev,tostring(entry))
+                pdfflushobject(prev,entry)
             end
             entry = pdfdictionary {
                 Title  = pdfunicode(title),
@@ -573,7 +560,7 @@ local function build(levels,start,parent,method)
             n = n + 1
             i = i + 1
         elseif level < startlevel then
-            pdfimmediateobj(child,tostring(entry))
+            pdfflushobject(child,entry)
             return i, n, first, last
         elseif i < #levels and level > startlevel then
             i, m, f, l = build(levels,i,pdfreference(child),method)
@@ -588,16 +575,16 @@ local function build(levels,start,parent,method)
             if m > 0 then
                 entry.First, entry.Last = pdfreference(f), pdfreference(l)
             end
-            pdfimmediateobj(child,tostring(entry))
+            pdfflushobject(child,entry)
             return i, n, first, last
         end
     end
-    pdfimmediateobj(child,tostring(entry))
+    pdfflushobject(child,entry)
     return nil, n, first, last
 end
 
 function codeinjections.addbookmarks(levels,method)
-    local parent = pdfreserveobj()
+    local parent = pdfreserveobject()
     local _, m, first, last = build(levels,1,pdfreference(parent),method or "internal")
     local dict = pdfdictionary {
         Type  = pdfconstant("Outlines"),
@@ -605,6 +592,6 @@ function codeinjections.addbookmarks(levels,method)
         Last  = pdfreference(last),
         Count = m,
     }
-    pdfimmediateobj(parent,tostring(dict))
+    pdfflushobject(parent,dict)
     lpdf.addtocatalog("Outlines",lpdf.reference(parent))
 end
