@@ -11,9 +11,12 @@ if not modules then modules = { } end modules ['lpdf-fld'] = {
 -- some optimizations removed (will come bakc if needed)
 
 local gmatch, lower, format = string.gmatch, string.lower, string.format
+local lpegmatch = lpeg.match
 
 local trace_fields = false  trackers.register("widgets.fields",   function(v) trace_fields   = v end)
 
+local report_fields = logs.new("fields")
+
 local texsprint, ctxcatcodes = tex.sprint, tex.ctxcatcodes
 
 local variables = interfaces.variables
@@ -24,17 +27,18 @@ local registrations  = backends.pdf.registrations
 
 local registeredsymbol = codeinjections.registeredsymbol
 
-local pdfstream     = lpdf.stream
-local pdfdictionary = lpdf.dictionary
-local pdfarray      = lpdf.array
-local pdfreference  = lpdf.reference
-local pdfunicode    = lpdf.unicode
-local pdfstring     = lpdf.string
-local pdfconstant   = lpdf.constant
-local pdftoeight    = lpdf.toeight
+local pdfstream          = lpdf.stream
+local pdfdictionary      = lpdf.dictionary
+local pdfarray           = lpdf.array
+local pdfreference       = lpdf.reference
+local pdfunicode         = lpdf.unicode
+local pdfstring          = lpdf.string
+local pdfconstant        = lpdf.constant
+local pdftoeight         = lpdf.toeight
+local pdfflushobject     = lpdf.flushobject
+local pdfreserveobject   = lpdf.reserveobject
 
-local pdfimmediateobj = pdf.immediateobj
-local pdfreserveobj   = pdf.reserveobj
+local pdfannotation_node = nodes.pdfannotation
 
 local submitoutputformat = 0 --  0=unknown 1=HTML 2=FDF 3=XML   => not yet used, needs to be checked
 
@@ -116,7 +120,7 @@ end
 local function checked(what)
     if what and what ~= "" then
         local set, bug = jobreferences.identify("",what)
-        return not bug and #set > 0 and lpdf.pdfaction(set)
+        return not bug and #set > 0 and lpdf.action(set)
     end
 end
 
@@ -143,30 +147,53 @@ end
 -- fonts and color
 
 local fontnames = {
-    rmtf = "Times-Roman",           rmbf = "Times-Bold",
-    rmit = "Times-Italic",          rmsl = "Times-Italic",
-    rmbi = "Times-BoldItalic",      rmbs = "Times-BoldItalic",
-    sstf = "Helvetica",             ssbf = "Helvetica-Bold",
-    ssit = "Helvetica-Oblique",     sssl = "Helvetica-Oblique",
-    ssbi = "Helvetica-BoldOblique", ssbs = "Helvetica-BoldOblique",
-    tttf = "Courier",               ttbf = "Courier-Bold",
-    ttit = "Courier-Oblique",       ttsl = "Courier-Oblique",
-    ttbi = "Courier-BoldOblique",   ttbs = "Courier-BoldOblique",
+    rm = {
+        tf = "Times-Roman",
+        bf = "Times-Bold",
+        it = "Times-Italic",
+        sl = "Times-Italic",
+        bi = "Times-BoldItalic",
+        bs = "Times-BoldItalic",
+    },
+    ss = {
+        tf = "Helvetica",
+        bf = "Helvetica-Bold",
+        it = "Helvetica-Oblique",
+        sl = "Helvetica-Oblique",
+        bi = "Helvetica-BoldOblique",
+        bs = "Helvetica-BoldOblique",
+    },
+    tt = {
+        tf = "Courier",
+        bf = "Courier-Bold",
+        it = "Courier-Oblique",
+        sl = "Courier-Oblique",
+        bi = "Courier-BoldOblique",
+        bs = "Courier-BoldOblique",
+    }
 }
 
 local usedfonts = { }
 
 local function fieldsurrounding(specification)
-    local tag = (specification.fontstyle or "tt") .. (specification.fontalternative or "tf")
-    if not fontnames[tag] then
-        tag = "tttf"
+    local size        = specification.fontsize or "12pt"
+    local style       = specification.fontstyle or "rm"
+    local alternative = specification.fontalternative or "tf"
+    local s = fontnames[style]
+    if not s then
+        style, s = "rm", fontnames.rm
     end
-    local size = specification.fontsize
+    local a = s[alternative]
+    if not a then
+        alternative, a = "tf", s.tf
+    end
+    local tag = style .. alternative
+    size = string.todimen(size)
     local stream = pdfstream {
         pdfconstant(tag),
-        format("%s Tf",(size and (number.dimenfactors.bp * size)) or 12),
+        format("%0.4f Tf",(size and (number.dimenfactors.bp * size)) or 12),
     }
-    usedfonts[tag] = true
+    usedfonts[tag] = a -- the name
     -- add color to stream: 0 g
     -- move up with "x.y Ts"
     return tostring(stream)
@@ -175,14 +202,14 @@ end
 local function registerfonts()
     if next(usedfonts) then
         local d = pdfdictionary()
-        for tag, _ in next, usedfonts do
+        for tag, name in next, usedfonts do
             local f = pdfdictionary {
                 Type     = pdfconstant("Font"),
                 Subtype  = pdfconstant("Type1"), -- todo
                 Name     = pdfconstant(tag),
-                BaseFont = pdfconstant(fontnames[tag]),
+                BaseFont = pdfconstant(name),
             }
-            d[tag] = pdfreference(pdfimmediateobj(tostring(f)))
+            d[tag] = pdfreference(pdfflushobject(f))
         end
         return d
     end
@@ -203,7 +230,7 @@ end
 local function fieldappearances(specification)
     -- todo: caching
     local values = specification.values
-    local default = specification.default
+    local default = specification.default -- todo
     if not values then
         -- error
         return
@@ -231,18 +258,18 @@ local function fieldstates(specification,forceyes,values,default)
         return
     end
     local v = aux.settings_to_array(values)
-    local yes, off
+    local yes, off, yesn, yesr, yesd, offn, offr, offd
     if #v == 1 then
         yes, off = v[1], v[1]
     else
         yes, off = v[1], v[2]
     end
-    local yesshown, yesvalue = splitter:match(yes)
+    local yesshown, yesvalue = lpegmatch(splitter,yes)
     if not (yesshown and yesvalue) then
         yesshown = yes, yes
     end
     yes = aux.settings_to_array(yesshown)
-    local offshown, offvalue = splitter:match(off)
+    local offshown, offvalue = lpegmatch(splitter,off)
     if not (offshown and offvalue) then
         offshown = off, off
     end
@@ -293,7 +320,7 @@ local function fieldoptions(specification)
         local v = aux.settings_to_array(values)
         for i=1,#v do
             local vi = v[i]
-            local shown, value = splitter:match(vi)
+            local shown, value = lpegmatch(splitter,vi)
             if shown and value then
                 v[i] = pdfarray { pdfunicode(value), shown }
             else
@@ -364,8 +391,12 @@ end
 local function predefinesymbols(specification)
     local values = specification.values
     if values then
-        local a, b = splitter:match(values)
-        codeinjections.presetsymbollist(a or values)
+        local symbols = aux.settings_to_array(values)
+        for i=1,#symbols do
+            local symbol = symbols[i]
+            local a, b = lpegmatch(splitter,symbol)
+            codeinjections.presetsymbol(a or symbol)
+        end
     end
 end
 
@@ -375,11 +406,11 @@ function codeinjections.getdefaultfieldvalue(name)
         local values  = f.values
         local default = f.default
         if not default or default == "" then
-            local a, b = splitter:match(values)
-            values = a or values
-            for name in gmatch(list,"[^, ]+") do
-                default = name
-                break
+            local symbols = aux.settings_to_array(values)
+            local symbol = symbols[1]
+            if symbol then
+                local a, b = lpegmatch(splitter,symbol) -- splits at =>
+                default = a or symbol
             end
         end
         if default then
@@ -395,7 +426,7 @@ function codeinjections.definefield(specification)
         local kind = specification.kind
         if not kind then
             if trace_fields then
-                logs.report("fields","invalid definition of '%s': unknown type",n)
+                report_fields("invalid definition of '%s': unknown type",n)
             end
         elseif kind == "radio" then
             local values = specification.values
@@ -406,10 +437,10 @@ function codeinjections.definefield(specification)
                 end
                 fields[n] = specification
                 if trace_fields then
-                    logs.report("fields","defining '%s' as radio",n or "?")
+                    report_fields("defining '%s' as radio",n or "?")
                 end
             elseif trace_fields then
-                logs.report("fields","invalid definition of radio '%s': missing values",n)
+                report_fields("invalid definition of radio '%s': missing values",n)
             end
         elseif kind == "sub" then
             -- not in main field list !
@@ -421,21 +452,29 @@ function codeinjections.definefield(specification)
                 end
                 if trace_fields then
                     local p = radios[n] and radios[n].parent
-                    logs.report("fields","defining '%s' as sub of radio '%s'",n or "?",p or "?")
+                    report_fields("defining '%s' as sub of radio '%s'",n or "?",p or "?")
                 end
             elseif trace_fields then
-                logs.report("fields","invalid definition of radio sub '%s': no parent",n)
+                report_fields("invalid definition of radio sub '%s': no parent",n)
             end
             predefinesymbols(specification)
+        elseif kind == "text" or kind == "line" then
+            fields[n] = specification
+            if trace_fields then
+                report_fields("defining '%s' as %s",n,kind)
+            end
+            if specification.values ~= "" and specification.default == "" then
+                specification.default, specification.values = specification.values, nil
+            end
         else
             fields[n] = specification
             if trace_fields then
-                logs.report("fields","defining '%s' as %s",n,kind)
+                report_fields("defining '%s' as %s",n,kind)
             end
             predefinesymbols(specification)
         end
     elseif trace_fields then
-        logs.report("fields","invalid definition of '%s': already defined",n)
+        report_fields("invalid definition of '%s': already defined",n)
     end
 end
 
@@ -443,22 +482,22 @@ function codeinjections.clonefield(specification)
     local p, c, v = specification.parent, specification.children, specification.variant
     if not p or not c then
         if trace_fields then
-            logs.report("fields","invalid clone: children: '%s', parent '%s', variant: '%s'",p or "?",c or "?", v or "?")
+            report_fields("invalid clone: children: '%s', parent '%s', variant: '%s'",p or "?",c or "?", v or "?")
         end
     else
         for n in gmatch(c,"[^, ]+") do
             local f, r, c, x = fields[n], radios[n], clones[n], fields[p]
             if f or r or c then
                 if trace_fields then
-                    logs.report("fields","already cloned: child: '%s', parent '%s', variant: '%s'",p or "?",n or "?", v or "?")
+                    report_fields("already cloned: child: '%s', parent '%s', variant: '%s'",p or "?",n or "?", v or "?")
                 end
             elseif x then
                 if trace_fields then
-                    logs.report("fields","invalid clone: child: '%s', variant: '%s', no parent",n or "?", v or "?")
+                    report_fields("invalid clone: child: '%s', variant: '%s', no parent",n or "?", v or "?")
                 end
             else
                 if trace_fields then
-                    logs.report("fields","cloning: child: '%s', parent '%s', variant: '%s'",p or "?",n or "?", v or "?")
+                    report_fields("cloning: child: '%s', parent '%s', variant: '%s'",p or "?",n or "?", v or "?")
                 end
                 clones[n] = specification
                 predefinesymbols(specification)
@@ -521,33 +560,35 @@ end
 
 local collected = pdfarray()
 
-function codeinjections.finishfields()
+local function finishfields()
     for name, field in next, fields do
         local kids = field.kids
         if kids then
-            pdfimmediateobj(field.kobj,tostring(kids))
+            pdfflushobject(field.kobj,kids)
         end
         local pobj = field.pobj
     end
     for name, field in next, radios do
         local kids = field.kids
         if kids then
-            pdfimmediateobj(field.kobj,tostring(kids))
+            pdfflushobject(field.kobj,kids)
         end
     end
     if #collected > 0 then
+        usedfonts.tttf = fontnames.tt.tf
         local acroform = pdfdictionary {
             NeedAppearances = true,
-            Fields = pdfreference(pdfimmediateobj(tostring(collected))),
+            Fields = pdfreference(pdfflushobject(collected)),
             DR     = pdfdictionary { Font = registerfonts() },
             CO     = fieldsetlist(calculationset),
             DA     = "/tttf 12 Tf 0 g",
         }
-        lpdf.addtocatalog("AcroForm",pdfreference(pdfimmediateobj(tostring(acroform))))
+        lpdf.addtocatalog("AcroForm",pdfreference(pdfflushobject(acroform)))
     end
-    lpdf.finishfields = function() end
 end
 
+lpdf.registerdocumentfinalizer(finishfields,"form fields")
+
 local pdf_widget = pdfconstant("Widget")
 local pdf_tx     = pdfconstant("Tx")
 local pdf_ch     = pdfconstant("Ch")
@@ -563,7 +604,7 @@ local methods = { }
 function codeinjections.typesetfield(name,specification)
     local field = fields[name] or radios[name] or clones[name]
     if not field then
-        logs.report("fields", "unknown child '%s'",name)
+        report_fields( "unknown child '%s'",name)
         -- unknown field
         return
     end
@@ -575,49 +616,50 @@ function codeinjections.typesetfield(name,specification)
     if method then
         method(name,specification,variant)
     else
-        logs.report("fields", "unknown method '%s' for child '%s'",field.kind,name)
+        report_fields( "unknown method '%s' for child '%s'",field.kind,name)
     end
 end
 
 -- can be optional multipass optimization (share objects)
 
 local function save_parent(field,specification,d)
-    local kn = pdfreserveobj()
+    local kn = pdfreserveobject()
     d.Kids = pdfreference(kn)
     field.kobj = kn
     field.kids = pdfarray()
-    local pn = pdfimmediateobj(tostring(d))
+    local pn = pdfflushobject(d)
     field.pobj = pn
     collected[#collected+1] = pdfreference(pn)
 end
 
 local function save_kid(field,specification,d)
-    local kn = pdfreserveobj()
+    local kn = pdfreserveobject()
     field.kids[#field.kids+1] = pdfreference(kn)
-    node.write(nodes.pdfannot(specification.width,specification.height,0,d(),kn))
+    node.write(pdfannotation_node(specification.width,specification.height,0,d(),kn))
 end
 
 function methods.line(name,specification,variant,extras)
     local field = fields[name]
     if variant == "copy" or variant == "clone" then
-        logs.report("fields","todo: clones of text fields")
+        report_fields("todo: clones of text fields")
     end
     local kind = field.kind
     if not field.pobj then
         if trace_fields then
-            logs.report("fields","using parent text '%s'",name)
+            report_fields("using parent text '%s'",name)
         end
         if extras then
             enhance(specification,extras)
         end
-        local text = pdfunicode(specification.default)
+        local text = pdfunicode(field.default)
         local d = pdfdictionary {
             Subtype  = pdf_widget,
             T        = pdfunicode(specification.title),
             F        = fieldplus(specification),
             Ff       = fieldflag(specification),
             OC       = fieldlayer(specification),
-            MK       = fieldsurrounding(specification),
+            MK       = fieldsurrounding(specification), -- needed ?
+            DA       = fieldsurrounding(specification),
             AA       = fieldactions(specification),
             FT       = pdf_tx,
             Q        = fieldalignment(specification),
@@ -630,7 +672,7 @@ function methods.line(name,specification,variant,extras)
     end
     specification = field.specification or { } -- todo: radio spec
     if trace_fields then
-        logs.report("fields","using child text '%s'",name)
+        report_fields("using child text '%s'",name)
     end
     local d = pdfdictionary {
         Subtype = pdf_widget,
@@ -639,8 +681,9 @@ function methods.line(name,specification,variant,extras)
         DA      = fieldattributes(specification),
         OC      = fieldlayer(specification),
         MK      = fieldsurrounding(specification),
+        DA      = fieldsurrounding(specification),
         AA      = fieldactions(specification),
-        Q       = fieldalignment(specification)
+        Q       = fieldalignment(specification),
     }
     save_kid(field,specification,d)
 end
@@ -652,13 +695,13 @@ end
 function methods.choice(name,specification,variant,extras)
     local field = fields[name]
     if variant == "copy" or variant == "clone" then
-        logs.report("fields","todo: clones of choice fields")
+        report_fields("todo: clones of choice fields")
     end
     local kind = field.kind
     local d
     if not field.pobj then
         if trace_fields then
-            logs.report("fields","using parent choice '%s'",name)
+            report_fields("using parent choice '%s'",name)
         end
         if extras then
             enhance(specification,extras)
@@ -678,7 +721,7 @@ function methods.choice(name,specification,variant,extras)
     end
     specification = field.specification or { }
     if trace_fields then
-        logs.report("fields","using child choice '%s'",name)
+        report_fields("using child choice '%s'",name)
     end
     local d = pdfdictionary {
         Subtype = pdf_widget,
@@ -706,13 +749,13 @@ function methods.check(name,specification,variant)
     -- contrary to radio there is no way to associate then
     local field = fields[name]
     if variant == "copy" or variant == "clone" then
-        logs.report("fields","todo: clones of check fields")
+        report_fields("todo: clones of check fields")
     end
     local kind = field.kind
     local appearance, default = fieldstates(field,true)
     if not field.pobj then
         if trace_fields then
-            logs.report("fields","using parent check '%s'",name)
+            report_fields("using parent check '%s'",name)
         end
         local d = pdfdictionary {
             Subtype  = pdf_widget,
@@ -733,7 +776,7 @@ function methods.check(name,specification,variant)
     end
     specification = field.specification or { } -- todo: radio spec
     if trace_fields then
-        logs.report("fields","using child check '%s'",name)
+        report_fields("using child check '%s'",name)
     end
     local d = pdfdictionary {
         Subtype = pdf_widget,
@@ -754,12 +797,12 @@ end
 function methods.push(name,specification,variant)
     local field = fields[name]
     if variant == "copy" or variant == "clone" then
-        logs.report("fields","todo: clones of push fields")
+        report_fields("todo: clones of push fields")
     end
     local kind = field.kind
     if not field.pobj then
         if trace_fields then
-            logs.report("fields","using parent push '%s'",name)
+            report_fields("using parent push '%s'",name)
         end
         enhance(specification,"PushButton")
         local d = pdfdictionary {
@@ -778,7 +821,7 @@ function methods.push(name,specification,variant)
     end
     specification = field.specification or { } -- todo: radio spec
     if trace_fields then
-        logs.report("fields","using child push '%s'",name)
+        report_fields("using child push '%s'",name)
     end
     local d = pdfdictionary {
         Subtype = pdf_widget,
@@ -811,7 +854,7 @@ function methods.sub(name,specification,variant)
     local default = radiodefault(parent,field)
     if not parent.pobj then
         if trace_fields then
-            logs.report("fields","using parent '%s' of radio '%s' with values '%s' and default '%s'",parent.name,name,parent.values or "?",parent.default or "?")
+            report_fields("using parent '%s' of radio '%s' with values '%s' and default '%s'",parent.name,name,parent.values or "?",parent.default or "?")
         end
         local specification = parent.specification or { }
     --  enhance(specification,"Radio,RadiosInUnison")
@@ -828,7 +871,7 @@ function methods.sub(name,specification,variant)
         save_parent(parent,specification,d)
     end
     if trace_fields then
-        logs.report("fields","using child radio '%s' with values '%s'",name,values or "?")
+        report_fields("using child radio '%s' with values '%s'",name,values or "?")
     end
     local d = pdfdictionary {
         Subtype = pdf_widget,
