@@ -6,6 +6,11 @@ if not modules then modules = { } end modules ['luat-exe'] = {
     license   = "see context related readme files"
 }
 
+local match, find = string.match, string.find
+local concat = table.concat
+
+local report_executer = logs.new("executer")
+
 if not executer then executer = { } end
 
 executer.permitted = { }
@@ -13,7 +18,9 @@ executer.execute   = os.execute
 
 function executer.register(...)
     local ep = executer.permitted
-    for k,v in pairs({...}) do
+    local t = { ... }
+    for k=1,#t do
+        local v = t[k]
         ep[#ep+1] = (v == "*" and ".*") or v
     end
 end
@@ -21,34 +28,36 @@ end
 function executer.finalize() -- todo: os.exec, todo: report ipv print
     local execute = os.execute
     function executer.execute(...)
-        local t, name, arguments = {...}, "", ""
+        -- todo: make more clever first split
+        local t, name, arguments = { ... }, "", ""
+        local one = t[1]
         if #t == 1 then
-            if type(t[1]) == 'table' then
-                name, arguments = t[1], table.concat(t," ",2,#t)
+            if type(one) == 'table' then
+                name, arguments = one, concat(t," ",2,#t)
             else
-                name, arguments = t[1]:match("^(.-)%s+(.+)$")
+                name, arguments = match(one,"^(.-)%s+(.+)$")
                 if not (name and arguments) then
-                    name, arguments = t[1], ""
+                    name, arguments = one, ""
                 end
             end
         else
-            name, arguments = t[1], table.concat(t," ",2,#t)
+            name, arguments = one, concat(t," ",2,#t)
         end
-        for _,v in pairs(executer.permitted) do
-            if name:find(v) then
+        local permitted = executer.permitted
+        for k=1,#permitted do
+            local v = permitted[k]
+            if find(name,v) then
                 execute(name .. " " .. arguments)
             --  print("executed: " .. name .. " " .. arguments)
             else
-                print("not permitted: " .. name .. " " .. arguments)
+                report_executer("not permitted: %s %s",name,arguments)
             end
         end
     end
     function executer.finalize()
-        print("executer is already finalized")
-    end
-    function executer.register(name)
-        print("executer is already finalized")
+        report_executer("already finalized")
     end
+    executer.register = executer.finalize
     os.execute = executer.execute
 end
 
@@ -62,3 +71,20 @@ end
 --~ executer.execute("dir *.tex")
 --~ executer.execute("ls *.tex")
 --~ os.execute('ls')
+
+function executer.check()
+    local mode = resolvers.variable("command_mode")
+    local list = resolvers.variable("command_list")
+    if mode == "none" then
+        executer.finalize()
+    elseif mode == "list" and list ~= "" then
+        for s in string.gmatch("[^%s,]",list) do
+            executer.register(s)
+        end
+        executer.finalize()
+    else
+        -- all
+    end
+end
+
+--~ executer.check()
