@@ -11,6 +11,10 @@ if not modules then modules = { } end modules ['lxml-pth'] = {
 local concat, remove, insert = table.concat, table.remove, table.insert
 local type, next, tonumber, tostring, setmetatable, loadstring = type, next, tonumber, tostring, setmetatable, loadstring
 local format, upper, lower, gmatch, gsub, find, rep = string.format, string.upper, string.lower, string.gmatch, string.gsub, string.find, string.rep
+local lpegmatch = lpeg.match
+
+-- beware, this is not xpath ... e.g. position is different (currently) and
+-- we have reverse-sibling as reversed preceding sibling
 
 --[[ldx--
 <p>This module can be used stand alone but also inside <l n='mkiv'/> in
@@ -39,6 +43,8 @@ local trace_lpath    = false  if trackers then trackers.register("xml.path",
 local trace_lparse   = false  if trackers then trackers.register("xml.parse",   function(v) trace_lparse = v end) end
 local trace_lprofile = false  if trackers then trackers.register("xml.profile", function(v) trace_lpath  = v trace_lparse = v trace_lprofile = v end) end
 
+local report_lpath = logs.new("lpath")
+
 --[[ldx--
 <p>We've now arrived at an interesting part: accessing the tree using a subset
 of <l n='xpath'/> and since we're not compatible we call it <l n='lpath'/>. We
@@ -65,7 +71,7 @@ local function fallback (t, name)
     if fn then
         t[name] = fn
     else
-        logs.report("xml","unknown sub finalizer '%s'",tostring(name))
+        report_lpath("unknown sub finalizer '%s'",tostring(name))
         fn = function() end
     end
     return fn
@@ -137,14 +143,19 @@ end
 apply_axis['child'] = function(list)
     local collected = { }
     for l=1,#list do
-        local dt = list[l].dt
+        local ll = list[l]
+        local dt = ll.dt
+        local en = 0
         for k=1,#dt do
             local dk = dt[k]
             if dk.tg then
                 collected[#collected+1] = dk
                 dk.ni = k -- refresh
+            en = en + 1
+            dk.ei = en
             end
         end
+        ll.en = en
     end
     return collected
 end
@@ -152,14 +163,18 @@ end
 local function collect(list,collected)
     local dt = list.dt
     if dt then
+        local en = 0
         for k=1,#dt do
             local dk = dt[k]
             if dk.tg then
                 collected[#collected+1] = dk
                 dk.ni = k -- refresh
+                en = en + 1
+                dk.ei = en
                 collect(dk,collected)
             end
         end
+        list.en = en
     end
 end
 apply_axis['descendant'] = function(list)
@@ -173,14 +188,18 @@ end
 local function collect(list,collected)
     local dt = list.dt
     if dt then
+        local en = 0
         for k=1,#dt do
             local dk = dt[k]
             if dk.tg then
                 collected[#collected+1] = dk
                 dk.ni = k -- refresh
+                en = en + 1
+                dk.ei = en
                 collect(dk,collected)
             end
         end
+        list.en = en
     end
 end
 apply_axis['descendant-or-self'] = function(list)
@@ -239,26 +258,94 @@ apply_axis['attribute'] = function(list)
     return { }
 end
 
-apply_axis['following'] = function(list)
-    return { }
-end
-
-apply_axis['following-sibling'] = function(list)
-    return { }
-end
-
 apply_axis['namespace'] = function(list)
     return { }
 end
 
-apply_axis['preceding'] = function(list)
+apply_axis['following'] = function(list) -- incomplete
+--~     local collected = { }
+--~     for l=1,#list do
+--~         local ll = list[l]
+--~         local p = ll.__p__
+--~         local d = p.dt
+--~         for i=ll.ni+1,#d do
+--~             local di = d[i]
+--~             if type(di) == "table" then
+--~                 collected[#collected+1] = di
+--~                 break
+--~             end
+--~         end
+--~     end
+--~     return collected
     return { }
 end
 
-apply_axis['preceding-sibling'] = function(list)
+apply_axis['preceding'] = function(list) -- incomplete
+--~     local collected = { }
+--~     for l=1,#list do
+--~         local ll = list[l]
+--~         local p = ll.__p__
+--~         local d = p.dt
+--~         for i=ll.ni-1,1,-1 do
+--~             local di = d[i]
+--~             if type(di) == "table" then
+--~                 collected[#collected+1] = di
+--~                 break
+--~             end
+--~         end
+--~     end
+--~     return collected
     return { }
 end
 
+apply_axis['following-sibling'] = function(list)
+    local collected = { }
+    for l=1,#list do
+        local ll = list[l]
+        local p = ll.__p__
+        local d = p.dt
+        for i=ll.ni+1,#d do
+            local di = d[i]
+            if type(di) == "table" then
+                collected[#collected+1] = di
+            end
+        end
+    end
+    return collected
+end
+
+apply_axis['preceding-sibling'] = function(list)
+    local collected = { }
+    for l=1,#list do
+        local ll = list[l]
+        local p = ll.__p__
+        local d = p.dt
+        for i=1,ll.ni-1 do
+            local di = d[i]
+            if type(di) == "table" then
+                collected[#collected+1] = di
+            end
+        end
+    end
+    return collected
+end
+
+apply_axis['reverse-sibling'] = function(list) -- reverse preceding
+    local collected = { }
+    for l=1,#list do
+        local ll = list[l]
+        local p = ll.__p__
+        local d = p.dt
+        for i=ll.ni-1,1,-1 do
+            local di = d[i]
+            if type(di) == "table" then
+                collected[#collected+1] = di
+            end
+        end
+    end
+    return collected
+end
+
 apply_axis['auto-descendant-or-self'] = apply_axis['descendant-or-self']
 apply_axis['auto-descendant']         = apply_axis['descendant']
 apply_axis['auto-child']              = apply_axis['child']
@@ -362,13 +449,19 @@ local function apply_nodes(list,directive,nodes)
     end
 end
 
+local quit_expression = false
+
 local function apply_expression(list,expression,order)
     local collected = { }
+    quit_expression = false
     for l=1,#list do
         local ll = list[l]
-        if expression(list,ll,l,order) then -- nasty, alleen valid als n=1
+        if expression(list,ll,l,order) then -- nasty, order alleen valid als n=1
             collected[#collected+1] = ll
         end
+        if quit_expression then
+            break
+        end
     end
     return collected
 end
@@ -384,15 +477,21 @@ local lp_or      = P("|")  / " or "
 local lp_and     = P("&")  / " and "
 
 local lp_builtin = P (
+        P("firstindex")   / "1" +
+        P("lastindex")    / "(#ll.__p__.dt or 1)" +
+        P("firstelement") / "1" +
+        P("lastelement")  / "(ll.__p__.en or 1)" +
         P("first")        / "1" +
         P("last")         / "#list" +
-        P("position")     / "l" + -- is element in finalizer
         P("rootposition") / "order" +
+        P("position")     / "l" + -- is element in finalizer
         P("order")        / "order" +
+        P("element")      / "(ll.ei or 1)" +
         P("index")        / "(ll.ni or 1)" +
         P("match")        / "(ll.mi or 1)" +
         P("text")         / "(ll.dt[1] or '')" +
-        P("name")         / "(ll.ns~='' and ll.ns..':'..ll.tg)" +
+    --  P("name")         / "(ll.ns~='' and ll.ns..':'..ll.tg)" +
+        P("name")         / "((ll.ns~='' and ll.ns..':'..ll.tg) or ll.tg)" +
         P("tag")          / "ll.tg" +
         P("ns")           / "ll.ns"
     ) * ((spaces * P("(") * spaces * P(")"))/"")
@@ -422,6 +521,7 @@ local nested   = lpeg.P{lparent * (noparent + lpeg.V(1))^0 * rparent}
 local value    = lpeg.P(lparent * lpeg.C((noparent + nested)^0) * rparent) -- lpeg.P{"("*C(((1-S("()"))+V(1))^0)*")"}
 
 local lp_child   = Cc("expr.child(ll,'") * R("az","AZ","--","__")^1 * Cc("')")
+local lp_number  = S("+-") * R("09")^1
 local lp_string  = Cc("'") * R("az","AZ","--","__")^1 * Cc("'")
 local lp_content = (P("'") * (1-P("'"))^0 * P("'") + P('"') * (1-P('"'))^0 * P('"'))
 
@@ -429,7 +529,7 @@ local cleaner
 
 local lp_special = (C(P("name")+P("text")+P("tag")+P("count")+P("child"))) * value / function(t,s)
     if expressions[t] then
-        s = s and s ~= "" and cleaner:match(s)
+        s = s and s ~= "" and lpegmatch(cleaner,s)
         if s and s ~= "" then
             return "expr." .. t .. "(ll," .. s ..")"
         else
@@ -459,9 +559,11 @@ local converter = Cs (
 cleaner = Cs ( (
 --~     lp_fastpos +
     lp_reserved +
+    lp_number +
     lp_string +
 1 )^1 )
 
+
 --~ expr
 
 local template_e = [[
@@ -484,12 +586,42 @@ local template_f_n = [[
 
 --
 
+local register_self                    = { kind = "axis", axis = "self"                    } -- , apply = apply_axis["self"]               }
+local register_parent                  = { kind = "axis", axis = "parent"                  } -- , apply = apply_axis["parent"]             }
+local register_descendant              = { kind = "axis", axis = "descendant"              } -- , apply = apply_axis["descendant"]         }
+local register_child                   = { kind = "axis", axis = "child"                   } -- , apply = apply_axis["child"]              }
+local register_descendant_or_self      = { kind = "axis", axis = "descendant-or-self"      } -- , apply = apply_axis["descendant-or-self"] }
+local register_root                    = { kind = "axis", axis = "root"                    } -- , apply = apply_axis["root"]               }
+local register_ancestor                = { kind = "axis", axis = "ancestor"                } -- , apply = apply_axis["ancestor"]           }
+local register_ancestor_or_self        = { kind = "axis", axis = "ancestor-or-self"        } -- , apply = apply_axis["ancestor-or-self"]   }
+local register_attribute               = { kind = "axis", axis = "attribute"               } -- , apply = apply_axis["attribute"]          }
+local register_namespace               = { kind = "axis", axis = "namespace"               } -- , apply = apply_axis["namespace"]          }
+local register_following               = { kind = "axis", axis = "following"               } -- , apply = apply_axis["following"]          }
+local register_following_sibling       = { kind = "axis", axis = "following-sibling"       } -- , apply = apply_axis["following-sibling"]  }
+local register_preceding               = { kind = "axis", axis = "preceding"               } -- , apply = apply_axis["preceding"]          }
+local register_preceding_sibling       = { kind = "axis", axis = "preceding-sibling"       } -- , apply = apply_axis["preceding-sibling"]  }
+local register_reverse_sibling         = { kind = "axis", axis = "reverse-sibling"         } -- , apply = apply_axis["reverse-sibling"]    }
+
+local register_auto_descendant_or_self = { kind = "axis", axis = "auto-descendant-or-self" } -- , apply = apply_axis["auto-descendant-or-self"] }
+local register_auto_descendant         = { kind = "axis", axis = "auto-descendant"         } -- , apply = apply_axis["auto-descendant"] }
+local register_auto_self               = { kind = "axis", axis = "auto-self"               } -- , apply = apply_axis["auto-self"] }
+local register_auto_child              = { kind = "axis", axis = "auto-child"              } -- , apply = apply_axis["auto-child"] }
+
+local register_initial_child           = { kind = "axis", axis = "initial-child"           } -- , apply = apply_axis["initial-child"] }
+
+local register_all_nodes               = { kind = "nodes", nodetest = true, nodes = { true, false, false } }
+
+local skip = { }
+
 local function errorrunner_e(str,cnv)
-    logs.report("lpath","error in expression: %s => %s",str,cnv)
+    if not skip[str] then
+        report_lpath("error in expression: %s => %s",str,cnv)
+        skip[str] = cnv or str
+    end
     return false
 end
 local function errorrunner_f(str,arg)
-    logs.report("lpath","error in finalizer: %s(%s)",str,arg or "")
+    report_lpath("error in finalizer: %s(%s)",str,arg or "")
     return false
 end
 
@@ -498,7 +630,7 @@ local function register_nodes(nodetest,nodes)
 end
 
 local function register_expression(expression)
-    local converted = converter:match(expression)
+    local converted = lpegmatch(converter,expression)
     local runner = loadstring(format(template_e,converted))
     runner = (runner and runner()) or function() errorrunner_e(expression,converted) end
     return { kind = "expression", expression = expression, converted = converted, evaluator = runner }
@@ -530,38 +662,22 @@ local arguments = P { "ar",
 
 -- todo: better arg parser
 
-local register_self                    = { kind = "axis", axis = "self"                    } -- , apply = apply_axis["self"]               }
-local register_parent                  = { kind = "axis", axis = "parent"                  } -- , apply = apply_axis["parent"]             }
-local register_descendant              = { kind = "axis", axis = "descendant"              } -- , apply = apply_axis["descendant"]         }
-local register_child                   = { kind = "axis", axis = "child"                   } -- , apply = apply_axis["child"]              }
-local register_descendant_or_self      = { kind = "axis", axis = "descendant-or-self"      } -- , apply = apply_axis["descendant-or-self"] }
-local register_root                    = { kind = "axis", axis = "root"                    } -- , apply = apply_axis["root"]               }
-local register_ancestor                = { kind = "axis", axis = "ancestor"                } -- , apply = apply_axis["ancestor"]           }
-local register_ancestor_or_self        = { kind = "axis", axis = "ancestor-or-self"        } -- , apply = apply_axis["ancestor-or-self"]   }
-local register_attribute               = { kind = "axis", axis = "attribute"               } -- , apply = apply_axis["attribute"]          }
-local register_namespace               = { kind = "axis", axis = "namespace"               } -- , apply = apply_axis["namespace"]          }
-local register_following               = { kind = "axis", axis = "following"               } -- , apply = apply_axis["following"]          }
-local register_following_sibling       = { kind = "axis", axis = "following-sibling"       } -- , apply = apply_axis["following-sibling"]  }
-local register_preceding               = { kind = "axis", axis = "preceding"               } -- , apply = apply_axis["preceding"]          }
-local register_preceding_sibling       = { kind = "axis", axis = "preceding-sibling"       } -- , apply = apply_axis["preceding-sibling"]  }
-
-local register_auto_descendant_or_self = { kind = "axis", axis = "auto-descendant-or-self" } -- , apply = apply_axis["auto-descendant-or-self"] }
-local register_auto_descendant         = { kind = "axis", axis = "auto-descendant"         } -- , apply = apply_axis["auto-descendant"] }
-local register_auto_self               = { kind = "axis", axis = "auto-self"               } -- , apply = apply_axis["auto-self"] }
-local register_auto_child              = { kind = "axis", axis = "auto-child"              } -- , apply = apply_axis["auto-child"] }
-
-local register_initial_child           = { kind = "axis", axis = "initial-child"           } -- , apply = apply_axis["initial-child"] }
-
-local register_all_nodes               = { kind = "nodes", nodetest = true, nodes = { true, false, false } }
-
 local function register_error(str)
-    return { kind = "error", comment = format("unparsed: %s",str) }
+    return { kind = "error", error = format("unparsed: %s",str) }
 end
 
+-- there is a difference in * and /*/ and so we need to catch a few special cases
+
+local special_1 = P("*")  * Cc(register_auto_descendant) * Cc(register_all_nodes) -- last one not needed
+local special_2 = P("/")  * Cc(register_auto_self)
+local special_3 = P("")   * Cc(register_auto_self)
+
 local parser = Ct { "patterns", -- can be made a bit faster by moving pattern outside
 
-    patterns             = spaces * V("protocol") * spaces * V("initial") * spaces * V("step") * spaces *
-                           (P("/") * spaces * V("step") * spaces)^0,
+    patterns             = spaces * V("protocol") * spaces * (
+                              ( V("special") * spaces * P(-1)                                                         ) +
+                              ( V("initial") * spaces * V("step") * spaces * (P("/") * spaces * V("step") * spaces)^0 )
+                           ),
 
     protocol             = Cg(V("letters"),"protocol") * P("://") + Cg(Cc(nil),"protocol"),
 
@@ -570,10 +686,12 @@ local parser = Ct { "patterns", -- can be made a bit faster by moving pattern ou
     step                 = ((V("shortcuts") + P("/") + V("axis")) * spaces * V("nodes")^0 + V("error")) * spaces * V("expressions")^0 * spaces * V("finalizer")^0,
 
     axis                 = V("descendant") + V("child") + V("parent") + V("self") + V("root") + V("ancestor") +
-                           V("descendant_or_self") + V("following") + V("following_sibling") +
-                           V("preceding") + V("preceding_sibling") + V("ancestor_or_self") +
+                           V("descendant_or_self") + V("following_sibling") + V("following") +
+                           V("reverse_sibling") + V("preceding_sibling") + V("preceding") + V("ancestor_or_self") +
                            #(1-P(-1)) * Cc(register_auto_child),
 
+    special              = special_1 + special_2 + special_3,
+
     initial              = (P("/") * spaces * Cc(register_initial_child))^-1,
 
     error                = (P(1)^1) / register_error,
@@ -585,7 +703,8 @@ local parser = Ct { "patterns", -- can be made a bit faster by moving pattern ou
     s_descendant_or_self = (P("***/") + P("/"))  * Cc(register_descendant_or_self), --- *** is a bonus
  -- s_descendant_or_self = P("/")                * Cc(register_descendant_or_self),
     s_descendant         = P("**")               * Cc(register_descendant),
-    s_child              = P("*")  * #(1-P(":")) * Cc(register_child     ),
+    s_child              = P("*") * #(1-P(":"))  * Cc(register_child     ),
+--  s_child              = P("*") * #(P("/")+P(-1)) * Cc(register_child     ),
     s_parent             = P("..")               * Cc(register_parent    ),
     s_self               = P("." )               * Cc(register_self      ),
     s_root               = P("^^")               * Cc(register_root      ),
@@ -605,6 +724,7 @@ local parser = Ct { "patterns", -- can be made a bit faster by moving pattern ou
     following_sibling    = P('following-sibling::')  * Cc(register_following_sibling  ),
     preceding            = P('preceding::')          * Cc(register_preceding          ),
     preceding_sibling    = P('preceding-sibling::')  * Cc(register_preceding_sibling  ),
+    reverse_sibling      = P('reverse-sibling::')    * Cc(register_reverse_sibling    ),
 
     nodes                = (V("nodefunction") * spaces * P("(") * V("nodeset") * P(")") + V("nodetest") * V("nodeset")) / register_nodes,
 
@@ -660,19 +780,30 @@ end
 
 xml.nodesettostring = nodesettostring
 
+local parse_pattern -- we have a harmless kind of circular reference
+
 local function lshow(parsed)
     if type(parsed) == "string" then
         parsed = parse_pattern(parsed)
     end
     local s = table.serialize_functions -- ugly
     table.serialize_functions = false -- ugly
-    logs.report("lpath","%s://%s => %s",parsed.protocol or xml.defaultprotocol,parsed.pattern,table.serialize(parsed,false))
+    report_lpath("%s://%s => %s",parsed.protocol or xml.defaultprotocol,parsed.pattern,table.serialize(parsed,false))
     table.serialize_functions = s -- ugly
 end
 
 xml.lshow = lshow
 
-local function parse_pattern(pattern) -- the gain of caching is rather minimal
+local function add_comment(p,str)
+    local pc = p.comment
+    if not pc then
+        p.comment = { str }
+    else
+        pc[#pc+1] = str
+    end
+end
+
+parse_pattern = function (pattern) -- the gain of caching is rather minimal
     lpathcalls = lpathcalls + 1
     if type(pattern) == "table" then
         return pattern
@@ -681,27 +812,41 @@ local function parse_pattern(pattern) -- the gain of caching is rather minimal
         if parsed then
             lpathcached = lpathcached + 1
         else
-            parsed = parser:match(pattern)
+            parsed = lpegmatch(parser,pattern)
             if parsed then
                 parsed.pattern = pattern
                 local np = #parsed
                 if np == 0 then
                     parsed = { pattern = pattern, register_self, state = "parsing error" }
-                    logs.report("lpath","parsing error in '%s'",pattern)
+                    report_lpath("parsing error in '%s'",pattern)
                     lshow(parsed)
                 else
-                    -- we could have done this with a more complex parsed but this
+                    -- we could have done this with a more complex parser but this
                     -- is cleaner
                     local pi = parsed[1]
                     if pi.axis == "auto-child" then
-                        parsed.comment = "auto-child replaced by auto-descendant-or-self"
-                        parsed[1] = register_auto_descendant_or_self
-                    --~ parsed.comment = "auto-child replaced by auto-descendant"
-                    --~ parsed[1] = register_auto_descendant
+                        if false then
+                            add_comment(parsed, "auto-child replaced by auto-descendant-or-self")
+                            parsed[1] = register_auto_descendant_or_self
+                        else
+                            add_comment(parsed, "auto-child replaced by auto-descendant")
+                            parsed[1] = register_auto_descendant
+                        end
                     elseif pi.axis == "initial-child" and np > 1 and parsed[2].axis then
-                        parsed.comment = "initial-child removed" -- we could also make it a auto-self
+                        add_comment(parsed, "initial-child removed") -- we could also make it a auto-self
                         remove(parsed,1)
                     end
+                    local np = #parsed -- can have changed
+                    if np > 1 then
+                        local pnp = parsed[np]
+                        if pnp.kind == "nodes" and pnp.nodetest == true then
+                            local nodes = pnp.nodes
+                            if nodes[1] == true and nodes[2] == false and nodes[3] == false then
+                                add_comment(parsed, "redundant final wildcard filter removed")
+                                remove(parsed,np)
+                            end
+                        end
+                    end
                 end
             else
                 parsed = { pattern = pattern }
@@ -727,6 +872,10 @@ end
 -- caching found lookups saves not that much (max .1 sec on a 8 sec run)
 -- and it also messes up finalizers
 
+-- watch out: when there is a finalizer, it's always called as there
+-- can be cases that a finalizer returns (or does) something in case
+-- there is no match; an example of this is count()
+
 local profiled = { }  xml.profiled = profiled
 
 local function profiled_apply(list,parsed,nofparsed,order)
@@ -754,6 +903,12 @@ local function profiled_apply(list,parsed,nofparsed,order)
             return collected
         end
         if not collected or #collected == 0 then
+            local pn = i < nofparsed and parsed[nofparsed]
+            if pn and pn.kind == "finalizer" then
+                collected = pn.finalizer(collected)
+                p.finalized = p.finalized + 1
+                return collected
+            end
             return nil
         end
     end
@@ -767,28 +922,34 @@ local function traced_apply(list,parsed,nofparsed,order)
     if trace_lparse then
         lshow(parsed)
     end
-    logs.report("lpath", "collecting : %s",parsed.pattern)
-    logs.report("lpath", " root tags : %s",tagstostring(list))
-    logs.report("lpath", "     order : %s",order or "unset")
+    report_lpath("collecting : %s",parsed.pattern)
+    report_lpath(" root tags : %s",tagstostring(list))
+    report_lpath("     order : %s",order or "unset")
     local collected = list
     for i=1,nofparsed do
         local pi = parsed[i]
         local kind = pi.kind
         if kind == "axis" then
             collected = apply_axis[pi.axis](collected)
-            logs.report("lpath", "% 10i : ax : %s",(collected and #collected) or 0,pi.axis)
+            report_lpath("% 10i : ax : %s",(collected and #collected) or 0,pi.axis)
         elseif kind == "nodes" then
             collected = apply_nodes(collected,pi.nodetest,pi.nodes)
-            logs.report("lpath", "% 10i : ns : %s",(collected and #collected) or 0,nodesettostring(pi.nodes,pi.nodetest))
+            report_lpath("% 10i : ns : %s",(collected and #collected) or 0,nodesettostring(pi.nodes,pi.nodetest))
         elseif kind == "expression" then
             collected = apply_expression(collected,pi.evaluator,order)
-            logs.report("lpath", "% 10i : ex : %s -> %s",(collected and #collected) or 0,pi.expression,pi.converted)
+            report_lpath("% 10i : ex : %s -> %s",(collected and #collected) or 0,pi.expression,pi.converted)
         elseif kind == "finalizer" then
             collected = pi.finalizer(collected)
-            logs.report("lpath", "% 10i : fi : %s : %s(%s)",(collected and #collected) or 0,parsed.protocol or xml.defaultprotocol,pi.name,pi.arguments or "")
+            report_lpath("% 10i : fi : %s : %s(%s)",(type(collected) == "table" and #collected) or 0,parsed.protocol or xml.defaultprotocol,pi.name,pi.arguments or "")
             return collected
         end
         if not collected or #collected == 0 then
+            local pn = i < nofparsed and parsed[nofparsed]
+            if pn and pn.kind == "finalizer" then
+                collected = pn.finalizer(collected)
+                report_lpath("% 10i : fi : %s : %s(%s)",(type(collected) == "table" and #collected) or 0,parsed.protocol or xml.defaultprotocol,pn.name,pn.arguments or "")
+                return collected
+            end
             return nil
         end
     end
@@ -813,6 +974,10 @@ local function normal_apply(list,parsed,nofparsed,order)
             return pi.finalizer(collected)
         end
         if not collected or #collected == 0 then
+            local pf = i < nofparsed and parsed[nofparsed].finalizer
+            if pf then
+                return pf(collected) -- can be anything
+            end
             return nil
         end
     end
@@ -836,13 +1001,17 @@ local function parse_apply(list,pattern)
     end
     local nofparsed = #parsed
     if nofparsed == 0 then
-        -- something is wrong
+        return -- something is wrong
+    end
+    local one = list[1]
+    if not one then
+        return -- something is wrong
     elseif not trace_lpath then
-        return normal_apply(list,parsed,nofparsed,list[1].mi)
+        return normal_apply(list,parsed,nofparsed,one.mi)
     elseif trace_lprofile then
-        return profiled_apply(list,parsed,nofparsed,list[1].mi)
-    else -- trace_lpath
-        return traced_apply(list,parsed,nofparsed,list[1].mi)
+        return profiled_apply(list,parsed,nofparsed,one.mi)
+    else
+        return traced_apply(list,parsed,nofparsed,one.mi)
     end
 end
 
@@ -869,6 +1038,18 @@ expressions.undefined = function(s)
     return s == nil
 end
 
+expressions.quit = function(s)
+    if s or s == nil then
+        quit_expression = true
+    end
+    return true
+end
+
+expressions.print = function(...)
+    print(...)
+    return true
+end
+
 expressions.contains  = find
 expressions.find      = find
 expressions.upper     = upper
@@ -879,7 +1060,7 @@ expressions.boolean   = toboolean
 -- user interface
 
 local function traverse(root,pattern,handle)
-    logs.report("xml","use 'xml.selection' instead for '%s'",pattern)
+    report_lpath("use 'xml.selection' instead for '%s'",pattern)
     local collected = parse_apply({ root },pattern)
     if collected then
         for c=1,#collected do
@@ -927,7 +1108,7 @@ local function dofunction(collected,fnc)
                 f(collected[c])
             end
         else
-            logs.report("xml","unknown function '%s'",fnc)
+            report_lpath("unknown function '%s'",fnc)
         end
     end
 end
