@@ -12,6 +12,8 @@ if not modules then modules = { } end modules ['lxml-tab'] = {
 
 local trace_entities = false  trackers.register("xml.entities", function(v) trace_entities = v end)
 
+local report_xml = logs.new("xml")
+
 --[[ldx--
 <p>The parser used here is inspired by the variant discussed in the lua book, but
 handles comment and processing instructions, has a different structure, provides
@@ -30,8 +32,10 @@ xml = xml or { }
 
 local concat, remove, insert = table.concat, table.remove, table.insert
 local type, next, setmetatable, getmetatable, tonumber = type, next, setmetatable, getmetatable, tonumber
-local format, lower, find = string.format, string.lower, string.find
+local format, lower, find, match, gsub = string.format, string.lower, string.find, string.match, string.gsub
 local utfchar = unicode.utf8.char
+local lpegmatch = lpeg.match
+local P, S, R, C, V, C, Cs = lpeg.P, lpeg.S, lpeg.R, lpeg.C, lpeg.V, lpeg.C, lpeg.Cs
 
 --[[ldx--
 <p>First a hack to enable namespace resolving. A namespace is characterized by
@@ -43,7 +47,7 @@ much cleaner.</p>
 
 xml.xmlns = xml.xmlns or { }
 
-local check = lpeg.P(false)
+local check = P(false)
 local parse = check
 
 --[[ldx--
@@ -56,8 +60,8 @@ xml.registerns("mml","mathml")
 --ldx]]--
 
 function xml.registerns(namespace, pattern) -- pattern can be an lpeg
-    check = check + lpeg.C(lpeg.P(lower(pattern))) / namespace
-    parse = lpeg.P { lpeg.P(check) + 1 * lpeg.V(1) }
+    check = check + C(P(lower(pattern))) / namespace
+    parse = P { P(check) + 1 * V(1) }
 end
 
 --[[ldx--
@@ -71,7 +75,7 @@ xml.checkns("m","http://www.w3.org/mathml")
 --ldx]]--
 
 function xml.checkns(namespace,url)
-    local ns = parse:match(lower(url))
+    local ns = lpegmatch(parse,lower(url))
     if ns and namespace ~= ns then
         xml.xmlns[namespace] = ns
     end
@@ -89,7 +93,7 @@ This returns <t>mml</t>.
 --ldx]]--
 
 function xml.resolvens(url)
-     return parse:match(lower(url)) or ""
+     return lpegmatch(parse,lower(url)) or ""
 end
 
 --[[ldx--
@@ -143,12 +147,12 @@ element.</p>
 local nsremap, resolvens = xml.xmlns, xml.resolvens
 
 local stack, top, dt, at, xmlns, errorstr, entities = { }, { }, { }, { }, { }, nil, { }
-local strip, cleanup, utfize, resolve = false, false, false, false
+local strip, cleanup, utfize, resolve, resolve_predefined, unify_predefined = false, false, false, false, false, false
 local dcache, hcache, acache = { }, { }, { }
 
 local mt = { }
 
-function initialize_mt(root)
+local function initialize_mt(root)
     mt = { __index = root } -- will be redefined later
 end
 
@@ -167,11 +171,14 @@ local function add_attribute(namespace,tag,value)
     if tag == "xmlns" then
         xmlns[#xmlns+1] = resolvens(value)
         at[tag] = value
+    elseif namespace == "" then
+        at[tag] = value
     elseif namespace == "xmlns" then
         xml.checkns(tag,value)
         at["xmlns:" .. tag] = value
     else
-        at[tag] = value
+        -- for the moment this way:
+        at[namespace .. ":" .. tag] = value
     end
 end
 
@@ -237,7 +244,7 @@ local function add_special(what, spacing, text)
     if strip and (what == "@cm@" or what == "@dt@") then
         -- forget it
     else
-        dt[#dt+1] = { special=true, ns="", tg=what, dt={text} }
+        dt[#dt+1] = { special=true, ns="", tg=what, dt={ text } }
     end
 end
 
@@ -249,7 +256,7 @@ local reported_attribute_errors = { }
 
 local function attribute_value_error(str)
     if not reported_attribute_errors[str] then
-        logs.report("xml","invalid attribute value: %q",str)
+        report_xml("invalid attribute value: %q",str)
         reported_attribute_errors[str] = true
         at._error_ = str
     end
@@ -257,62 +264,16 @@ local function attribute_value_error(str)
 end
 local function attribute_specification_error(str)
     if not reported_attribute_errors[str] then
-        logs.report("xml","invalid attribute specification: %q",str)
+        report_xml("invalid attribute specification: %q",str)
         reported_attribute_errors[str] = true
         at._error_ = str
     end
     return str
 end
 
-function xml.unknown_dec_entity_format(str) return format("&%s;",  str) end
+function xml.unknown_dec_entity_format(str) return (str == "" and "&error;") or format("&%s;",str) end
 function xml.unknown_hex_entity_format(str) return format("&#x%s;",str) end
-function xml.unknown_any_entity_format(str) return format("&%s;",  str) end
-
-local function handle_hex_entity(str)
-    local h = hcache[str]
-    if not h then
-        if utfize then
-            local n = tonumber(str,16)
-            h = (n and utfchar(n)) or xml.unknown_hex_entity_format(str) or ""
-            if not n then
-                logs.report("xml","utfize, ignoring hex entity &#x%s;",str)
-            elseif trace_entities then
-                logs.report("xml","utfize, converting hex entity &#x%s; into %s",str,c)
-            end
-        else
-            if trace_entities then
-                logs.report("xml","found entity &#x%s;",str)
-            end
-            h = "&#c" .. str .. ";"
-        end
-        hcache[str] = h
-    end
-    return h
-end
-
-local function handle_dec_entity(str)
-    local d = dcache[str]
-    if not d then
-        if utfize then
-            local n = tonumber(str)
-            d = (n and utfchar(n)) or xml.unknown_dec_entity_format(str) or ""
-            if not n then
-                logs.report("xml","utfize, ignoring dec entity &#%s;",str)
-            elseif trace_entities then
-                logs.report("xml","utfize, converting dec entity &#%s; into %s",str,c)
-            end
-        else
-            if trace_entities then
-                logs.report("xml","found entity &#%s;",str)
-            end
-            d = "&#" .. str .. ";"
-        end
-        dcache[str] = d
-    end
-    return d
-end
-
--- one level expansion (simple case)
+function xml.unknown_any_entity_format(str) return format("&#x%s;",str) end
 
 local function fromhex(s)
     local n = tonumber(s,16)
@@ -332,7 +293,7 @@ local function fromdec(s)
     end
 end
 
-local P, S, R, C, V, Cs = lpeg.P, lpeg.S, lpeg.R, lpeg.C, lpeg.V, lpeg.Cs
+-- one level expansion (simple case), no checking done
 
 local rest = (1-P(";"))^0
 local many = P(1)^0
@@ -341,41 +302,120 @@ local parsedentity =
     P("&") * (P("#x")*(rest/fromhex) + P("#")*(rest/fromdec)) * P(";") * P(-1) +
              (P("#x")*(many/fromhex) + P("#")*(many/fromdec))
 
+-- parsing in the xml file
+
+local predefined_unified = {
+    [38] = "&amp;",
+    [42] = "&quot;",
+    [47] = "&apos;",
+    [74] = "&lt;",
+    [76] = "&gr;",
+}
+
+local predefined_simplified = {
+    [38] = "&", amp  = "&",
+    [42] = '"', quot = '"',
+    [47] = "'", apos = "'",
+    [74] = "<", lt   = "<",
+    [76] = ">", gt   = ">",
+}
+
+local function handle_hex_entity(str)
+    local h = hcache[str]
+    if not h then
+        local n = tonumber(str,16)
+        h = unify_predefined and predefined_unified[n]
+        if h then
+            if trace_entities then
+                report_xml("utfize, converting hex entity &#x%s; into %s",str,h)
+            end
+        elseif utfize then
+            h = (n and utfchar(n)) or xml.unknown_hex_entity_format(str) or ""
+            if not n then
+                report_xml("utfize, ignoring hex entity &#x%s;",str)
+            elseif trace_entities then
+                report_xml("utfize, converting hex entity &#x%s; into %s",str,h)
+            end
+        else
+            if trace_entities then
+                report_xml("found entity &#x%s;",str)
+            end
+            h = "&#x" .. str .. ";"
+        end
+        hcache[str] = h
+    end
+    return h
+end
+
+local function handle_dec_entity(str)
+    local d = dcache[str]
+    if not d then
+        local n = tonumber(str)
+        d = unify_predefined and predefined_unified[n]
+        if d then
+            if trace_entities then
+                report_xml("utfize, converting dec entity &#%s; into %s",str,d)
+            end
+        elseif utfize then
+            d = (n and utfchar(n)) or xml.unknown_dec_entity_format(str) or ""
+            if not n then
+                report_xml("utfize, ignoring dec entity &#%s;",str)
+            elseif trace_entities then
+                report_xml("utfize, converting dec entity &#%s; into %s",str,h)
+            end
+        else
+            if trace_entities then
+                report_xml("found entity &#%s;",str)
+            end
+            d = "&#" .. str .. ";"
+        end
+        dcache[str] = d
+    end
+    return d
+end
+
 xml.parsedentitylpeg = parsedentity
 
 local function handle_any_entity(str)
     if resolve then
         local a = acache[str] -- per instance ! todo
         if not a then
-            if type(resolve) == "function" then
+            a = resolve_predefined and predefined_simplified[str]
+            if a then
+                -- one of the predefined
+            elseif type(resolve) == "function" then
                 a = resolve(str) or entities[str]
             else
                 a = entities[str]
             end
             if a then
                 if trace_entities then
-                    logs.report("xml","resolved entity &%s; -> %s (internal)",str,a)
+                    report_xml("resolved entity &%s; -> %s (internal)",str,a)
                 end
-                a = parsedentity:match(a) or a
+                a = lpegmatch(parsedentity,a) or a
             else
                 if xml.unknown_any_entity_format then
                     a = xml.unknown_any_entity_format(str) or ""
                 end
                 if a then
                     if trace_entities then
-                        logs.report("xml","resolved entity &%s; -> %s (external)",str,a)
+                        report_xml("resolved entity &%s; -> %s (external)",str,a)
                     end
                 else
                     if trace_entities then
-                        logs.report("xml","keeping entity &%s;",str)
+                        report_xml("keeping entity &%s;",str)
+                    end
+                    if str == "" then
+                        a = "&error;"
+                    else
+                        a = "&" .. str .. ";"
                     end
-                    a = "&" .. str .. ";"
                 end
             end
             acache[str] = a
         elseif trace_entities then
             if not acache[str] then
-                logs.report("xml","converting entity &%s; into %s",str,a)
+                report_xml("converting entity &%s; into %s",str,a)
                 acache[str] = a
             end
         end
@@ -384,15 +424,28 @@ local function handle_any_entity(str)
         local a = acache[str]
         if not a then
             if trace_entities then
-                logs.report("xml","found entity &%s;",str)
+                report_xml("found entity &%s;",str)
+            end
+            a = resolve_predefined and predefined_simplified[str]
+            if a then
+                -- one of the predefined
+                acache[str] = a
+            elseif str == "" then
+                a = "&error;"
+                acache[str] = a
+            else
+                a = "&" .. str .. ";"
+                acache[str] = a
             end
-            a = "&" .. str .. ";"
-            acache[str] = a
         end
         return a
     end
 end
 
+local function handle_end_entity(chr)
+    report_xml("error in entity, %q found instead of ';'",chr)
+end
+
 local space            = S(' \r\n\t')
 local open             = P('<')
 local close            = P('>')
@@ -407,18 +460,18 @@ local valid            = R('az', 'AZ', '09') + S('_-.')
 local name_yes         = C(valid^1) * colon * C(valid^1)
 local name_nop         = C(P(true)) * C(valid^1)
 local name             = name_yes + name_nop
-
-local utfbom           = P('\000\000\254\255') + P('\255\254\000\000') +
-                         P('\255\254') + P('\254\255') + P('\239\187\191') -- no capture
-
+local utfbom           = lpeg.patterns.utfbom -- no capture
 local spacing          = C(space^0)
 
-local entitycontent    = (1-open-semicolon)^0
+----- entitycontent    = (1-open-semicolon)^0
+local anyentitycontent = (1-open-semicolon-space-close)^0
+local hexentitycontent = R("AF","af","09")^0
+local decentitycontent = R("09")^0
 local parsedentity     = P("#")/"" * (
-                                P("x")/"" * (entitycontent/handle_hex_entity) +
-                                            (entitycontent/handle_dec_entity)
-                            ) +             (entitycontent/handle_any_entity)
-local entity           = ampersand/"" * parsedentity * (semicolon/"")
+                                P("x")/"" * (hexentitycontent/handle_hex_entity) +
+                                            (decentitycontent/handle_dec_entity)
+                            ) +             (anyentitycontent/handle_any_entity)
+local entity           = ampersand/"" * parsedentity * ( (semicolon/"") + #(P(1)/handle_end_entity))
 
 local text_unparsed    = C((1-open)^1)
 local text_parsed      = Cs(((1-open-ampersand)^1 + entity)^1)
@@ -426,13 +479,15 @@ local text_parsed      = Cs(((1-open-ampersand)^1 + entity)^1)
 local somespace        = space^1
 local optionalspace    = space^0
 
-local value            = (squote * C((1 - squote)^0) * squote) + (dquote * C((1 - dquote)^0) * dquote) -- ampersand and < also invalid in value
+----- value            = (squote * C((1 - squote)^0) * squote) + (dquote * C((1 - dquote)^0) * dquote) -- ampersand and < also invalid in value
+local value            = (squote * Cs((entity + (1 - squote))^0) * squote) + (dquote * Cs((entity + (1 - dquote))^0) * dquote) -- ampersand and < also invalid in value
 
 local endofattributes  = slash * close + close -- recovery of flacky html
 local whatever         = space * name * optionalspace * equal
 local wrongvalue       = C(P(1-whatever-close)^1 + P(1-close)^1) / attribute_value_error
------ local wrongvalue = C(P(1-whatever-endofattributes)^1 + P(1-endofattributes)^1) / attribute_value_error
-local wrongvalue       = C(P(1-space-endofattributes)^1) / attribute_value_error
+----- wrongvalue       = C(P(1-whatever-endofattributes)^1 + P(1-endofattributes)^1) / attribute_value_error
+----- wrongvalue       = C(P(1-space-endofattributes)^1) / attribute_value_error
+local wrongvalue       = Cs(P(entity + (1-space-endofattributes))^1) / attribute_value_error
 
 local attributevalue   = value + wrongvalue
 
@@ -490,10 +545,10 @@ local doctype          = (spacing * begindoctype     * somedoctype     * enddoct
 
 --  nicer but slower:
 --
---  local instruction = (lpeg.Cc("@pi@") * spacing * begininstruction * someinstruction * endinstruction) / add_special
---  local comment     = (lpeg.Cc("@cm@") * spacing * begincomment     * somecomment     * endcomment    ) / add_special
---  local cdata       = (lpeg.Cc("@cd@") * spacing * begincdata       * somecdata       * endcdata      ) / add_special
---  local doctype     = (lpeg.Cc("@dt@") * spacing * begindoctype     * somedoctype     * enddoctype    ) / add_special
+--  local instruction = (Cc("@pi@") * spacing * begininstruction * someinstruction * endinstruction) / add_special
+--  local comment     = (Cc("@cm@") * spacing * begincomment     * somecomment     * endcomment    ) / add_special
+--  local cdata       = (Cc("@cd@") * spacing * begincdata       * somecdata       * endcdata      ) / add_special
+--  local doctype     = (Cc("@dt@") * spacing * begindoctype     * somedoctype     * enddoctype    ) / add_special
 
 local trailer = space^0 * (text_unparsed/set_message)^0
 
@@ -513,13 +568,17 @@ local grammar_unparsed_text = P { "preamble",
     children = unparsedtext + V("parent") + emptyelement + comment + cdata + instruction,
 }
 
+-- maybe we will add settinsg to result as well
+
 local function xmlconvert(data, settings)
     settings = settings or { } -- no_root strip_cm_and_dt given_entities parent_root error_handler
     strip = settings.strip_cm_and_dt
     utfize = settings.utfize_entities
     resolve = settings.resolve_entities
+    resolve_predefined = settings.resolve_predefined_entities -- in case we have escaped entities
+    unify_predefined = settings.unify_predefined_entities -- &#038; -> &amp;
     cleanup = settings.text_cleanup
-    stack, top, at, xmlns, errorstr, result, entities = { }, { }, { }, { }, nil, nil, settings.entities or { }
+    stack, top, at, xmlns, errorstr, entities = { }, { }, { }, { }, nil, settings.entities or { }
     acache, hcache, dcache = { }, { }, { } -- not stored
     reported_attribute_errors = { }
     if settings.parent_root then
@@ -533,20 +592,23 @@ local function xmlconvert(data, settings)
     if not data or data == "" then
         errorstr = "empty xml file"
     elseif utfize or resolve then
-        if grammar_parsed_text:match(data) then
+        if lpegmatch(grammar_parsed_text,data) then
             errorstr = ""
         else
             errorstr = "invalid xml file - parsed text"
         end
-    else
-        if grammar_unparsed_text:match(data) then
+    elseif type(data) == "string" then
+        if lpegmatch(grammar_unparsed_text,data) then
             errorstr = ""
         else
             errorstr = "invalid xml file - unparsed text"
         end
+    else
+        errorstr = "invalid xml file - no text at all"
     end
+    local result
     if errorstr and errorstr ~= "" then
-        result = { dt = { { ns = "", tg = "error", dt = { errorstr }, at={}, er = true } } }
+        result = { dt = { { ns = "", tg = "error", dt = { errorstr }, at={ }, er = true } } }
         setmetatable(stack, mt)
         local error_handler = settings.error_handler
         if error_handler == false then
@@ -561,13 +623,14 @@ local function xmlconvert(data, settings)
         result = stack[1]
     end
     if not settings.no_root then
-        result = { special = true, ns = "", tg = '@rt@', dt = result.dt, at={}, entities = entities, settings = settings }
+        result = { special = true, ns = "", tg = '@rt@', dt = result.dt, at={ }, entities = entities, settings = settings }
         setmetatable(result, mt)
         local rdt = result.dt
         for k=1,#rdt do
             local v = rdt[k]
             if type(v) == "table" and not v.special then -- always table -)
                 result.ri = k -- rootindex
+v.__p__ = result  -- new, experiment, else we cannot go back to settings, we need to test this !
                 break
             end
         end
@@ -580,6 +643,19 @@ end
 
 xml.convert = xmlconvert
 
+function xml.inheritedconvert(data,xmldata)
+    local settings = xmldata.settings
+    settings.parent_root = xmldata -- to be tested
+ -- settings.no_root = true
+    local xc = xmlconvert(data,settings)
+ -- xc.settings = nil
+ -- xc.entities = nil
+ -- xc.special = nil
+ -- xc.ri = nil
+ -- print(xc.tg)
+    return xc
+end
+
 --[[ldx--
 <p>Packaging data in an xml like table is done with the following
 function. Maybe it will go away (when not used).</p>
@@ -590,7 +666,7 @@ function xml.is_valid(root)
 end
 
 function xml.package(tag,attributes,data)
-    local ns, tg = tag:match("^(.-):?([^:]+)$")
+    local ns, tg = match(tag,"^(.-):?([^:]+)$")
     local t = { ns = ns, tg = tg, dt = data or "", at = attributes or {} }
     setmetatable(t, mt)
     return t
@@ -608,21 +684,19 @@ the whole file first. The function accepts a string representing
 a filename or a file handle.</p>
 --ldx]]--
 
-function xml.load(filename)
+function xml.load(filename,settings)
+    local data = ""
     if type(filename) == "string" then
+     -- local data = io.loaddata(filename) - -todo: check type in io.loaddata
         local f = io.open(filename,'r')
         if f then
-            local root = xmlconvert(f:read("*all"))
+            data = f:read("*all")
             f:close()
-            return root
-        else
-            return xmlconvert("")
         end
     elseif filename then -- filehandle
-        return xmlconvert(filename:read("*all"))
-    else
-        return xmlconvert("")
+        data = filename:read("*all")
     end
+    return xmlconvert(data,settings)
 end
 
 --[[ldx--
@@ -654,7 +728,7 @@ local function copy(old,tables)
         if not tables[old] then
             tables[old] = new
         end
-        for k,v in pairs(old) do
+        for k,v in next, old do
             new[k] = (type(v) == "table" and (tables[v] or copy(v, tables))) or v
         end
         local mt = getmetatable(old)
@@ -684,7 +758,7 @@ function xml.checkbom(root) -- can be made faster
         local dt, found = root.dt, false
         for k=1,#dt do
             local v = dt[k]
-            if type(v) == "table" and v.special and v.tg == "@pi" and find(v.dt,"xml.*version=") then
+            if type(v) == "table" and v.special and v.tg == "@pi@" and find(v.dt[1],"xml.*version=") then
                 found = true
                 break
             end
@@ -713,7 +787,7 @@ local function verbose_element(e,handlers)
             ats[#ats+1] = format('%s=%q',k,v)
         end
     end
-    if ern and trace_remap and ern ~= ens then
+    if ern and trace_entities and ern ~= ens then
         ens = ern
     end
     if ens ~= "" then
@@ -844,7 +918,7 @@ local function newhandlers(settings)
     if settings then
         for k,v in next, settings do
             if type(v) == "table" then
-                tk = t[k] if not tk then tk = { } t[k] = tk end
+                local tk = t[k] if not tk then tk = { } t[k] = tk end
                 for kk,vv in next, v do
                     tk[kk] = vv
                 end
@@ -955,7 +1029,7 @@ local function xmltext(root) -- inline
     return (root and xmltostring(root)) or ""
 end
 
-function initialize_mt(root)
+initialize_mt = function(root) -- redefinition
     mt = { __tostring = xmltext, __index = root }
 end
 
@@ -990,12 +1064,37 @@ xml.string = xmlstring
 <p>A few helpers:</p>
 --ldx]]--
 
+--~ xmlsetproperty(root,"settings",settings)
+
+function xml.settings(e)
+    while e do
+        local s = e.settings
+        if s then
+            return s
+        else
+            e = e.__p__
+        end
+    end
+    return nil
+end
+
+function xml.root(e)
+    local r = e
+    while e do
+        e = e.__p__
+        if e then
+            r = e
+        end
+    end
+    return r
+end
+
 function xml.parent(root)
     return root.__p__
 end
 
 function xml.body(root)
-    return (root.ri and root.dt[root.ri]) or root
+    return (root.ri and root.dt[root.ri]) or root -- not ok yet
 end
 
 function xml.name(root)
@@ -1040,3 +1139,39 @@ function xml.assign(dt,k,root)
         return xml.body(root)
     end
 end
+
+-- the following helpers may move
+
+--[[ldx--
+<p>The next helper assigns a tree (or string). Usage:</p>
+<typing>
+xml.tocdata(e)
+xml.tocdata(e,"error")
+</typing>
+--ldx]]--
+
+function xml.tocdata(e,wrapper)
+    local whatever = xmltostring(e.dt)
+    if wrapper then
+        whatever = format("<%s>%s</%s>",wrapper,whatever,wrapper)
+    end
+    local t = { special = true, ns = "", tg = "@cd@", at = {}, rn = "", dt = { whatever }, __p__ = e }
+    setmetatable(t,getmetatable(e))
+    e.dt = { t }
+end
+
+function xml.makestandalone(root)
+    if root.ri then
+        local dt = root.dt
+        for k=1,#dt do
+            local v = dt[k]
+            if type(v) == "table" and v.special and v.tg == "@pi@" then
+                local txt = v.dt[1]
+                if find(txt,"xml.*version=") then
+                    v.dt[1] = txt .. " standalone='yes'"
+                    break
+                end
+            end
+        end
+    end
+end
