@@ -10,10 +10,12 @@ local utf = unicode.utf8
 
 local utfchar = utf.char
 local concat, insert, remove, gsub, find = table.concat, table.insert, table.remove
-local format, sub, gsub, find, gmatch = string.format, string.sub, string.gsub, string.find, string.gmatch
+local format, sub, gsub, find, gmatch, match = string.format, string.sub, string.gsub, string.find, string.gmatch, string.match
 local type, next, tonumber, tostring = type, next, tonumber, tostring
+local lpegmatch = lpeg.match
+local P, S, C, Cc = lpeg.P, lpeg.S, lpeg.C, lpeg.Cc
 
-if not tex and not tex.sprint then
+if not tex and not tex.sprint then -- no longer needed
     tex = {
         sprint = function(catcodes,...) texio.write(table.concat{...}) end,
         print  = function(catcodes,...) texio.write(table.concat{...}) end,
@@ -26,7 +28,7 @@ if not tex and not tex.sprint then
 end
 
 local texsprint, texprint, texwrite = tex.sprint, tex.print, tex.write
-local texcatcodes, ctxcatcodes, vrbcatcodes = tex.texcatcodes, tex.ctxcatcodes, tex.vrbcatcodes
+local texcatcodes, ctxcatcodes, vrbcatcodes, notcatcodes = tex.texcatcodes, tex.ctxcatcodes, tex.vrbcatcodes, tex.notcatcodes
 
 local xmlelements, xmlcollected, xmlsetproperty = xml.elements, xml.collected, xml.setproperty
 local xmlparseapply, xmlwithelements = xml.parse_apply, xml.withelements
@@ -41,10 +43,10 @@ local trace_loading  = false  trackers.register("lxml.loading",  function(v) tra
 local trace_access   = false  trackers.register("lxml.access",   function(v) trace_access   = v end)
 local trace_comments = false  trackers.register("lxml.comments", function(v) trace_comments = v end)
 
+local report_lxml = logs.new("lxml")
+
 lxml              = lxml or { }
 lxml.loaded       = lxml.loaded or { }
-lxml.noffiles     = 0
-lxml.nofconverted = 0
 
 local loaded = lxml.loaded
 
@@ -62,23 +64,27 @@ finalizers.tex = finalizers.tex or { }
 -- because we avoid tokenization of leading spaces and xml can be
 -- rather verbose (indented)
 
-local crlf      = lpeg.P("\r\n")
-local cr        = lpeg.P("\r")
-local lf        = lpeg.P("\n")
-local newline   = crlf + cr + lf
-local space     = lpeg.S(" \t\f\v")
-local ampersand = lpeg.P("&")
-local semicolon = lpeg.P(";")
+local newline   = lpeg.patterns.newline
+local space     = lpeg.patterns.spacer
+local ampersand = P("&")
+local semicolon = P(";")
 local spacing   = newline * space^0
-local content   = lpeg.C((1-spacing-ampersand)^1)
-local verbose   = lpeg.C((1-(space+newline))^1)
-local entity    = ampersand * lpeg.C((1-semicolon)^1) * semicolon
+local content   = C((1-spacing-ampersand)^1)
+local verbose   = C((1-(space+newline))^1)
+local entity    = ampersand * C((1-semicolon)^1) * semicolon
 
 local xmltextcapture = (
-    space^0 * newline^2  * lpeg.Cc("")            / texprint  + -- better ^-2 ?
-    space^0 * newline    * space^0 * lpeg.Cc(" ") / texsprint +
-    content                                       / texsprint + -- current catcodes regime is notcatcodes
-    entity                                        / xml.resolved_entity
+    space^0 * newline^2  * Cc("")            / texprint  + -- better ^-2 ?
+    space^0 * newline    * space^0 * Cc(" ") / texsprint +
+    content                                  / function(str) return texsprint(notcatcodes,str) end + -- was just texsprint, current catcodes regime is notcatcodes
+    entity                                  / xml.resolved_entity
+)^0
+
+local ctxtextcapture = (
+    space^0 * newline^2  * Cc("")            / texprint  + -- better ^-2 ?
+    space^0 * newline    * space^0 * Cc(" ") / texsprint +
+    content                                  / function(str) return texsprint(ctxcatcodes,str) end + -- was just texsprint, current catcodes regime is notcatcodes
+    entity                                   / xml.resolved_entity
 )^0
 
 local forceraw, rawroot = false, nil
@@ -129,7 +135,7 @@ local xmlverbosecapture = (
 
 local function toverbatim(str)
     if beforecommand then texsprint(texcatcodes,beforecommand,"{}") end
-    xmlverbosecapture:match(str)
+    lpegmatch(xmlverbosecapture,str)
     if aftercommand  then texsprint(texcatcodes,aftercommand,"{}")  end
 end
 
@@ -158,7 +164,7 @@ function lxml.toverbatim(str)
     -- todo: add this to capture
     str = gsub(str,"^[ \t]+[\n\r]+","")
     str = gsub(str,"[ \t\n\r]+$","")
-    xmlverbosecapture:match(str)
+    lpegmatch(xmlverbosecapture,str)
     if aftercommand  then texsprint(texcatcodes,aftercommand,"{}")  end
 end
 
@@ -172,13 +178,12 @@ function lxml.store(id,root,filename)
     end
 end
 
---~ local splitter = lpeg.C((1-lpeg.P(":"))^1) * lpeg.P("::") * lpeg.C(lpeg.P(1)^1)
 local splitter = lpeg.splitat("::")
 
 lxml.idsplitter = splitter
 
 function lxml.splitid(id)
-    local d, i = splitter:match(id)
+    local d, i = lpegmatch(splitter,id)
     if d then
         return d, i
     else
@@ -194,7 +199,7 @@ local function get_id(id, qualified)
         elseif type(id) == "table" then
             return id
         else
-            local d, i = splitter:match(id)
+            local d, i = lpegmatch(splitter,id)
             if d then
                 local ld = loaded[d]
                 if ld then
@@ -208,20 +213,20 @@ local function get_id(id, qualified)
                                 return root
                             end
                         elseif trace_access then
-                            logs.report("lxml","'%s' has no index entry '%s'",d,i)
+                            report_lxml("'%s' has no index entry '%s'",d,i)
                         end
                     elseif trace_access then
-                        logs.report("lxml","'%s' has no index",d)
+                        report_lxml("'%s' has no index",d)
                     end
                 elseif trace_access then
-                    logs.report("lxml","'%s' is not loaded",d)
+                    report_lxml("'%s' is not loaded",d)
                 end
             elseif trace_access then
-                logs.report("lxml","'%s' is not loaded",i)
+                report_lxml("'%s' is not loaded",i)
             end
         end
     elseif trace_access then
-        logs.report("lxml","invalid id (nil)")
+        report_lxml("invalid id (nil)")
     end
 end
 
@@ -267,7 +272,7 @@ local function addindex(name,check_sum,force)
         root.index = index
         root.maxindex = maxindex
         if trace_access then
-            logs.report("lxml","%s indexed, %s nodes",tostring(name),maxindex)
+            report_lxml("%s indexed, %s nodes",tostring(name),maxindex)
         end
     end
 end
@@ -301,7 +306,7 @@ function lxml.checkindex(name)
 end
 
 function lxml.withindex(name,n,command) -- will change as name is always there now
-    local i, p = splitter:match(n)
+    local i, p = lpegmatch(splitter,n)
     if p then
         texsprint(ctxcatcodes,"\\xmlw{",command,"}{",n,"}")
     else
@@ -310,7 +315,7 @@ function lxml.withindex(name,n,command) -- will change as name is always there n
 end
 
 function lxml.getindex(name,n) -- will change as name is always there now
-    local i, p = splitter:match(n)
+    local i, p = lpegmatch(splitter,n)
     if p then
         texsprint(ctxcatcodes,n)
     else
@@ -324,12 +329,11 @@ xml.originalload = xml.originalload or xml.load
 
 local noffiles, nofconverted = 0, 0
 
-function xml.load(filename)
-    noffiles = noffiles + 1
-    nofconverted = nofconverted + 1
+function xml.load(filename,settings)
+    noffiles, nofconverted = noffiles + 1, nofconverted + 1
     starttiming(xml)
     local ok, data = resolvers.loadbinfile(filename)
-    local xmltable = xml.convert((ok and data) or "")
+    local xmltable = xml.convert((ok and data) or "",settings)
     stoptiming(xml)
     return xmltable
 end
@@ -341,7 +345,9 @@ local function entityconverter(id,str)
 end
 
 function lxml.convert(id,data,entities,compress)
-    local settings = { }
+    local settings = {
+        unify_predefined_entities = true,
+    }
     if compress and compress == variables.yes then
         settings.strip_cm_and_dt = true
     end
@@ -357,6 +363,7 @@ function lxml.load(id,filename,compress,entities)
     if trace_loading then
         commands.writestatus("lxml","loading file '%s' as '%s'",filename,id)
     end
+    noffiles, nofconverted = noffiles + 1, nofconverted + 1
  -- local xmltable = xml.load(filename)
     local ok, data = resolvers.loadbinfile(filename)
     local xmltable = lxml.convert(id,(ok and data) or "",compress,entities)
@@ -439,7 +446,7 @@ end
 
 local function tex_comment(e,handlers)
     if trace_comments then
-        logs.report("lxml","comment: %s",e.dt[1])
+        report_lxml("comment: %s",e.dt[1])
     end
 end
 
@@ -465,7 +472,7 @@ local function tex_element(e,handlers)
                 end
                 texsprint(ctxcatcodes,"\\xmlw{",command,"}{",rootname,"::",ix,"}")
             else
-                logs.report("lxml", "fatal error: no index for '%s'",command)
+                report_lxml( "fatal error: no index for '%s'",command)
                 texsprint(ctxcatcodes,"\\xmlw{",command,"}{",ix or 0,"}")
             end
         elseif tc == "function" then
@@ -476,19 +483,19 @@ end
 
 local pihandlers = { }  xml.pihandlers = pihandlers
 
-local kind   = lpeg.P("context-") * lpeg.C((1-lpeg.P("-"))^1) * lpeg.P("-directive")
-local space  = lpeg.S(" \n\r")
+local kind   = P("context-") * C((1-P("-"))^1) * P("-directive")
+local space  = S(" \n\r")
 local spaces = space^0
-local class  = lpeg.C((1-space)^0)
+local class  = C((1-space)^0)
 local key    = class
-local value  = lpeg.C(lpeg.P(1-(space * -1))^0)
+local value  = C(P(1-(space * -1))^0)
 
 local parser = kind * spaces * class * spaces * key * spaces * value
 
 pihandlers[#pihandlers+1] = function(str)
---  local kind, class, key, value = parser:match(str)
+--  local kind, class, key, value = lpegmatch(parser,str)
     if str then
-        local a, b, c, d = parser:match(str)
+        local a, b, c, d = lpegmatch(parser,str)
         if d then
             texsprint(ctxcatcodes,"\\xmlcontextdirective{",a",}{",b,"}{",c,"}{",d,"}")
         end
@@ -508,12 +515,16 @@ local function tex_cdata(e,handlers)
     end
 end
 
-local function tex_text(e,handlers)
-    xmltextcapture:match(e)
+local function tex_text(e)
+    lpegmatch(xmltextcapture,e)
+end
+
+local function ctx_text(e)
+    lpegmatch(ctxtextcapture,e)
 end
 
 local function tex_handle(...)
---  logs.report("lxml", "error while flushing: %s", concat { ... })
+--  report_lxml( "error while flushing: %s", concat { ... })
     texsprint(...) -- notcatcodes is active anyway
 end
 
@@ -532,6 +543,8 @@ local xmltexhandler = xml.newhandlers {
     }
 }
 
+lxml.xmltexhandler = xmltexhandler
+
 function lxml.serialize(root)
     xmlserialize(root,xmltexhandler)
 end
@@ -550,7 +563,7 @@ local function sprint(root)
         local tr = type(root)
         if tr == "string" then -- can also be result of lpath
          -- rawroot = false
-            xmltextcapture:match(root)
+            lpegmatch(xmltextcapture,root)
         elseif tr == "table" then
             if forceraw then
                 rawroot = root
@@ -574,7 +587,7 @@ local function tprint(root) -- we can move sprint inline
             end
         end
     elseif tr == "string" then
-        xmltextcapture:match(root)
+        lpegmatch(xmltextcapture,root)
     end
 end
 
@@ -584,7 +597,7 @@ local function cprint(root) -- content
         -- quit
     elseif type(root) == 'string' then
      -- rawroot = false
-        xmltextcapture:match(root)
+        lpegmatch(xmltextcapture,root)
     else
         local rootdt = root.dt
         if forceraw then
@@ -618,7 +631,7 @@ end
 --~
 --~ local xmllineshandler = table.copy(xmltexhandler)
 --~
---~ xmllineshandler.handle = function(...) xmllinescapture:match(concat{ ... }) end
+--~ xmllineshandler.handle = function(...) lpegmatch(xmllinescapture,concat{ ... }) end
 --~
 --~ function lines(root)
 --~     if not root then
@@ -626,7 +639,7 @@ end
 --~      -- quit
 --~     elseif type(root) == 'string' then
 --~      -- rawroot = false
---~         xmllinescapture:match(root)
+--~         lpegmatch(xmllinescapture,root)
 --~     elseif next(root) then -- tr == 'table'
 --~         xmlserialize(root,xmllineshandler)
 --~     end
@@ -698,8 +711,9 @@ function lxml.installsetup(what,document,setup,where)
 end
 
 function lxml.flushsetups(id,...)
-    local done = { }
-    for _, document in ipairs({...}) do
+    local done, list = { }, { ... }
+    for i=1,#list do
+        local document = list[i]
         local sd = setups[document]
         if sd then
             for k=1,#sd do
@@ -708,7 +722,6 @@ function lxml.flushsetups(id,...)
                     if trace_loading then
                         commands.writestatus("lxml","applying setup %02i = %s to %s",k,v,document)
                     end
---~                     texsprint(ctxcatcodes,"\\directsetup{",v,"}")
                     texsprint(ctxcatcodes,"\\xmlsetup{",id,"}{",v,"}")
                     done[v] = true
                 end
@@ -751,19 +764,19 @@ function lxml.setsetup(id,pattern,setup)
                     local ix = e.ix or 0
                     if setup == "-" then
                         e.command = false
-                        logs.report("lxml","lpath matched (a) %5i: %s = %s -> skipped",c,ix,setup)
+                        report_lxml("lpath matched (a) %5i: %s = %s -> skipped",c,ix,setup)
                     elseif setup == "+" then
                         e.command = true
-                        logs.report("lxml","lpath matched (b) %5i: %s = %s -> text",c,ix,setup)
+                        report_lxml("lpath matched (b) %5i: %s = %s -> text",c,ix,setup)
                     else
                         local tg = e.tg
                         if tg then -- to be sure
                             e.command = tg
                             local ns = e.rn or e.ns
                             if ns == "" then
-                                logs.report("lxml","lpath matched (c) %5i: %s = %s -> %s",c,ix,tg,tg)
+                                report_lxml("lpath matched (c) %5i: %s = %s -> %s",c,ix,tg,tg)
                             else
-                                logs.report("lxml","lpath matched (d) %5i: %s = %s:%s -> %s",c,ix,ns,tg,tg)
+                                report_lxml("lpath matched (d) %5i: %s = %s:%s -> %s",c,ix,ns,tg,tg)
                             end
                         end
                     end
@@ -781,10 +794,10 @@ function lxml.setsetup(id,pattern,setup)
                 end
             end
         elseif trace_setups then
-            logs.report("lxml","no lpath matches for %s",pattern)
+            report_lxml("no lpath matches for %s",pattern)
         end
     else
-        local a, b = setup:match("^(.+:)([%*%-])$")
+        local a, b = match(setup,"^(.+:)([%*%-])$")
         if a and b then
             local collected = lxmlparseapply(id,pattern)
             if collected then
@@ -795,23 +808,23 @@ function lxml.setsetup(id,pattern,setup)
                         if b == "-" then
                             e.command = false
                             if ns == "" then
-                                logs.report("lxml","lpath matched (e) %5i: %s = %s -> skipped",c,ix,tg)
+                                report_lxml("lpath matched (e) %5i: %s = %s -> skipped",c,ix,tg)
                             else
-                                logs.report("lxml","lpath matched (f) %5i: %s = %s:%s -> skipped",c,ix,ns,tg)
+                                report_lxml("lpath matched (f) %5i: %s = %s:%s -> skipped",c,ix,ns,tg)
                             end
                         elseif b == "+" then
                             e.command = true
                             if ns == "" then
-                                logs.report("lxml","lpath matched (g) %5i: %s = %s -> text",c,ix,tg)
+                                report_lxml("lpath matched (g) %5i: %s = %s -> text",c,ix,tg)
                             else
-                                logs.report("lxml","lpath matched (h) %5i: %s = %s:%s -> text",c,ix,ns,tg)
+                                report_lxml("lpath matched (h) %5i: %s = %s:%s -> text",c,ix,ns,tg)
                             end
                         else
                             e.command = a .. tg
                             if ns == "" then
-                                logs.report("lxml","lpath matched (i) %5i: %s = %s -> %s",c,ix,tg,e.command)
+                                report_lxml("lpath matched (i) %5i: %s = %s -> %s",c,ix,tg,e.command)
                             else
-                                logs.report("lxml","lpath matched (j) %5i: %s = %s:%s -> %s",c,ix,ns,tg,e.command)
+                                report_lxml("lpath matched (j) %5i: %s = %s:%s -> %s",c,ix,ns,tg,e.command)
                             end
                         end
                     end
@@ -828,7 +841,7 @@ function lxml.setsetup(id,pattern,setup)
                     end
                 end
             elseif trace_setups then
-                logs.report("lxml","no lpath matches for %s",pattern)
+                report_lxml("no lpath matches for %s",pattern)
             end
         else
             local collected = lxmlparseapply(id,pattern)
@@ -839,9 +852,9 @@ function lxml.setsetup(id,pattern,setup)
                         e.command = setup
                         local ns, tg, ix = e.rn or e.ns, e.tg, e.ix or 0
                         if ns == "" then
-                            logs.report("lxml","lpath matched (k) %5i: %s = %s -> %s",c,ix,tg,setup)
+                            report_lxml("lpath matched (k) %5i: %s = %s -> %s",c,ix,tg,setup)
                         else
-                            logs.report("lxml","lpath matched (l) %5i: %s = %s:%s -> %s",c,ix,ns,tg,setup)
+                            report_lxml("lpath matched (l) %5i: %s = %s:%s -> %s",c,ix,ns,tg,setup)
                         end
                     end
                 else
@@ -850,7 +863,7 @@ function lxml.setsetup(id,pattern,setup)
                     end
                 end
             elseif trace_setups then
-                logs.report("lxml","no lpath matches for %s",pattern)
+                report_lxml("no lpath matches for %s",pattern)
             end
         end
     end
@@ -893,6 +906,7 @@ local function count(collected)
 end
 
 local function position(collected,n)
+    -- todo: if not n then == match
     if collected then
         n = tonumber(n) or 0
         if n < 0 then
@@ -904,7 +918,7 @@ local function position(collected,n)
     end
 end
 
-local function match(collected)
+local function match(collected) -- is match in preceding collected, never change, see bibxml
     texwrite((collected and collected[1].mi) or 0)
 end
 
@@ -920,9 +934,10 @@ local function index(collected,n)
     end
 end
 
-local function command(collected,cmd)
-    if collected then
-        for c=1,#collected do
+local function command(collected,cmd,otherwise)
+    local n = collected and #collected
+    if n and n > 0 then
+        for c=1,n do
             local e = collected[c]
             local ix = e.ix
             if not ix then
@@ -931,18 +946,20 @@ local function command(collected,cmd)
             end
             texsprint(ctxcatcodes,"\\xmlw{",cmd,"}{",e.name,"::",ix,"}")
         end
+    elseif otherwise then
+        texsprint(ctxcatcodes,"\\xmlw{",otherwise,"}{#1}")
     end
 end
 
 local function attribute(collected,a,default)
-    if collected then
+    if collected and #collected > 0 then
         local at = collected[1].at
         local str = (at and at[a]) or default
         if str and str ~= "" then
-            texsprint(ctxcatcodes,str)
+            texsprint(notcatcodes,str)
         end
     elseif default then
-        texsprint(ctxcatcodes,default)
+        texsprint(notcatcodes,default)
     end
 end
 
@@ -954,7 +971,7 @@ local function chainattribute(collected,arguments) -- todo: optional levels
             if at then
                 local a = at[arguments]
                 if a then
-                    texsprint(ctxcatcodes,a)
+                    texsprint(notcatcodes,a)
                 end
             else
                 break -- error
@@ -975,6 +992,32 @@ local function text(collected)
     end
 end
 
+local function ctxtext(collected)
+    if collected then
+        for c=1,#collected do
+            texsprint(ctxcatcodes,collected[1].dt)
+        end
+    end
+end
+
+local lowerchars, upperchars = characters.lower, characters.upper
+
+local function lower(collected)
+    if collected then
+        for c=1,#collected do
+            texsprint(ctxcatcodes,lowerchars(collected[1].dt[1]))
+        end
+    end
+end
+
+local function upper(collected)
+    if collected then
+        for c=1,#collected do
+            texsprint(ctxcatcodes,upperchars(collected[1].dt[1]))
+        end
+    end
+end
+
 local function number(collected)
     if collected then
         local n = 0
@@ -1021,6 +1064,10 @@ finalizers.count          = count
 finalizers.command        = command
 finalizers.attribute      = attribute
 finalizers.text           = text
+finalizers.lower          = lower
+finalizers.upper          = upper
+finalizers.ctxtext        = ctxtext
+finalizers.context        = ctxtext
 finalizers.position       = position
 finalizers.match          = match
 finalizers.index          = index
@@ -1142,6 +1189,19 @@ function lxml.raw(id,pattern) -- the content, untouched by commands
     end
 end
 
+function lxml.context(id,pattern) -- the content, untouched by commands
+    if not pattern then
+        local collected = get_id(id)
+    --  texsprint(ctxcatcodes,collected.dt[1])
+        ctx_text(collected.dt[1])
+    else
+        local collected = lxmlparseapply(id,pattern) or get_id(id)
+        if collected and #collected > 0 then
+            texsprint(ctxcatcodes,collected[1].dt)
+        end
+    end
+end
+
 function lxml.text(id,pattern)
     local collected = (pattern and lxmlparseapply(id,pattern)) or get_id(id)
     if collected then
@@ -1190,10 +1250,10 @@ function lxml.att(id,a,default)
         local at = root.at
         local str = (at and at[a]) or default
         if str and str ~= "" then
-            texsprint(ctxcatcodes,str)
+            texsprint(notcatcodes,str)
         end
     elseif default then
-        texsprint(ctxcatcodes,default)
+        texsprint(notcatcodes,default)
     end
 end
 
@@ -1301,9 +1361,10 @@ statistics.register("xml load time", function()
 end)
 
 statistics.register("lxml preparation time", function()
-    if noffiles > 0 or nofconverted > 0 then
+    local calls, cached = xml.lpathcalls(), xml.lpathcached()
+    if calls > 0 or cached > 0 then
         return format("%s seconds, %s nodes, %s lpath calls, %s cached calls",
-            statistics.elapsedtime(lxml), nofindices, xml.lpathcalls(), xml.lpathcached())
+            statistics.elapsedtime(lxml), nofindices, calls, cached)
     else
         return nil
     end
@@ -1330,15 +1391,14 @@ statistics.register("lxml lpath profile", function()
     end
 end)
 
-
 -- misc
 
 function lxml.nonspace(id,pattern) -- slow, todo loop
     xmltprint(xmlcollect(get_id(id),pattern,true))
 end
 
-function lxml.strip(id,pattern,nolines)
-    xml.strip(get_id(id),pattern,nolines)
+function lxml.strip(id,pattern,nolines,anywhere)
+    xml.strip(get_id(id),pattern,nolines,anywhere)
 end
 
 function lxml.stripped(id,pattern,nolines)
