@@ -20,12 +20,14 @@ if not modules then modules = { } end modules ['math-map'] = {
 -- maybe: script/scriptscript dynamic,
 
 local type, next = type, next
-local floor = math.floor
+local floor, div = math.floor, math.div
 
 local texattribute = tex.attribute
 
 local trace_greek  = false  trackers.register("math.greek",  function(v) trace_greek = v end)
 
+local report_math = logs.new("mathematics")
+
 mathematics = mathematics or { }
 
 -- we could use one level less and have tf etc be tables directly but the
@@ -213,6 +215,15 @@ mathematics.alphabets = {
                 [0x0005A]=0x02124,
             },
             lcletters = 0x1D552,
+            lcgreek = { -- gamma pi
+                [0x03B3]=0x0213C, [0x03C0]=0x0213D,
+            },
+            ucgreek = { -- Gamma pi
+                [0x0393]=0x0213E, [0x03A0]=0x0213F,
+            },
+            symbols = { -- sum
+              [0x2211]=0x02140,
+            },
         },
     },
     fraktur = { -- ok
@@ -294,9 +305,10 @@ alphabets.monospaced.it         = alphabets.sansserif.tf
 alphabets.monospaced.bf         = alphabets.sansserif.tf
 alphabets.monospaced.bi         = alphabets.sansserif.bf
 
-alphabets.blackboard.tf.symbols = alphabets.regular.tf.symbols
-alphabets.blackboard.tf.lcgreek = alphabets.regular.tf.lcgreek
-alphabets.blackboard.tf.ucgreek = alphabets.regular.tf.ucgreek
+alphabets.blackboard.tf.symbols = table.merged(alphabets.regular.tf.symbols, alphabets.blackboard.tf.symbols)
+alphabets.blackboard.tf.lcgreek = table.merged(alphabets.regular.tf.lcgreek, alphabets.blackboard.tf.lcgreek)
+alphabets.blackboard.tf.ucgreek = table.merged(alphabets.regular.tf.ucgreek, alphabets.blackboard.tf.ucgreek)
+
 alphabets.blackboard.it         = alphabets.blackboard.tf
 alphabets.blackboard.bf         = alphabets.blackboard.tf
 alphabets.blackboard.bi         = alphabets.blackboard.bf
@@ -312,16 +324,16 @@ alphabets.fraktur.bf.ucgreek    = alphabets.regular.bf.ucgreek
 alphabets.fraktur.it            = alphabets.fraktur.tf
 alphabets.fraktur.bi            = alphabets.fraktur.bf
 
-alphabets.script.tf.digits    = alphabets.regular.tf.digits
-alphabets.script.tf.symbols   = alphabets.regular.tf.symbols
-alphabets.script.tf.lcgreek   = alphabets.regular.tf.lcgreek
-alphabets.script.tf.ucgreek   = alphabets.regular.tf.ucgreek
-alphabets.script.bf.digits    = alphabets.regular.bf.digits
-alphabets.script.bf.symbols   = alphabets.regular.bf.symbols
-alphabets.script.bf.lcgreek   = alphabets.regular.bf.lcgreek
-alphabets.script.bf.ucgreek   = alphabets.regular.bf.ucgreek
-alphabets.script.it           = alphabets.script.tf
-alphabets.script.bi           = alphabets.script.bf
+alphabets.script.tf.digits      = alphabets.regular.tf.digits
+alphabets.script.tf.symbols     = alphabets.regular.tf.symbols
+alphabets.script.tf.lcgreek     = alphabets.regular.tf.lcgreek
+alphabets.script.tf.ucgreek     = alphabets.regular.tf.ucgreek
+alphabets.script.bf.digits      = alphabets.regular.bf.digits
+alphabets.script.bf.symbols     = alphabets.regular.bf.symbols
+alphabets.script.bf.lcgreek     = alphabets.regular.bf.lcgreek
+alphabets.script.bf.ucgreek     = alphabets.regular.bf.ucgreek
+alphabets.script.it             = alphabets.script.tf
+alphabets.script.bi             = alphabets.script.bf
 
 alphabets.tt = alphabets.monospaced
 alphabets.ss = alphabets.sansserif
@@ -390,7 +402,7 @@ function mathematics.remap_alphabets(char,mathalphabet,mathgreek)
                 local alphabet = r and r.alphabet or "regular"
                 local style = r and r.style or "tf"
                 if trace_greek then
-                    logs.report("math","before: char: %05X, alphabet: %s %s, lcgreek: %s, ucgreek: %s",char,alphabet,style,remapping[lc].what,remapping[uc].what)
+                    report_math("before: char: %05X, alphabet: %s %s, lcgreek: %s, ucgreek: %s",char,alphabet,style,remapping[lc].what,remapping[uc].what)
                 end
                 local s = remapping[islc or isuc][style]
                 if s then
@@ -398,7 +410,7 @@ function mathematics.remap_alphabets(char,mathalphabet,mathgreek)
                     mathalphabet, style = data and data.attribute or mathalphabet, s
                 end
                 if trace_greek then
-                    logs.report("math","after : char: %05X, alphabet: %s %s, lcgreek: %s, ucgreek: %s",char,alphabet,style,remapping[lc].what,remapping[uc].what)
+                    report_math("after : char: %05X, alphabet: %s %s, lcgreek: %s, ucgreek: %s",char,alphabet,style,remapping[lc].what,remapping[uc].what)
                 end
             end
         end
@@ -410,13 +422,13 @@ function mathematics.remap_alphabets(char,mathalphabet,mathgreek)
             -- nothing to remap
         elseif char >= 0x030 and char <= 0x039 then
             local o = offset.digits
-            newchar = (type(o) == "table" and (o[char] or char)) or (char - 0x030 + o)
+            newchar = o and ((type(o) == "table" and (o[char] or char)) or (char - 0x030 + o))
         elseif char >= 0x041 and char <= 0x05A then
             local o = offset.ucletters
-            newchar = (type(o) == "table" and (o[char] or char)) or (char - 0x041 + o)
+            newchar = o and ((type(o) == "table" and (o[char] or char)) or (char - 0x041 + o))
         elseif char >= 0x061 and char <= 0x07A then
             local o = offset.lcletters
-            newchar = (type(o) == "table" and (o[char] or char)) or (char - 0x061 + o)
+            newchar = o and ((type(o) == "table" and (o[char] or char)) or (char - 0x061 + o))
         elseif islcgreek[char] then
             newchar = offset.lcgreek[char]
         elseif isucgreek[char] then
