@@ -10,10 +10,12 @@ if not modules then modules = { } end modules ['meta-pdf'] = {
 -- meta-pdh.lua but since we no longer want to overload functione we use
 -- more locals now. This module keeps changing as it is also a testbed.
 
-local concat, format, gsub, find = table.concat, string.format, string.gsub, string.find
-local byte, round = string.byte, math.round
-local texsprint = tex.sprint
-local ctxcatcodes = tex.ctxcatcodes
+local concat, format, gsub, find, byte, gmatch, match = table.concat, string.format, string.gsub, string.find, string.byte, string.gmatch, string.match
+local lpegmatch = lpeg.match
+local round = math.round
+local texsprint, ctxcatcodes = tex.sprint, tex.ctxcatcodes
+
+local report_mptopdf = logs.new("mptopdf")
 
 local pdfrgbcode                = lpdf.rgbcode
 local pdfcmykcode               = lpdf.cmykcode
@@ -55,7 +57,7 @@ local function texcode(str)
     texsprint(ctxcatcodes,str)
 end
 
-function mpscode(str)
+local function mpscode(str)
     if ignore_path then
         pdfcode("h W n")
         if extra_path_code then
@@ -304,9 +306,9 @@ end
 
 -- not supported in mkiv , use mplib instead
 
-handlers[10] = function() logs.report("mptopdf","skipping special %s",10) end
-handlers[20] = function() logs.report("mptopdf","skipping special %s",20) end
-handlers[50] = function() logs.report("mptopdf","skipping special %s",50) end
+handlers[10] = function() report_mptopdf("skipping special %s",10) end
+handlers[20] = function() report_mptopdf("skipping special %s",20) end
+handlers[50] = function() report_mptopdf("skipping special %s",50) end
 
 --end of not supported
 
@@ -320,7 +322,7 @@ function mps.setrgbcolor(r,g,b) -- extra check
         if handler then
             handler(s)
         else
-            logs.report("mptopdf","unknown special handler %s (1)",h)
+            report_mptopdf("unknown special handler %s (1)",h)
         end
     elseif r == 0.123 and g < 0.1 then
         g, b = round(g*1000), round(b*1000)
@@ -330,7 +332,7 @@ function mps.setrgbcolor(r,g,b) -- extra check
         if handler then
             handler(s)
         else
-            logs.report("mptopdf","unknown special handler %s (2)",h)
+            report_mptopdf("unknown special handler %s (2)",h)
         end
     else
         pdfcode(pdffinishtransparencycode())
@@ -390,7 +392,7 @@ local text  = lpegCc("{") * (
 local package = lpegCs(spec + text^0)
 
 function mps.fshow(str,font,scale) -- lpeg parser
-    mps.textext(font,scale,package:match(str))
+    mps.textext(font,scale,lpegmatch(package,str))
 end
 
 local cnumber = lpegC(number)
@@ -510,9 +512,9 @@ local captures_new = ( space + verbose + procset + preamble )^0
 
 local function parse(m_data)
     if find(m_data,"%%%%BeginResource: procset mpost") then
-        captures_new:match(m_data)
+        lpegmatch(captures_new,m_data)
     else
-        captures_old:match(m_data)
+        lpegmatch(captures_old,m_data)
     end
 end
 
