@@ -6,10 +6,12 @@ if not modules then modules = { } end modules ['mlib-pdf'] = {
     license   = "see context related readme files",
 }
 
-local format, concat = string.format, table.concat
+local format, concat, gsub = string.format, table.concat, string.gsub
 local texsprint = tex.sprint
 local abs, sqrt, round = math.abs, math.sqrt, math.round
 
+local report_mplib = logs.new("mplib")
+
 local copy_node, write_node = node.copy, node.write
 
 local ctxcatcodes = tex.ctxcatcodes
@@ -69,7 +71,7 @@ function metapost.flush_literal(d) -- \def\MPLIBtoPDF#1{\ctxlua{metapost.flush_l
         literal.data = savedliterals[d]
         write_node(literal)
     else
-        logs.report("metapost","problem flushing literal %s",d)
+        report_mplib("problem flushing literal %s",d)
     end
 end
 
@@ -119,7 +121,7 @@ function metapost.flushers.pdf.flushfigure(pdfliterals) -- table
 end
 
 function metapost.flushers.pdf.textfigure(font,size,text,width,height,depth) -- we could save the factor
-    text = text:gsub(".","\\hbox{%1}") -- kerning happens in metapost (i have to check if this is true for mplib)
+    text = gsub(text,".","\\hbox{%1}") -- kerning happens in metapost (i have to check if this is true for mplib)
     texsprint(ctxcatcodes,format("\\MPLIBtextext{%s}{%s}{%s}{%s}{%s}",font,size,text,0,-number.dimenfactors.bp*depth))
 end
 
@@ -155,6 +157,7 @@ local function curved(ith,pth)
 end
 
 local function flushnormalpath(path, t, open)
+    t = t or { }
     local pth, ith
     for i=1,#path do
         pth = path[i]
@@ -183,6 +186,7 @@ local function flushnormalpath(path, t, open)
 end
 
 local function flushconcatpath(path, t, open)
+    t = t or { }
     t[#t+1] = format("%f %f %f %f %f %f cm", sx, rx, ry, sy, tx ,ty)
     local pth, ith
     for i=1,#path do
