@@ -8,10 +8,15 @@ if not modules then modules = { } end modules ['mlib-pps'] = { -- prescript, pos
 
 -- current limitation: if we have textext as well as a special color then due to
 -- prescript/postscript overload we can have problems
+--
+-- todo: report max textexts
 
 local format, gmatch, concat, round, match = string.format, string.gmatch, table.concat, math.round, string.match
 local sprint = tex.sprint
 local tonumber, type = tonumber, type
+local lpegmatch = lpeg.match
+local texbox = tex.box
+local copy_list = node.copy_list
 
 local starttiming, stoptiming = statistics.starttiming, statistics.stoptiming
 
@@ -19,6 +24,8 @@ local ctxcatcodes = tex.ctxcatcodes
 
 local trace_textexts = false  trackers.register("metapost.textexts", function(v) trace_textexts = v end)
 
+local report_mplib = logs.new("mplib")
+
 colors = colors or { }
 
 local rgbtocmyk  = colors.rgbtocmyk   or function() return 0,0,0,1 end
@@ -112,14 +119,14 @@ function metapost.specials.register(str) -- only colors
             if cc then
                 cc[n] = data
             else
-                logs.report("mplib","problematic special: %s (no colordata class %s)", str or "?",class)
+                report_mplib("problematic special: %s (no colordata class %s)", str or "?",class)
             end
         else
          -- there is some bug to be solved, so we issue a message
-            logs.report("mplib","problematic special: %s", str or "?")
+            report_mplib("problematic special: %s", str or "?")
         end
     end
---~     if str:match("^%%%%MetaPostOption: multipass") then
+--~     if match(str,"^%%%%MetaPostOption: multipass") then
 --~         metapost.multipass = true
 --~     end
 end
@@ -203,11 +210,11 @@ local colorsplitter         = lpeg.Ct(lpeg.splitter(":",tonumber))
 -- This is also an example of a simple plugin.
 
 --~ function metapost.specials.cc(specification,object,result)
---~     object.color = specificationsplitter:match(specification)
+--~     object.color = lpegmatch(specificationsplitter,specification)
 --~     return object, nil, nil, nil
 --~ end
 --~ function metapost.specials.cc(specification,object,result)
---~     local c = specificationsplitter:match(specification)
+--~     local c = lpegmatch(specificationsplitter,specification)
 --~     local o = object.color[1]
 --~     c[1],c[2],c[3],c[4] = o*c[1],o*c[2],o*c[3],o*c[4]
 --~     return object, nil, nil, nil
@@ -243,7 +250,7 @@ function metapost.specials.ps(specification,object,result) -- positions
     local label = specification
     x = x - metapost.llx
     y = metapost.ury - y
- -- logs.report("mplib", "todo: position '%s' at (%s,%s) with (%s,%s)",label,x,y,w,h)
+ -- report_mplib( "todo: position '%s' at (%s,%s) with (%s,%s)",label,x,y,w,h)
     sprint(ctxcatcodes,format("\\dosavepositionwhd{%s}{0}{%sbp}{%sbp}{%sbp}{%sbp}{0pt}",label,x,y,w,h))
     return { }, nil, nil, nil
 end
@@ -276,10 +283,10 @@ function metapost.specials.cs(specification,object,result,flusher) -- spot color
     nofshades = nofshades + 1
     flusher.flushfigure(result)
     result = { }
-    local t = specificationsplitter:match(specification)
+    local t = lpegmatch(specificationsplitter,specification)
     -- we need a way to move/scale
-    local ca = colorsplitter:match(t[4])
-    local cb = colorsplitter:match(t[8])
+    local ca = lpegmatch(colorsplitter,t[4])
+    local cb = lpegmatch(colorsplitter,t[8])
     if round(ca[1]*10000) == 123 then ca = metapost.colorspec(ca) end
     if round(cb[1]*10000) == 123 then cb = metapost.colorspec(cb) end
     local name = format("MplSh%s",nofshades)
@@ -348,10 +355,10 @@ function metapost.specials.ls(specification,object,result,flusher)
     nofshades = nofshades + 1
     flusher.flushfigure(result)
     result = { }
-    local t = specificationsplitter:match(specification)
+    local t = lpegmatch(specificationsplitter,specification)
     -- we need a way to move/scale
-    local ca = colorsplitter:match(t[4])
-    local cb = colorsplitter:match(t[7])
+    local ca = lpegmatch(colorsplitter,t[4])
+    local cb = lpegmatch(colorsplitter,t[7])
     if round(ca[1]*10000) == 123 then ca = metapost.colorspec(ca) end
     if round(cb[1]*10000) == 123 then cb = metapost.colorspec(cb) end
     local name = format("MpSh%s",nofshades)
@@ -417,36 +424,61 @@ end
 
 -- no need for a before here
 
-local current_format, current_graphic
+local current_format, current_graphic, current_initializations
 
-metapost.first_box       = metapost.first_box or 1000
-metapost.last_box        = metapost.last_box or 1100
-metapost.textext_current = metapost.first_box
+-- metapost.first_box       = metapost.first_box or 1000
+-- metapost.last_box        = metapost.last_box or 1100
+--~ metapost.textext_current = metapost.first_box
 metapost.multipass       = false
 
+local textexts = { }
+
 function metapost.free_boxes() -- todo: mp direct list ipv box
-    local tb = tex.box
-    for i = metapost.first_box,metapost.last_box do
-        local b = tb[i]
-        if b then
-            tb[i] = nil -- no node.flush_list(b) needed, else double free error
-        else
-            break
+ -- for i = metapost.first_box,metapost.last_box do
+ --     local b = texbox[i]
+ --     if b then
+ --         texbox[i] = nil -- no node.flush_list(b) needed, else double free error
+ --     else
+ --         break
+ --     end
+ -- end
+    for n, box in next, textexts do
+        local tn = textexts[n]
+        if tn then
+            -- somehow not flushed (used)
+            textexts[n] = nil
         end
     end
+    textexts = { }
+end
+
+function metapost.settext(box,slot)
+    textexts[slot] = copy_list(texbox[box])
+    texbox[box] = nil
+    -- this will become
+    -- textexts[slot] = texbox[box]
+    -- unsetbox(box)
+end
+
+function metapost.gettext(box,slot)
+    texbox[box] = copy_list(textexts[slot])
+--  textexts[slot] = nil -- no, pictures can be placed several times
 end
 
 function metapost.specials.tf(specification,object)
 --~ print("setting", metapost.textext_current)
     local n, str = match(specification,"^(%d+):(.+)$")
     if n and str then
-        if metapost.textext_current < metapost.last_box then
-            metapost.textext_current = metapost.first_box + n - 1
-        end
+        n = tonumber(n)
+     -- if metapost.textext_current < metapost.last_box then
+     --     metapost.textext_current = metapost.first_box + n - 1
+     -- end
         if trace_textexts then
-            logs.report("metapost","first pass: order %s, box %s",n,metapost.textext_current)
+         -- report_mplib("first pass: order %s, box %s",n,metapost.textext_current)
+            report_mplib("first pass: order %s",n)
         end
-        sprint(ctxcatcodes,format("\\MPLIBsettext{%s}{%s}",metapost.textext_current,str))
+     -- sprint(ctxcatcodes,format("\\MPLIBsettext{%s}{%s}",metapost.textext_current,str))
+        sprint(ctxcatcodes,format("\\MPLIBsettext{%s}{%s}",n,str))
         metapost.multipass = true
     end
     return { }, nil, nil, nil
@@ -456,8 +488,9 @@ function metapost.specials.ts(specification,object,result,flusher)
     -- print("getting", metapost.textext_current)
     local n, str = match(specification,"^(%d+):(.+)$")
     if n and str then
+        n = tonumber(n)
         if trace_textexts then
-            logs.report("metapost","second pass: order %s, box %s",n,metapost.textext_current)
+            report_mplib("second pass: order %s",n)
         end
         local op = object.path
         local first, second, fourth = op[1], op[2], op[4]
@@ -469,17 +502,24 @@ function metapost.specials.ts(specification,object,result,flusher)
         if not trace_textexts then
             object.path = nil
         end
-        local before = function() -- no need for function
+        local before = function() -- no need for before function (just do it directly)
         --~ flusher.flushfigure(result)
         --~ sprint(ctxcatcodes,format("\\MPLIBgettext{%f}{%f}{%f}{%f}{%f}{%f}{%s}",sx,rx,ry,sy,tx,ty,metapost.textext_current))
         --~ result = { }
             result[#result+1] = format("q %f %f %f %f %f %f cm", sx,rx,ry,sy,tx,ty)
             flusher.flushfigure(result)
-            if metapost.textext_current < metapost.last_box then
-                metapost.textext_current = metapost.first_box + n - 1
+            -- if metapost.textext_current < metapost.last_box then
+            --     metapost.textext_current = metapost.first_box + n - 1
+            --  end
+         -- local b = metapost.textext_current
+         -- local box = texbox[b]
+            local box = textexts[n]
+            if box then
+             -- sprint(ctxcatcodes,format("\\MPLIBgettextscaled{%s}{%s}{%s}",b,metapost.sxsy(box.width,box.height,box.depth)))
+                sprint(ctxcatcodes,format("\\MPLIBgettextscaled{%s}{%s}{%s}",n,metapost.sxsy(box.width,box.height,box.depth)))
+            else
+                -- error
             end
-            local b = metapost.textext_current
-            sprint(ctxcatcodes,format("\\MPLIBgettextscaled{%s}{%s}{%s}",b, metapost.sxsy(tex.wd[b],tex.ht[b],tex.dp[b])))
             result = { "Q" }
             return object, result
         end
@@ -646,7 +686,7 @@ do
 
     local function convert(str)
         found = true
-        return "textext(\"" .. str .. "\")"
+        return "rawtextext(\"" .. str .. "\")" -- centered
     end
     local function ditto(str)
         return "\" & ditto & \""
@@ -658,20 +698,33 @@ do
         forced = true
     end
 
+    local texmess   = (dquote/ditto + (1 - etex))^0
+
+    local function ignore(s)
+        report_mplib("ignoring verbatim tex: %s",s)
+        return ""
+    end
+
     local parser = P {
-        [1] = Cs((V(2)/register + V(3)/convert + V(4)/force + 1)^0),
+        [1] = Cs((V(2)/register + V(4)/ignore + V(3)/convert + V(5)/force + 1)^0),
         [2] = ttex + gtex,
-        [3] = (btex + vtex) * spacing * Cs((dquote/ditto + (1 - etex))^0) * etex,
-        [4] = multipass, -- experimental, only for testing
+        [3] = btex * spacing * Cs(texmess) * etex,
+        [4] = vtex * spacing * Cs(texmess) * etex,
+        [5] = multipass, -- experimental, only for testing
     }
 
     -- currently a a one-liner produces less code
 
-    local parser = Cs(((ttex + gtex)/register + ((btex + vtex) * spacing * Cs((dquote/ditto + (1 - etex))^0) * etex)/convert + 1)^0)
+    local parser = Cs((
+        (ttex + gtex)/register
+      + (btex * spacing * Cs(texmess) * etex)/convert
+      + (vtex * spacing * Cs(texmess) * etex)/ignore
+      + 1
+    )^0)
 
     function metapost.check_texts(str)
         found, forced = false, false
-        return parser:match(str), found, forced
+        return lpegmatch(parser,str), found, forced
     end
 
 end
@@ -689,19 +742,29 @@ function metapost.sxsy(wd,ht,dp) -- helper for text
     return (wd ~= 0 and factor/wd) or 0, (hd ~= 0 and factor/hd) or 0
 end
 
+local no_trial_run       = "_trial_run_ := false ;"
+local do_trial_run       = "if unknown _trial_run_ : boolean _trial_run_ fi ; _trial_run_ := true ;"
+local text_data_template = "_tt_w_[%i]:=%f;_tt_h_[%i]:=%f;_tt_d_[%i]:=%f;"
+local do_begin_fig       = "; beginfig(1); "
+local do_end_fig         = "; endfig ;"
+local do_safeguard       = ";"
+
 function metapost.text_texts_data()
     local t, n = { }, 0
-    for i = metapost.first_box, metapost.last_box do
-        n = n + 1
+--~     for i = metapost.first_box, metapost.last_box do
+--~         n = n + 1
+--~         local box = texbox[i]
+    for n, box in next, textexts do
         if trace_textexts then
-            logs.report("metapost","passed data: order %s, box %s",n,i)
+            report_mplib("passed data: order %s",n)
         end
-        if tex.box[i] then
-            t[#t+1] = format("_tt_w_[%i]:=%f;_tt_h_[%i]:=%f;_tt_d_[%i]:=%f;", n,tex.wd[i]/factor, n,tex.ht[i]/factor, n,tex.dp[i]/factor)
+        if box then
+            t[#t+1] = format(text_data_template,n,box.width/factor,n,box.height/factor,n,box.depth/factor)
         else
             break
         end
     end
+--~     print(table.serialize(t))
     return t
 end
 
@@ -711,31 +774,33 @@ metapost.intermediate.needed  = false
 
 metapost.method = 1 -- 1:dumb 2:clever
 
-function metapost.graphic_base_pass(mpsformat,str,preamble,askedfig)
+function metapost.graphic_base_pass(mpsformat,str,initializations,preamble,askedfig)
     local nofig = (askedfig and "") or false
     local done_1, done_2, forced_1, forced_2
     str, done_1, forced_1 = metapost.check_texts(str)
-    if preamble then
-        preamble, done_2, forced_2 = metapost.check_texts(preamble)
-    else
+    if not preamble or preamble == "" then
         preamble, done_2, forced_2 = "", false, false
+    else
+        preamble, done_2, forced_2 = metapost.check_texts(preamble)
     end
-    metapost.textext_current = metapost.first_box
+ -- metapost.textext_current = metapost.first_box
     metapost.intermediate.needed  = false
     metapost.multipass = false -- no needed here
-    current_format, current_graphic = mpsformat, str
+    current_format, current_graphic, current_initializations = mpsformat, str, initializations or ""
     if metapost.method == 1 or (metapost.method == 2 and (done_1 or done_2)) then
      -- first true means: trialrun, second true means: avoid extra run if no multipass
         local flushed = metapost.process(mpsformat, {
             preamble,
-            nofig or "beginfig(1); ",
-            "if unknown _trial_run_ : boolean _trial_run_ fi ; _trial_run_ := true ;",
-            str,
-            nofig or "endfig ;"
+            nofig or do_begin_fig,
+            do_trial_run,
+            current_initializations,
+            do_safeguard,
+            current_graphic,
+            nofig or do_end_fig
      -- }, true, nil, true )
         }, true, nil, not (forced_1 or forced_2), false, askedfig)
         if metapost.intermediate.needed then
-            for _, action in pairs(metapost.intermediate.actions) do
+            for _, action in next, metapost.intermediate.actions do
                 action()
             end
         end
@@ -746,11 +811,13 @@ function metapost.graphic_base_pass(mpsformat,str,preamble,askedfig)
         end
     else
         metapost.process(mpsformat, {
-            preamble or "",
-            nofig or "beginfig(1); ",
-            "_trial_run_ := false ;",
-            str,
-            nofig or "endfig ;"
+            preamble,
+            nofig or do_begin_fig,
+            no_trial_run,
+            current_initializations,
+            do_safeguard,
+            current_graphic,
+            nofig or do_end_fig
         }, false, nil, false, false, askedfig )
     end
     -- here we could free the textext boxes
@@ -759,34 +826,18 @@ end
 
 function metapost.graphic_extra_pass(askedfig)
     local nofig = (askedfig and "") or false
-    metapost.textext_current = metapost.first_box
+ -- metapost.textext_current = metapost.first_box
     metapost.process(current_format, {
-        nofig or "beginfig(1); ",
-        "_trial_run_ := false ;",
+        nofig or do_begin_fig,
+        no_trial_run,
         concat(metapost.text_texts_data()," ;\n"),
+        current_initializations,
+        do_safeguard,
         current_graphic,
-        nofig or "endfig ;"
+        nofig or do_end_fig
     }, false, nil, false, true, askedfig )
 end
 
-function metapost.getclippath(data)
-    local mpx = metapost.format("metafun")
-    if mpx and data then
-        starttiming(metapost)
-        starttiming(metapost.exectime)
-        local result = mpx:execute(format("beginfig(1);%s;endfig;",data))
-        stoptiming(metapost.exectime)
-        if result.status > 0 then
-            print("error", result.status, result.error or result.term or result.log)
-            result = ""
-        else
-            result = metapost.filterclippath(result)
-        end
-        stoptiming(metapost)
-        sprint(result)
-    end
-end
-
 metapost.tex = metapost.tex or { }
 
 do -- only used in graphictexts
