@@ -31,15 +31,24 @@ nears zero.</p>
 
 local trace_graphics = false  trackers.register("metapost.graphics", function(v) trace_graphics = v end)
 
-local format = string.format
+local report_mplib = logs.new("mplib")
+
+local texerrormessage = tracers.texerrormessage
+
+local format, gsub, match = string.format, string.gsub, string.match
 
 local starttiming, stoptiming = statistics.starttiming, statistics.stoptiming
 
-
 metapost = metapost or { }
 
-metapost.showlog = false
-metapost.lastlog = ""
+metapost.showlog      = false
+metapost.lastlog      = ""
+metapost.texerrors    = false
+metapost.exectime     = metapost.exectime or { } -- hack
+
+local mplibone = tonumber(mplib.version()) <= 1.50
+
+directives.register("mplib.texerrors", function(v) metapost.texerrors = v end)
 
 function metapost.resetlastlog()
     metapost.lastlog = ""
@@ -57,59 +66,163 @@ end
 
 metapost.finder = finder
 
-metapost.parameters = {
-    hash_size = 100000,
-    main_memory = 2000000,
-    max_in_open = 50,
-    param_size = 100000,
-}
-
-metapost.exectime = metapost.exectime or { } -- hack
-
-local preamble = [[
-boolean mplib; string mp_parent_version;
-mplib := true;
-mp_parent_version := "%s";
-input %s ; dump ;
-]]
-
-function metapost.make(name, target, version)
-    starttiming(mplib)
-    target = file.replacesuffix(target or name, "mem")
-    local mpx = mplib.new ( table.merged (
-        metapost.parameters,
-        {
-            ini_version = true,
-            find_file = finder,
-            job_name = file.removesuffix(target),
-        }
-    ) )
-    if mpx then
-        starttiming(metapost.exectime)
-        local result = mpx:execute(format(preamble,version or "unknown",name))
-        stoptiming(metapost.exectime)
-        mpx:finish()
+function metapost.reporterror(result)
+    if not result then
+        report_mplib("mp error: no result object returned")
+    elseif result.status > 0 then
+        local t, e, l = result.term, result.error, result.log
+        if t and t ~= "" then
+            (metapost.texerrors and texerrormessage or report_mplib)("mp terminal: %s",t)
+        end
+        if e then
+            (metapost.texerrors and texerrormessage or report_mplib)("mp error: %s",(e=="" and "?") or e)
+        end
+        if not t and not e and l then
+            metapost.lastlog = metapost.lastlog .. "\n" .. l
+            report_mplib("mp log: %s",l)
+        else
+            report_mplib("mp error: unknown, no error, terminal or log messages")
+        end
+    else
+        return false
     end
-    stoptiming(mplib)
+    return true
 end
 
-function metapost.load(name)
-    starttiming(mplib)
-    local mpx = mplib.new ( table.merged (
-        metapost.parameters,
-        {
-            ini_version = false,
-            mem_name = file.replacesuffix(name,"mem"),
-            find_file = finder,
-         -- job_name = "mplib",
-        }
-    ) )
-    local result
-    if not mpx then
-        result = { status = 99, error = "out of memory"}
+if mplibone then
+
+    local preamble = [[
+        boolean mplib ; mplib := true ;
+        string mp_parent_version ; mp_parent_version := "%s" ;
+        input %s ; dump ;
+    ]]
+
+    metapost.parameters = {
+        hash_size = 100000,
+        main_memory = 4000000,
+        max_in_open = 50,
+        param_size = 100000,
+    }
+
+    function metapost.make(name, target, version)
+        starttiming(mplib)
+        target = file.replacesuffix(target or name, "mem")
+        local mpx = mplib.new ( table.merged (
+            metapost.parameters,
+            {
+                ini_version = true,
+                find_file = finder,
+                job_name = file.removesuffix(target),
+            }
+        ) )
+        if mpx then
+            starttiming(metapost.exectime)
+            local result = mpx:execute(format(preamble,version or "unknown",name))
+            stoptiming(metapost.exectime)
+            mpx:finish()
+        end
+        stoptiming(mplib)
     end
-    stoptiming(mplib)
-    return mpx, result
+
+    function metapost.load(name)
+        starttiming(mplib)
+        local mpx = mplib.new ( table.merged (
+            metapost.parameters,
+            {
+                ini_version = false,
+                mem_name = file.replacesuffix(name,"mem"),
+                find_file = finder,
+             -- job_name = "mplib",
+            }
+        ) )
+        local result
+        if not mpx then
+            result = { status = 99, error = "out of memory"}
+        end
+        stoptiming(mplib)
+        return mpx, result
+    end
+
+    function metapost.checkformat(mpsinput)
+        local mpsversion = environment.version or "unset version"
+        local mpsinput   = file.addsuffix(mpsinput or "metafun", "mp")
+        local mpsformat  = file.removesuffix(file.basename(texconfig.formatname or (tex and tex.formatname) or mpsinput))
+        local mpsbase    = file.removesuffix(file.basename(mpsinput))
+        if mpsbase ~= mpsformat then
+            mpsformat = mpsformat .. "-" .. mpsbase
+        end
+        mpsformat = file.addsuffix(mpsformat, "mem")
+        local mpsformatfullname = caches.getfirstreadablefile(mpsformat,"formats") or ""
+        if mpsformatfullname ~= "" then
+            commands.writestatus("mplib","loading '%s' from '%s'", mpsinput, mpsformatfullname)
+            local mpx, result = metapost.load(mpsformatfullname)
+            if mpx then
+                local result = mpx:execute("show mp_parent_version ;")
+                if not result.log then
+                    metapost.reporterror(result)
+                else
+                    local version = match(result.log,">> *(.-)[\n\r]") or "unknown"
+                    version = gsub(version,"[\'\"]","")
+                    if version ~= mpsversion then
+                        commands.writestatus("mplib","version mismatch: %s <> %s", version or "unknown", mpsversion)
+                    else
+                        return mpx
+                    end
+                end
+            else
+                commands.writestatus("mplib","error in loading '%s' from '%s'", mpsinput, mpsformatfullname)
+                metapost.reporterror(result)
+            end
+        end
+        local mpsformatfullname = caches.setfirstwritablefile(mpsformat,"formats")
+        commands.writestatus("mplib","making '%s' into '%s'", mpsinput, mpsformatfullname)
+        metapost.make(mpsinput,mpsformatfullname,mpsversion) -- somehow return ... fails here
+        if lfs.isfile(mpsformatfullname) then
+            commands.writestatus("mplib","loading '%s' from '%s'", mpsinput, mpsformatfullname)
+            return metapost.load(mpsformatfullname)
+        else
+            commands.writestatus("mplib","problems with '%s' from '%s'", mpsinput, mpsformatfullname)
+        end
+    end
+
+else
+
+    local preamble = [[
+        boolean mplib ; mplib := true ;
+        let dump = endinput ;
+        input %s ;
+    ]]
+
+    function metapost.load(name)
+        starttiming(mplib)
+        local mpx = mplib.new {
+            ini_version = true,
+            find_file = finder,
+        }
+        local result
+        if not mpx then
+            result = { status = 99, error = "out of memory"}
+        else
+            result = mpx:execute(format(preamble, file.replacesuffix(name,"mp")))
+        end
+        stoptiming(mplib)
+        metapost.reporterror(result)
+        return mpx, result
+    end
+
+    function metapost.checkformat(mpsinput)
+        local mpsversion = environment.version or "unset version"
+        local mpsinput   = file.addsuffix(mpsinput or "metafun", "mp")
+        commands.writestatus("mplib","loading '%s' (experimental metapost version two)",mpsinput)
+        local mpx, result = metapost.load(mpsinput)
+        if mpx then
+            return mpx
+        else
+            commands.writestatus("mplib","error in loading '%s'",mpsinput)
+            metapost.reporterror(result)
+        end
+    end
+
 end
 
 function metapost.unload(mpx)
@@ -120,73 +233,6 @@ function metapost.unload(mpx)
     stoptiming(mplib)
 end
 
-function metapost.reporterror(result)
-    if not result then
-        metapost.report("mp error: no result object returned")
-    elseif result.status > 0 then
-        local t, e, l = result.term, result.error, result.log
-        if t and t ~= "" then
-            metapost.report("mp terminal: %s",t)
-        end
-        if e then
-            metapost.report("mp error: %s",(e=="" and "?") or e)
-        end
-        if not t and not e and l then
-            metapost.lastlog = metapost.lastlog .. "\n" .. l
-            metapost.report("mp log: %s",l)
-        else
-            metapost.report("mp error: unknown, no error, terminal or log messages")
-        end
-    else
-        return false
-    end
-    return true
-end
-
-function metapost.checkformat(mpsinput, mpsformat, dirname)
-    mpsinput  = file.addsuffix(mpsinput or "metafun", "mp")
-    mpsformat = file.removesuffix(file.basename(mpsformat or texconfig.formatname or (tex and tex.formatname) or mpsinput))
-    local mpsbase = file.removesuffix(file.basename(mpsinput))
-    if mpsbase ~= mpsformat then
-        mpsformat = mpsformat .. "-" .. mpsbase
-    end
-    mpsformat = file.addsuffix(mpsformat, "mem")
-    local pth = dirname or file.dirname(texconfig.formatname or "")
-    if pth ~= "" then
-        mpsformat = file.join(pth,mpsformat)
-    end
-    local the_version = environment.version or "unset version"
-    if lfs.isfile(mpsformat) then
-        commands.writestatus("mplib","loading '%s' from '%s'", mpsinput, mpsformat)
-        local mpx, result = metapost.load(mpsformat)
-        if mpx then
-            local result = mpx:execute("show mp_parent_version ;")
-            if not result.log then
-                metapost.reporterror(result)
-            else
-                local version = result.log:match(">> *(.-)[\n\r]") or "unknown"
-                version = version:gsub("[\'\"]","")
-                if version ~= the_version then
-                    commands.writestatus("mplib","version mismatch: %s <> %s", version or "unknown", the_version)
-                else
-                    return mpx
-                end
-            end
-        else
-            commands.writestatus("mplib","error in loading '%s' from '%s'", mpsinput, mpsformat)
-            metapost.reporterror(result)
-        end
-    end
-    commands.writestatus("mplib","making '%s' into '%s'", mpsinput, mpsformat)
-    metapost.make(mpsinput,mpsformat,the_version) -- somehow return ... fails here
-    if lfs.isfile(mpsformat) then
-        commands.writestatus("mplib","loading '%s' from '%s'", mpsinput, mpsformat)
-        return metapost.load(mpsformat)
-    else
-        commands.writestatus("mplib","problems with '%s' from '%s'", mpsinput, mpsformat)
-    end
-end
-
 local mpxformats = { }
 
 function metapost.format(instance,name)
@@ -209,7 +255,7 @@ function metapost.reset(mpx)
             mpxformats[mpx] = nil
         end
     else
-        for name, instance in pairs(mpxformats) do
+        for name, instance in next, mpxformats do
             if instance == mpx then
                 mpx:finish()
                 mpxformats[name] = nil
@@ -259,7 +305,7 @@ function metapost.process(mpx, data, trialrun, flusher, multipass, isextrapass,
                             local str = (result.term ~= "" and result.term) or "no terminal output"
                             if not str:is_empty() then
                                 metapost.lastlog = metapost.lastlog .. "\n" .. str
-                                metapost.report("mp log: %s",str)
+                                report_mplib("mp log: %s",str)
                             end
                         end
                         if result.fig then
@@ -267,7 +313,7 @@ function metapost.process(mpx, data, trialrun, flusher, multipass, isextrapass,
                         end
                     end
                 else
-                    metapost.report("mp error: invalid graphic component %s",i)
+                    report_mplib("mp error: invalid graphic component %s",i)
                 end
             end
        else
@@ -285,13 +331,13 @@ function metapost.process(mpx, data, trialrun, flusher, multipass, isextrapass,
             end
             -- todo: error message
             if not result then
-                metapost.report("mp error: no result object returned")
+                report_mplib("mp error: no result object returned")
             elseif result.status > 0 then
-                metapost.report("mp error: %s",(result.term or "no-term") .. "\n" .. (result.error or "no-error"))
+                report_mplib("mp error: %s",(result.term or "no-term") .. "\n" .. (result.error or "no-error"))
             else
                 if metapost.showlog then
                     metapost.lastlog = metapost.lastlog .. "\n" .. result.term
-                    metapost.report("mp info: %s",result.term or "no-term")
+                    report_mplib("mp info: %s",result.term or "no-term")
                 end
                 if result.fig then
                     converted = metapost.convert(result, trialrun, flusher, multipass, askedfig)
@@ -309,11 +355,7 @@ function metapost.process(mpx, data, trialrun, flusher, multipass, isextrapass,
 end
 
 function metapost.convert()
-    metapost.report('mp warning: no converter set')
-end
-
-function metapost.report(...)
-    logs.report("mplib",...)
+    report_mplib('mp warning: no converter set')
 end
 
 -- handy
@@ -327,7 +369,7 @@ function metapost.directrun(formatname,filename,outputformat,astable,mpdata)
     if not data then
         logs.simple("unknown file '%s'",filename or "?")
     else
-        local mpx = metapost.checkformat(formatname,formatname,caches.setpath("formats"))
+        local mpx = metapost.checkformat(formatname)
         if not mpx then
             logs.simple("unknown format '%s'",formatname or "?")
         else
@@ -348,7 +390,8 @@ function metapost.directrun(formatname,filename,outputformat,astable,mpdata)
                     if astable then
                         local result = { }
                         logs.simple("storing %s figures in table",#sorted)
-                        for k, v in ipairs(sorted) do
+                        for k=1,#sorted do
+                            local v = sorted[k]
                             if outputformat == "mps" then
                                 result[v] = figures[v]:postscript()
                             else
@@ -358,7 +401,8 @@ function metapost.directrun(formatname,filename,outputformat,astable,mpdata)
                         return result
                     else
                         local basename = file.removesuffix(file.basename(filename))
-                        for k, v in ipairs(sorted) do
+                        for k=1,#sorted do
+                            local v = sorted[k]
                             local output
                             if outputformat == "mps" then
                                 output = figures[v]:postscript()
