@@ -14,6 +14,7 @@ local texsprint = tex.sprint
 local ctxcatcodes = tex.ctxcatcodes
 
 local glyph   = node.id('glyph')
+local disc    = node.id('disc')
 local glue    = node.id('glue')
 local rule    = node.id('rule')
 local whatsit = node.id('whatsit')
@@ -75,7 +76,7 @@ end
 
 local function process_attribute(head,plugin) -- head,attribute,enabled,initializer,resolver,processor,finalizer
     local namespace = plugin.namespace
-    if namespace.enabled then
+    if namespace.enabled ~= false then -- this test will go away
         starttiming(attributes)
         local done, used, ok = false, nil, false
         local attribute = namespace.attribute or numbers[plugin.name] -- todo: plugin.attribute
@@ -170,7 +171,7 @@ end
 local insert_node_before = node.insert_before
 local insert_node_after  = node.insert_after
 
-local nsdata, nsdone, nsforced, nsselector, nstrigger
+local nsdata, nsnone, nslistwise, nsforced, nsselector, nstrigger
 local current, current_selector, done = 0, 0, false -- nb, stack has a local current !
 
 function states.initialize(namespace,attribute,head)
@@ -202,7 +203,8 @@ local function process(namespace,attribute,head,inheritance,default) -- one attr
         local id = stack.id
         -- we need to deal with literals too (reset as well as oval)
         -- if id == glyph or (id == whatsit and stack.subtype == 8) or (id == rule and stack.width ~= 0) or (id == glue and stack.leader) then -- or disc
-        if id == glyph or (id == rule and stack.width ~= 0) or (id == glue and stack.leader) then -- or disc
+        if id == glyph -- or id == disc
+                or (id == rule and stack.width ~= 0) or (id == glue and stack.leader) then -- or disc
             local c = has_attribute(stack,attribute)
             if c then
                 if default and c == inheritance then
@@ -287,7 +289,8 @@ local function selective(namespace,attribute,head,inheritance,default) -- two at
         local id = stack.id
         -- we need to deal with literals too (reset as well as oval)
         -- if id == glyph or (id == whatsit and stack.subtype == 8) or (id == rule and stack.width ~= 0) or (id == glue and stack.leader) then -- or disc
-        if id == glyph or (id == rule and stack.width ~= 0) or (id == glue and stack.leader) then -- or disc
+        if id == glyph -- or id == disc
+                or (id == rule and stack.width ~= 0) or (id == glue and stack.leader) then -- or disc
             local c = has_attribute(stack,attribute)
             if c then
                 if default and c == inheritance then
@@ -438,7 +441,5 @@ states.stacked = stacked
 -- -- --
 
 statistics.register("attribute processing time", function()
-    if statistics.elapsedindeed(attributes) then
-        return format("%s seconds (front- and backend)",statistics.elapsedtime(attributes))
-    end
+    return statistics.elapsedseconds(attributes,"front- and backend")
 end)
