@@ -26,7 +26,9 @@ local kind_to_number = scripts.kind_to_number
 local number_to_kind = scripts.number_to_kind
 local hash           = scripts.hash
 local cjk            = scripts.cjk
-local fontdata       = fonts.ids
+
+local fontdata       = fonts.identifiers
+local quaddata       = fonts.quads
 
 -- raggedleft is controlled by leftskip and we might end up with a situation where
 -- the intercharacter spacing interferes with this; the solution is to patch the
@@ -38,8 +40,7 @@ local inter_char_hangul_penalty = 0
 
 local function set_parameters(font,data)
     -- beware: parameters can be nil in e.g. punk variants
-    local parameters = fontdata[font].parameters
-    local quad = (parameters and parameters.quad or parameters[6]) or 0
+    local quad = quaddata[font]
     inter_char_half_shrink    = data.inter_char_half_shrink_factor * quad
     inter_char_stretch        = data.inter_char_stretch_factor * quad
     inter_char_hangul_penalty = data.inter_char_hangul_penalty
@@ -326,7 +327,7 @@ local function process(head,first,last)
                     if action then
                         local font = first.font
                         if font ~= lastfont then
-                            lastfont, done = font, true
+                            lastfont = font
                             set_parameters(font,dataset)
                         end
                         action(head,first)
@@ -358,7 +359,7 @@ local function process(head,first,last)
                     previous = "start"
                 end
             end
-            if upcoming == stop then
+            if upcoming == last then -- was stop
                 break
             else
                 first = upcoming
@@ -486,7 +487,7 @@ local injectors = { -- [previous] [current]
     full_width_close = {
         jami_initial     = nobreak_shrink_break_stretch,
         korean           = nobreak_shrink_break_stretch,
-        chinese          = nobreak_shrink_break_stretch,
+        chinese          = stretch_break, -- nobreak_shrink_break_stretch,
         half_width_open  = nobreak_shrink_break_stretch_nobreak_autoshrink,
         half_width_close = nobreak_shrink_nobreak_stretch,
         full_width_open  = nobreak_shrink_break_stretch_nobreak_shrink,
@@ -498,7 +499,7 @@ local injectors = { -- [previous] [current]
     half_width_close = {
         jami_initial     = nobreak_shrink_break_stretch,
         korean           = nobreak_autoshrink_break_stretch,
-        chinese          = nobreak_autoshrink_break_stretch,
+        chinese          = stretch_break, -- nobreak_autoshrink_break_stretch,
         half_width_open  = nobreak_autoshrink_break_stretch_nobreak_autoshrink,
         half_width_close = nobreak_autoshrink_nobreak_stretch,
         full_width_open  = nobreak_autoshrink_break_stretch_nobreak_shrink,
@@ -529,7 +530,7 @@ local function process(head,first,last)
                     if action then
                         local font = first.font
                         if font ~= lastfont then
-                            lastfont, done = font, true
+                            lastfont = font
                             set_parameters(font,dataset)
                         end
                         action(head,first)
@@ -561,7 +562,7 @@ local function process(head,first,last)
                     previous = "start"
                 end
             end
-            if upcoming == stop then
+            if upcoming == last then -- was stop
                 break
             else
                 first = upcoming
