@@ -6,6 +6,8 @@ if not modules then modules = { } end modules ['spac-ver'] = {
     license   = "see context related readme files"
 }
 
+-- we also need to call the spacer for inserts!
+
 -- this code dates from the beginning and is kind of experimental; it
 -- will be optimized and improved soon
 --
@@ -15,10 +17,14 @@ if not modules then modules = { } end modules ['spac-ver'] = {
 -- still needed tex -> lua -> tex > lua chain (needed because we can have
 -- expandable settings at the tex end
 
+-- todo: strip baselineskip around display math
+
 local next, type, tonumber = next, type, tonumber
 local format, gmatch, concat, match = string.format, string.gmatch, table.concat, string.match
 local ceil, floor, max, min, round = math.ceil, math.floor, math.max, math.min, math.round
-local texsprint, texlists, texdimen, texbox, texht, texdp = tex.sprint, tex.lists, tex.dimen, tex.box, tex.ht, tex.dp
+local texsprint, texlists, texdimen, texbox = tex.sprint, tex.lists, tex.dimen, tex.box
+local lpegmatch = lpeg.match
+local unpack = unpack or table.unpack
 
 local ctxcatcodes = tex.ctxcatcodes
 local variables = interfaces.variables
@@ -32,6 +38,11 @@ local trace_page_vspacing    = false  trackers.register("nodes.page_vspacing",
 local trace_collect_vspacing = false  trackers.register("nodes.collect_vspacing", function(v) trace_collect_vspacing = v end)
 local trace_vspacing         = false  trackers.register("nodes.vspacing",         function(v) trace_vspacing         = v end)
 local trace_vsnapping        = false  trackers.register("nodes.vsnapping",        function(v) trace_vsnapping        = v end)
+local trace_vpacking         = false  trackers.register("nodes.vpacking",         function(v) trace_vpacking         = v end)
+
+local report_vspacing  = logs.new("vspacing")
+local report_collapser = logs.new("collapser")
+local report_snapper   = logs.new("snapper")
 
 local skip_category = attributes.private('skip-category')
 local skip_penalty  = attributes.private('skip-penalty')
@@ -59,6 +70,7 @@ local count_nodes        = nodes.count
 local node_ids_to_string = nodes.ids_to_string
 local hpack_node         = node.hpack
 local vpack_node         = node.vpack
+local writable_spec      = nodes.writable_spec
 
 local glyph   = node.id("glyph")
 local penalty = node.id("penalty")
@@ -75,7 +87,7 @@ vspacing.data.snapmethods = vspacing.data.snapmethods or { }
 
 storage.register("vspacing/data/snapmethods", vspacing.data.snapmethods, "vspacing.data.snapmethods")
 
-local snapmethods = vspacing.data.snapmethods
+local snapmethods = vspacing.data.snapmethods --maybe some older code can go
 
 local default = {
     maxheight = true,
@@ -91,28 +103,40 @@ local fractions = {
     top       = "tlines",    bottom    = "blines",
 }
 
+local values = {
+    offset = "offset"
+}
+
 local colonsplitter = lpeg.splitat(":")
 
-function interfaces.listtohash(str)
+local function listtohash(str)
     local t = { }
     for s in gmatch(str,"[^, ]+") do
-        local key, fraction = colonsplitter:match(s)
+        local key, detail = lpegmatch(colonsplitter,s)
         local v = variables[key]
         if v then
             t[v] = true
-            if fraction then
+            if detail then
                 local k = fractions[key]
                 if k then
-                    fraction = tonumber("0" .. fraction)
-                    if fraction then
-                        t[k] = fraction
+                    detail = tonumber("0" .. detail)
+                    if detail then
+                        t[k] = detail
+                    end
+                else
+                    k = values[key]
+                    if k then
+                        detail = string.todimen(detail)
+                        if detail then
+                            t[k] = detail
+                        end
                     end
                 end
             end
         else
-            fraction = tonumber("0" .. key)
-            if fraction then
-                t.hfraction, t.dfraction = fraction, fraction
+            detail = tonumber("0" .. key)
+            if detail then
+                t.hfraction, t.dfraction = detail, detail
             end
         end
     end
@@ -127,21 +151,43 @@ end
 
 function vspacing.define_snap_method(name,method)
     local n = #snapmethods + 1
-    local t = interfaces.listtohash(method)
+    local t = listtohash(method)
     snapmethods[n] = t
     t.name, t.specification = name, method
     tex.write(n)
 end
 
+--~ local rule_id  = node.id("rule")
+--~ local vlist_id = node.id("vlist")
+--~ function nodes.makevtop(n)
+--~     if n.id == vlist_id then
+--~         local list = n.list
+--~         local height = (list and list.id <= rule_id and list.height) or 0
+--~         n.depth = n.depth - height + n.height
+--~         n.height = height
+--~     end
+--~ end
+
 local function snap_hlist(current,method,height,depth) -- method.strut is default
     local snapht, snapdp
+--~ print(table.serialize(method))
     if method["local"] then
+        -- snapping is done immediately here
         snapht, snapdp = texdimen.bodyfontstrutheight, texdimen.bodyfontstrutdepth
-    else
+    elseif method["global"] then
         snapht, snapdp = texdimen.globalbodyfontstrutheight, texdimen.globalbodyfontstrutdepth
+    else
+        -- maybe autolocal
+        -- snapping might happen later in the otr
+        snapht, snapdp = texdimen.globalbodyfontstrutheight, texdimen.globalbodyfontstrutdepth
+        local lsnapht, lsnapdp = texdimen.bodyfontstrutheight, texdimen.bodyfontstrutdepth
+        if snapht ~= lsnapht and snapdp ~= lsnapdp then
+            snapht, snapdp = lsnapht, lsnapdp
+        end
     end
     local h, d = height or current.height, depth or current.depth
     local hr, dr, ch, cd = method.hfraction or 1, method.dfraction or 1, h, d
+    local tlines, blines = method.tlines or 1, method.blines or 1
     local done, plusht, plusdp = false, snapht, snapdp
     local snaphtdp = snapht + snapdp
     if method.none then
@@ -153,6 +199,7 @@ local function snap_hlist(current,method,height,depth) -- method.strut is defaul
     if method.line then
         plusht, plusdp = plusht + snaphtdp, plusdp + snaphtdp
     end
+
     if method.first then
         if current.id == vlist then
             local list, lh, ld = current.list
@@ -186,6 +233,7 @@ local function snap_hlist(current,method,height,depth) -- method.strut is defaul
             end
         end
     end
+
     if done then
         -- first or last
     elseif method.minheight then
@@ -205,10 +253,18 @@ local function snap_hlist(current,method,height,depth) -- method.strut is defaul
         cd = plusdp
     end
     if method.top then
-        ch = ch + (method.tlines or 1) * snaphtdp
+        ch = ch + tlines * snaphtdp
     end
     if method.bottom then
-        cd = cd + (method.blines or 1) * snaphtdp
+        cd = cd + blines * snaphtdp
+    end
+
+    local offset = method.offset
+    if offset then
+        -- we need to set the attr
+        local shifted = vpack_node(current.list)
+        shifted.shift = offset
+        current.list = shifted
     end
     if not height then
         current.height = ch
@@ -216,16 +272,28 @@ local function snap_hlist(current,method,height,depth) -- method.strut is defaul
     if not depth then
         current.depth = cd
     end
+ -- set_attribute(current,snap_method,0)
     return h, d, ch, cd, (ch+cd)/snaphtdp
 end
 
+--~ local function snap_topskip(current,method)
+--~     local spec = current.spec
+--~     local w = spec.width
+--~     local wd = w
+--~     if spec then
+--~         wd = 0
+--~         spec = writable_spec(current)
+--~         spec.width = wd
+--~     end
+--~     return w, wd
+--~ end
+
 local function snap_topskip(current,method)
     local spec = current.spec
     local w = spec.width
     local wd = w
-    if spec then
-        wd = 0
-        spec.width = wd
+    if spec.writable then
+        spec.width, wd = 0, 0
     end
     return w, wd
 end
@@ -274,7 +342,7 @@ storage.register("vspacing/data/skip", vspacing.data.skip, "vspacing.data.skip")
 do -- todo: interface.variables
 
     local function logger(c,...)
-        logs.report("vspacing",concat {...})
+        report_vspacing(concat {...})
         texsprint(c,...)
     end
 
@@ -295,9 +363,9 @@ do -- todo: interface.variables
 
     local function analyse(str,oldcategory,texsprint) -- we could use shorter names
         for s in gmatch(str,"([^ ,]+)") do
-            local amount, keyword, detail = splitter:match(s)
+            local amount, keyword, detail = lpegmatch(splitter,s)
             if not keyword then
-                logs.report("vspacing","unknown directive: %s",s)
+                report_vspacing("unknown directive: %s",s)
             else
                 local mk = map[keyword]
                 if mk then
@@ -373,23 +441,19 @@ local trace_list, tracing_info, before, after = { }, false, "", ""
 
 local function glue_to_string(glue)
     local spec = glue.spec
-    if spec then
-        local t = { }
-        t[#t+1] = aux.strip_zeros(number.topoints(spec.width))
-        if spec.stretch_order and spec.stretch_order ~= 0 then
-            t[#t+1] = format("plus -%sfi%s",spec.stretch/65536,string.rep("l",math.abs(spec.stretch_order)-1))
-        elseif spec.stretch and spec.stretch ~= 0 then
-            t[#t+1] = format("plus %s",aux.strip_zeros(number.topoints(spec.stretch)))
-        end
-        if spec.shrink_order and spec.shrink_order ~= 0 then
-            t[#t+1] = format("minus -%sfi%s",spec.shrink/65536,string.rep("l",math.abs(spec.shrink_order)-1))
-        elseif spec.shrink and spec.shrink ~= 0 then
-            t[#t+1] = format("minus %s",aux.strip_zeros(number.topoints(spec.shrink)))
-        end
-        return concat(t," ")
-    else
-        return "[0pt]"
+    local t = { }
+    t[#t+1] = aux.strip_zeros(number.topoints(spec.width))
+    if spec.stretch_order and spec.stretch_order ~= 0 then
+        t[#t+1] = format("plus -%sfi%s",spec.stretch/65536,string.rep("l",math.abs(spec.stretch_order)-1))
+    elseif spec.stretch and spec.stretch ~= 0 then
+        t[#t+1] = format("plus %s",aux.strip_zeros(number.topoints(spec.stretch)))
     end
+    if spec.shrink_order and spec.shrink_order ~= 0 then
+        t[#t+1] = format("minus -%sfi%s",spec.shrink/65536,string.rep("l",math.abs(spec.shrink_order)-1))
+    elseif spec.shrink and spec.shrink ~= 0 then
+        t[#t+1] = format("minus %s",aux.strip_zeros(number.topoints(spec.shrink)))
+    end
+    return concat(t," ")
 end
 
 local function nodes_to_string(head)
@@ -400,11 +464,7 @@ local function nodes_to_string(head)
         if id == penalty then
             t[#t+1] = format("%s:%s",ty,current.penalty)
         elseif id == glue then
-            if current.spec then
-                t[#t+1] = format("%s:%s",ty,aux.strip_zeros(number.topoints(current.spec.width)))
-            else
-                t[#t+1] = format("%s:[0pt]",ty)
-            end
+            t[#t+1] = format("%s:%s",ty,aux.strip_zeros(number.topoints(current.spec.width)))
         elseif id == kern then
             t[#t+1] = format("%s:%s",ty,aux.strip_zeros(number.topoints(current.kern)))
         else
@@ -458,13 +518,13 @@ local function show_tracing(head)
         for i=1,#trace_list do
             local tag, text = unpack(trace_list[i])
             if tag == "info" then
-                logs.report("collapse",text)
+                report_collapser(text)
             else
-                logs.report("collapse","  %s: %s",tag,text)
+                report_collapser("  %s: %s",tag,text)
             end
         end
-        logs.report("collapse","before: %s",before)
-        logs.report("collapse","after : %s",after)
+        report_collapser("before: %s",before)
+        report_collapser("after : %s",after)
     end
 end
 
@@ -513,25 +573,27 @@ local skips = {
 }
 
 local free_glue_node = free_node
-local free_glue_spec = free_node
 local discard, largest, force, penalty, add, disable, nowhite, goback, together = 0, 1, 2, 3, 4, 5, 6, 7, 8
 
+--~ local function free_glue_node(n) free_node(n.spec) free_node(n) end
+
 function vspacing.snap_box(n,how)
     local sv = snapmethods[how]
     if sv then
-        local list = texbox[n].list
+        local box = texbox[n]
+        local list = box.list
 --~         if list and (list.id == hlist or list.id == vlist) then
         if list then
             local s = has_attribute(list,snap_method)
             if s == 0 then
                 if trace_vsnapping then
-                    logs.report("snapper", "hlist not snapped, already done")
+                --  report_snapper("hlist not snapped, already done")
                 end
             else
-                local h, d, ch, cd, lines = snap_hlist(list,sv,texht[n],texdp[n])
-                texht[n], texdp[n] = ch, cd
+                local h, d, ch, cd, lines = snap_hlist(box,sv,box.height,box.depth)
+                box.height, box.depth = ch, cd
                 if trace_vsnapping then
-                    logs.report("snapper", "hlist snapped from (%s,%s) to (%s,%s) using method '%s' (%s) for '%s' (%s lines)",h,d,ch,cd,sv.name,sv.specification,"direct",lines)
+                    report_snapper("hlist snapped from (%s,%s) to (%s,%s) using method '%s' (%s) for '%s' (%s lines)",h,d,ch,cd,sv.name,sv.specification,"direct",lines)
                 end
                 set_attribute(list,snap_method,0)
             end
@@ -552,7 +614,7 @@ local function forced_skip(head,current,width,where,trace)
         current = c
     end
     if trace then
-        logs.report("vspacing", "inserting forced skip of %s",width)
+        report_vspacing("inserting forced skip of %s",width)
     end
     return head, current
 end
@@ -572,17 +634,16 @@ local function collapser(head,where,what,trace,snap) -- maybe also pass tail
         if penalty_data then
             local p = make_penalty_node(penalty_data)
             if trace then trace_done("flushed due to " .. why,p) end
-            head, _ = insert_node_before(head,current,p)
+            head = insert_node_before(head,current,p)
         end
         if glue_data then
             if force_glue then
                 if trace then trace_done("flushed due to " .. why,glue_data) end
-                local spec = glue_data.spec
-                head, _ = forced_skip(head,current,(spec and spec.width) or 0,"before",trace)
+                head = forced_skip(head,current,glue_data.spec.width,"before",trace)
                 free_glue_node(glue_data)
-            elseif glue_data.spec then
+            elseif glue_data.spec.writable then
                 if trace then trace_done("flushed due to " .. why,glue_data) end
-                head, _ = insert_node_before(head,current,glue_data)
+                head = insert_node_before(head,current,glue_data)
             else
                 free_glue_node(glue_data)
             end
@@ -593,32 +654,34 @@ local function collapser(head,where,what,trace,snap) -- maybe also pass tail
         parskip, ignore_parskip, ignore_following, ignore_whitespace = nil, false, false, false
     end
     if trace_vsnapping then
-        logs.report("snapper", "global ht/dp = %s/%s, local ht/dp = %s/%s",
+        report_snapper("global ht/dp = %s/%s, local ht/dp = %s/%s",
             texdimen.globalbodyfontstrutheight, texdimen.globalbodyfontstrutdepth,
             texdimen.bodyfontstrutheight, texdimen.bodyfontstrutdepth)
     end
+    if trace then trace_info("start analyzing",where,what) end
     while current do
         local id, subtype = current.id, current.subtype
         if id == hlist or id == vlist then
+-- needs checking, why so many calls
             if snap then
                 local s = has_attribute(current,snap_method)
                 if not s then
                 --  if trace_vsnapping then
-                --      logs.report("snapper", "hlist not snapped")
+                --      report_snapper("hlist not snapped")
                 --  end
                 elseif s == 0 then
                     if trace_vsnapping then
-                        logs.report("snapper", "hlist not snapped, already done")
+                    --  report_snapper("hlist not snapped, already done")
                     end
                 else
                     local sv = snapmethods[s]
                     if sv then
                         local h, d, ch, cd, lines = snap_hlist(current,sv)
                         if trace_vsnapping then
-                            logs.report("snapper", "hlist snapped from (%s,%s) to (%s,%s) using method '%s' (%s) for '%s' (%s lines)",h,d,ch,cd,sv.name,sv.specification,where,lines)
+                            report_snapper("hlist snapped from (%s,%s) to (%s,%s) using method '%s' (%s) for '%s' (%s lines)",h,d,ch,cd,sv.name,sv.specification,where,lines)
                         end
                     elseif trace_vsnapping then
-                        logs.report("snapper", "hlist not snapped due to unknown snap specification")
+                        report_snapper("hlist not snapped due to unknown snap specification")
                     end
                     set_attribute(current,snap_method,0)
                 end
@@ -636,7 +699,7 @@ local function collapser(head,where,what,trace,snap) -- maybe also pass tail
         elseif id == kern then
             if snap and trace_vsnapping and current.kern ~= 0 then
             --~ current.kern = 0
-                logs.report("snapper", "kern of %s (kept)",current.kern)
+                report_snapper("kern of %s (kept)",current.kern)
             end
             flush("kern")
             current = current.next
@@ -668,11 +731,12 @@ local function collapser(head,where,what,trace,snap) -- maybe also pass tail
                     local previous = current.prev
                     if previous and previous.id == glue and previous.subtype == 0 then
                         local ps = previous.spec
-                        if ps then
+                        if ps.writable then
                             local cs = current.spec
-                            if cs and ps.stretch_order == 0 and ps.shrink_order == 0 and cs.stretch_order == 0 and cs.shrink_order == 0 then
+                            if cs.writable and ps.stretch_order == 0 and ps.shrink_order == 0 and cs.stretch_order == 0 and cs.shrink_order == 0 then
                                 local pw, pp, pm = ps.width, ps.stretch, ps.shrink
                                 local cw, cp, cm = cs.width, cs.stretch, cs.shrink
+                                ps = writable_spec(previous) -- no writable needed here
                                 ps.width, ps.stretch, ps.shrink = pw + cw, pp + cp, pm + cm
                                 if trace then trace_natural("removed",current) end
                                 head, current = remove_node(head, current, true)
@@ -723,8 +787,7 @@ local function collapser(head,where,what,trace,snap) -- maybe also pass tail
                 -- is now exclusive, maybe support goback as combi, else why a set
                 if sc == largest then
                     local cs, gs = current.spec, glue_data.spec
-                    local cw = (cs and cs.width) or 0
-                    local gw = (gs and gs.width) or 0
+                    local cw, gw = cs.width, gs.width
                     if cw > gw then
                         if trace then trace_skip('largest',sc,so,sp,current) end
                         free_glue_node(glue_data) -- also free spec
@@ -751,7 +814,8 @@ local function collapser(head,where,what,trace,snap) -- maybe also pass tail
                     head, current = remove_node(head, current, true)
                 elseif sc == add then
                     if trace then trace_skip('add',sc,so,sp,current) end
-                    local old, new = glue_data.spec, current.spec
+--~                     local old, new = glue_data.spec, current.spec
+                    local old, new = writable_spec(glue_data), current.spec
                     old.width   = old.width   + new.width
                     old.stretch = old.stretch + new.stretch
                     old.shrink  = old.shrink  + new.shrink
@@ -773,11 +837,11 @@ local function collapser(head,where,what,trace,snap) -- maybe also pass tail
                 local s = has_attribute(current,snap_method)
                 if s and s ~= 0 then
                     set_attribute(current,snap_method,0)
-                    local spec = current.spec
-                    if spec then
+                    if current.spec.writable then
+                        local spec = writable_spec(current)
                         spec.width = 0
                         if trace_vsnapping then
-                            logs.report("snapper", "lineskip set to zero")
+                            report_snapper("lineskip set to zero")
                         end
                     end
                 else
@@ -794,11 +858,11 @@ local function collapser(head,where,what,trace,snap) -- maybe also pass tail
                 local s = has_attribute(current,snap_method)
                 if s and s ~= 0 then
                     set_attribute(current,snap_method,0)
-                    local spec = current.spec
-                    if spec then
+                    if current.spec.writable then
+                        local spec = writable_spec(current)
                         spec.width = 0
                         if trace_vsnapping then
-                            logs.report("snapper", "baselineskip set to zero")
+                            report_snapper("baselineskip set to zero")
                         end
                     end
                 else
@@ -817,7 +881,7 @@ local function collapser(head,where,what,trace,snap) -- maybe also pass tail
                 head, current = remove_node(head, current, true)
             elseif glue_data then
                 local ps, gs = current.spec, glue_data.spec
-                if ps and gs and ps.width > gs.width then
+                if ps.writable and gs.writable and ps.width > gs.width then
                     glue_data.spec = copy_node(ps)
                     if trace then trace_natural("taking parskip",current) end
                 else
@@ -836,7 +900,7 @@ local function collapser(head,where,what,trace,snap) -- maybe also pass tail
                     local sv = snapmethods[s]
                     local w, cw = snap_topskip(current,sv)
                     if trace_vsnapping then
-                        logs.report("snapper", "topskip snapped from %s to %s for '%s'",w,cw,where)
+                        report_snapper("topskip snapped from %s to %s for '%s'",w,cw,where)
                     end
                 else
                     if trace then trace_skip("topskip",sc,so,sp,current) end
@@ -872,8 +936,8 @@ local function collapser(head,where,what,trace,snap) -- maybe also pass tail
             current = current.next
             --
         else -- other glue
-            if snap and trace_vsnapping and current.spec and current.spec.width ~= 0 then
-                logs.report("snapper", "%s of %s (kept)",skips[subtype],current.spec.width)
+            if snap and trace_vsnapping and current.spec.writable and current.spec.width ~= 0 then
+                report_snapper("%s of %s (kept)",skips[subtype],current.spec.width)
             --~ current.spec.width = 0
             end
             if trace then trace_skip(format("some glue (%s)",subtype),sc,so,sp,current) end
@@ -930,16 +994,16 @@ end
 local stackhead, stacktail, stackhack = nil, nil, false
 
 local function report(message,lst)
-    logs.report("vspacing",message,count_nodes(lst,true),node_ids_to_string(lst))
+    report_vspacing(message,count_nodes(lst,true),node_ids_to_string(lst))
 end
 
-function nodes.handle_page_spacing(where)
-    local newhead = texlists.contrib_head
+function nodes.handle_page_spacing(newhead,where)
+--~     local newhead = texlists.contrib_head
     if newhead then
-        starttiming(vspacing)
+--~         starttiming(vspacing)
         local newtail = find_node_tail(newhead)
         local flush = false
-stackhack = true -- todo: only when grid snapping once enabled
+        stackhack = true -- todo: only when grid snapping once enabled
         for n in traverse_nodes(newhead) do -- we could just look for glue nodes
             local id = n.id
             if id == glue then
@@ -967,10 +1031,11 @@ stackhack = true -- todo: only when grid snapping once enabled
             if stackhack then
                 stackhack = false
                 if trace_collect_vspacing then report("processing %s nodes: %s",newhead) end
-                texlists.contrib_head = collapser(newhead,"page",where,trace_page_vspacing,true)
+--~                 texlists.contrib_head = collapser(newhead,"page",where,trace_page_vspacing,true)
+newhead = collapser(newhead,"page",where,trace_page_vspacing,true)
             else
                 if trace_collect_vspacing then report("flushing %s nodes: %s",newhead) end
-                texlists.contrib_head = newhead
+--~                 texlists.contrib_head = newhead
             end
         else
             if stackhead then
@@ -982,10 +1047,12 @@ stackhack = true -- todo: only when grid snapping once enabled
                 stackhead = newhead
             end
             stacktail = newtail
-            texlists.contrib_head = nil
+--~             texlists.contrib_head = nil
+newhead = nil
         end
-        stoptiming(vspacing)
+--~         stoptiming(vspacing)
     end
+return newhead
 end
 
 local ignore = table.tohash {
@@ -996,9 +1063,9 @@ local ignore = table.tohash {
 
 function nodes.handle_vbox_spacing(head,where)
     if head and not ignore[where] and head.next then
-        starttiming(vspacing)
+    --  starttiming(vspacing)
         head = collapser(head,"vbox",where,trace_vbox_vspacing,false)
-        stoptiming(vspacing)
+    --  stoptiming(vspacing)
     end
     return head
 end
@@ -1006,33 +1073,12 @@ end
 function nodes.collapse_vbox(n) -- for boxes
     local list = texbox[n].list
     if list then
-        starttiming(vspacing)
+    --  starttiming(vspacing)
         texbox[n].list = vpack_node(collapser(list,"snapper","vbox",trace_vbox_vspacing,true))
-        stoptiming(vspacing)
+    --  stoptiming(vspacing)
     end
 end
 
-statistics.register("v-node processing time", function()
-    if statistics.elapsedindeed(vspacing) then
-        return format("%s seconds", statistics.elapsedtime(vspacing))
-    end
-end)
-
--- these are experimental callback definitions that definitely will
--- be moved elsewhere as part of a chain of vnode handling
-
-function vspacing.enable()
-    callback.register('vpack_filter', nodes.handle_vbox_spacing) -- enabled per 2009/10/16
-    callback.register('buildpage_filter', nodes.handle_page_spacing)
-end
-
-function vspacing.disable()
-    callback.register('vpack_filter', nil)
-    callback.register('buildpage_filter', nil)
-end
-
-vspacing.enable()
-
 -- we will split this module hence the locals
 
 local attribute = attributes.private('graphicvadjust')
@@ -1058,7 +1104,7 @@ function nodes.repackage_graphicvadjust(head,groupcode) -- we can make an action
                         head, h, n = remove_node(head,h)
                         local pl = p.list
                         if n.width ~= 0 then
-                            n = hpack_node(n,0,'exactly')
+                            n = hpack_node(n,0,'exactly') -- todo: dir
                         end
                         if pl then
                             pl.prev = n
@@ -1083,4 +1129,62 @@ function nodes.repackage_graphicvadjust(head,groupcode) -- we can make an action
     end
 end
 
+--~ function nodes.repackage_graphicvadjust(head,groupcode) -- we can make an actionchain for mvl only
+--~     if groupcode == "" then -- mvl only
+--~         return head, false
+--~     else
+--~         return head, false
+--~     end
+--~ end
+
 --~ tasks.appendaction("finalizers", "lists", "nodes.repackage_graphicvadjust")
+
+nodes.builders = nodes.builder or { }
+
+local builders = nodes.builders
+
+local actions = tasks.actions("vboxbuilders",5)
+
+function nodes.builders.vpack_filter(head,groupcode,size,packtype,maxdepth,direction)
+    local done = false
+    if head then
+        starttiming(builders)
+        if trace_vpacking then
+            local before = nodes.count(head)
+            head, done = actions(head,groupcode,size,packtype,maxdepth,direction)
+            local after = nodes.count(head)
+            if done then
+                nodes.processors.tracer("vpack","changed",head,groupcode,before,after,true)
+            else
+                nodes.processors.tracer("vpack","unchanged",head,groupcode,before,after,true)
+            end
+            stoptiming(builders)
+        else
+            head, done = actions(head,groupcode)
+            stoptiming(builders)
+        end
+    end
+    return head, done
+end
+
+-- This one is special in the sense that it has no head
+-- and we operate on the mlv. Also, we need to do the
+-- vspacing last as it removes items from the mvl.
+
+local actions = tasks.actions("mvlbuilders",1)
+
+function nodes.builders.buildpage_filter(groupcode)
+    starttiming(builders)
+    local head = texlists.contrib_head
+    local head, done = actions(head,groupcode)
+    texlists.contrib_head = head
+    stoptiming(builders)
+    return (done and head) or true
+end
+
+callbacks.register('vpack_filter',     nodes.builders.vpack_filter, "vertical spacing etc")
+callbacks.register('buildpage_filter', nodes.builders.buildpage_filter, "vertical spacing etc (mvl)")
+
+statistics.register("v-node processing time", function()
+    return statistics.elapsedseconds(builders)
+end)
