@@ -6,9 +6,15 @@ if not modules then modules = { } end modules ['strc-doc'] = {
     license   = "see context related readme files"
 }
 
+-- todo: associate counter with head
+
+-- we need to freeze and document this module
+
 local next, type = next, type
-local format, gsub, find, concat = string.format, string.gsub, string.find, table.concat
+local format, gsub, find, concat, gmatch, match = string.format, string.gsub, string.find, table.concat, string.gmatch, string.match
 local texsprint, texwrite = tex.sprint, tex.write
+local concat = table.concat
+local max, min = math.max, math.min
 
 local ctxcatcodes = tex.ctxcatcodes
 local variables   = interfaces.variables
@@ -16,11 +22,9 @@ local variables   = interfaces.variables
 --~ if not trackers then trackers = { register = function() end } end
 
 local trace_sectioning = false  trackers.register("structure.sectioning", function(v) trace_sectioning = v end)
+local trace_detail     = false  trackers.register("structure.detail",     function(v) trace_detail     = v end)
 
-local function report(...)
---~ print(...)
-    logs.report("sectioning:",...)
-end
+local report_structure = logs.new("structure")
 
 structure            = structure            or { }
 structure.helpers    = structure.helpers    or { }
@@ -190,15 +194,18 @@ end
 function sections.somelevel(given)
     -- old number
     local numbers, ownnumbers, forced, status, olddepth = data.numbers, data.ownnumbers, data.forced, data.status, data.depth
---~ print("old",olddepth,given.metadata.name,levelmap[given.metadata.name])
-    local newdepth = tonumber(levelmap[given.metadata.name] or (olddepth > 0 and olddepth) or 1) -- hm, levelmap only works for section-*
---~ print("new",newdepth)
+    local givenname = given.metadata.name
+    local mappedlevel = levelmap[givenname]
+    local newdepth = tonumber(mappedlevel or (olddepth > 0 and olddepth) or 1) -- hm, levelmap only works for section-*
     local directives = given.directives
     local resetset = (directives and directives.resetset) or ""
-    local resetter = sets.getall("structure:resets",data.block,resetset)
+ -- local resetter = sets.getall("structure:resets",data.block,resetset)
     -- a trick to permits userdata to overload title, ownnumber and reference
     -- normally these are passed as argument but nowadays we provide several
     -- interfaces (we need this because we want to be compatible)
+    if trace_detail then
+        report_structure("name '%s', mapped level '%s', old depth '%s', new depth '%s', reset set '%s'",givenname,mappedlevel,olddepth,newdepth,resetset)
+    end
     local u = given.userdata
     if u then
         -- kind of obsolete as we can pass them directly anyway
@@ -212,7 +219,9 @@ function sections.somelevel(given)
     if newdepth > olddepth then
         for i=olddepth+1,newdepth do
             local s = tonumber(sets.get("structure:resets",data.block,resetset,i))
---~ logs.report("structure >","old: %s, new:%s, reset: %s (%s: %s)",olddepth,newdepth,s,resetset,table.concat(resetter,","))
+            if trace_detail then
+                report_structure("new>old (%s>%s), reset set '%s', reset value '%s', current '%s'",olddepth,newdepth,resetset,s or "?",numbers[i] or "?")
+            end
             if not s or s == 0 then
                 numbers[i] = numbers[i] or 0
                 ownnumbers[i] = ownnumbers[i] or ""
@@ -225,7 +234,9 @@ function sections.somelevel(given)
     elseif newdepth < olddepth then
         for i=olddepth,newdepth+1,-1 do
             local s = tonumber(sets.get("structure:resets",data.block,resetset,i))
---~ logs.report("structure <","old: %s, new:%s, reset: %s (%s: %s)",olddepth,newdepth,s,resetset,table.concat(resetter,","))
+            if trace_detail then
+                report_structure("new<old (%s<%s), reset set '%s', reset value '%s', current '%s'",olddepth,newdepth,resetset,s or "?",numbers[i] or "?")
+            end
             if not s or s == 0 then
                 numbers[i] = numbers[i] or 0
                 ownnumbers[i] = ownnumbers[i] or ""
@@ -236,14 +247,14 @@ function sections.somelevel(given)
             status[i] = nil
         end
     end
-structure.counters.check(newdepth)
+    structure.counters.check(newdepth)
     ownnumbers[newdepth] = given.numberdata.ownnumber or ""
     given.numberdata.ownnumber = nil
     data.depth = newdepth
     -- new number
     olddepth = newdepth
     if given.metadata.increment then
-        local oldn, newn = numbers[newdepth], 0
+        local oldn, newn = numbers[newdepth] or 0, 0
         local fd = forced[newdepth]
         if fd then
             if fd[1] == "add" then
@@ -255,21 +266,31 @@ structure.counters.check(newdepth)
                 newn = 1 -- maybe zero is nicer
             end
             forced[newdepth] = nil
+            if trace_detail then
+                report_structure("old depth '%s', new depth '%s, old n '%s', new n '%s', forced '%s'",olddepth,newdepth,oldn,newn,concat(fd,""))
+            end
         elseif newn then
             newn = oldn + 1
+            if trace_detail then
+                report_structure("old depth '%s', new depth '%s, old n '%s', new n '%s', increment",olddepth,newdepth,oldn,newn)
+            end
         else
             local s = tonumber(sets.get("structure:resets",data.block,resetset,newdepth))
---~ logs.report("structure =","old: %s, new:%s, reset: %s (%s: %s)",olddepth,newdepth,s,resetset,table.concat(resetter,","))
-            if not s or s == 0 then
+            if not s then
+                newn = oldn or 0
+            elseif s == 0 then
                 newn = oldn or 0
             else
                 newn = s - 1
             end
+            if trace_detail then
+                report_structure("old depth '%s', new depth '%s, old n '%s', new n '%s', reset",olddepth,newdepth,oldn,newn)
+            end
         end
         numbers[newdepth] = newn
     end
     status[newdepth] = given or { }
-    for k, v in pairs(data.checkers) do
+    for k, v in next, data.checkers do
         if v[1] == newdepth and v[2] then
             v[2](k)
         end
@@ -288,6 +309,9 @@ structure.counters.check(newdepth)
     if #ownnumbers > 0 then
         numberdata.ownnumbers = table.fastcopy(ownnumbers)
     end
+    if trace_detail then
+        report_structure("name '%s', numbers '%s', own numbers '%s'",givenname,concat(numberdata.numbers, " "),concat(numberdata.ownnumbers, " "))
+    end
     given.references.section = sections.save(given)
  -- given.numberdata = nil
 end
@@ -297,7 +321,7 @@ function sections.writestatus()
         local numbers, ownnumbers, status, depth = data.numbers, data.ownnumbers, data.status, data.depth
         local d = status[depth]
         local o = concat(ownnumbers,".",1,depth)
-        local n = (numbers and concat(numbers,".",1,depth)) or 0
+        local n = (numbers and concat(numbers,".",1,min(depth,#numbers))) or 0
         local l = d.titledata.title or ""
         local t = (l ~= "" and l) or d.titledata.title or "[no title]"
         local m = d.metadata.name
@@ -314,7 +338,7 @@ end
 function sections.setnumber(depth,n)
     local forced, depth, new = data.forced, depth or data.depth, tonumber(n)
     if type(n) == "string" then
-        if n:find("^[%+%-]") then
+        if find(n,"^[%+%-]") then
             forced[depth] = { "add", new }
         else
             forced[depth] = { "set", new }
@@ -345,8 +369,8 @@ function sections.matching_till_depth(depth,numbers,parentnumbers)
     return ok
 end
 
-function sections.getnumber(depth)
-    return texwrite(data.numbers[depth] or 0)
+function sections.getnumber(depth) -- redefined later ...
+    texwrite(data.numbers[depth] or 0)
 end
 
 function sections.set(key,value)
@@ -362,7 +386,7 @@ function sections.structuredata(depth,key,default,honorcatcodetable) -- todo: sp
     if not depth or depth == 0 then depth = data.depth end
     local data = data.status[depth]
     local d = data
-    for k in key:gmatch("([^.]+)") do
+    for k in gmatch(key,"([^.]+)") do
         if type(d) == "table" then
             d = d[k]
             if not d then
@@ -453,16 +477,20 @@ function sections.typesetnumber(entry,kind,...) -- kind='section','number','pref
         local conversionset = ""
         local conversion    = ""
         local stopper       = ""
+        local starter       = ""
         local connector     = ""
         local set           = ""
         local segments      = ""
         local criterium     = ""
-        for _, data in ipairs { ... } do -- can be multiple parametersets
+        local dataset = { ... }
+        for d=1,#dataset do
+            local data = dataset[d] -- can be multiple parametersets
             if data then
                 if separatorset  == "" then separatorset  = data.separatorset  or "" end
                 if conversionset == "" then conversionset = data.conversionset or "" end
                 if conversion    == "" then conversion    = data.conversion    or "" end
                 if stopper       == "" then stopper       = data.stopper       or "" end
+                if starter       == "" then starter       = data.starter       or "" end
                 if connector     == "" then connector     = data.connector     or "" end
                 if set           == "" then set           = data.set           or "" end
                 if segments      == "" then segments      = data.segments      or "" end
@@ -473,6 +501,7 @@ function sections.typesetnumber(entry,kind,...) -- kind='section','number','pref
         if conversionset == "" then conversionset = "default"  end -- not used
         if conversion    == "" then conversion    = nil        end
         if stopper       == "" then stopper       = nil        end
+        if starter       == "" then starter       = nil        end
         if connector     == "" then connector     = nil        end
         if set           == "" then set           = "default"  end
         if segments      == "" then segments      = nil        end
@@ -489,7 +518,7 @@ function sections.typesetnumber(entry,kind,...) -- kind='section','number','pref
         --
         local firstprefix, lastprefix = 0, 16
         if segments then
-            local f, l = (tostring(segments)):match("^(.-):(.+)$")
+            local f, l = match(tostring(segments),"^(.-):(.+)$")
             if f and l then
                 -- 0:100, chapter:subsubsection
                 firstprefix = tonumber(f) or sections.getlevel(f) or 0
@@ -539,6 +568,9 @@ function sections.typesetnumber(entry,kind,...) -- kind='section','number','pref
             end
             --
             local prefixlist = set and sets.getall("structure:prefixes","",set) -- "" == block
+            if starter then
+                processors.sprint(ctxcatcodes,starter)
+            end
             if prefixlist and (kind == 'section' or kind == 'prefix') then
                 -- find valid set (problem: for sectionnumber we should pass the level)
             --  if kind == "section" then
@@ -609,17 +641,7 @@ function sections.typesetnumber(entry,kind,...) -- kind='section','number','pref
                 processors.sprint(ctxcatcodes,stopper)
             end
         else
-        --  report("error: no numbers")
-        end
-    end
-end
-
-function sections.fullnumber(depth)
-    local data = data.status[depth or data.depth]
-    if data then
-        local sectiondata = jobsections.collected[data.references.section] -- tobesaved?
-        if sectiondata and sectiondata.hidenumber ~= true then -- can be nil
-            sections.typesetnumber(sectiondata,'section',sectiondata)
+        --  report_structure("error: no numbers")
         end
     end
 end
@@ -630,3 +652,48 @@ function sections.title()
         helpers.title(sc.titledata.title,sc.metadata)
     end
 end
+
+function sections.findnumber(depth,what)
+    local data = data.status[depth or data.depth]
+    if data then
+        local index = data.references.section
+        local collected = jobsections.collected
+        local sectiondata = collected[index]
+        if sectiondata and sectiondata.hidenumber ~= true then -- can be nil
+            if what == variables.first then
+                for i=index,1,-1 do
+                    local s = collected[i]
+                    local n = s.numbers
+                    if #n == depth and n[depth] and n[depth] ~= 0 then
+                        sectiondata = s
+                    elseif #n < depth then
+                        break
+                    end
+                end
+            elseif what == variables.last then
+                for i=index,#collected do
+                    local s = collected[i]
+                    local n = s.numbers
+                    if #n == depth and n[depth] and n[depth] ~= 0 then
+                        sectiondata = s
+                    elseif #n < depth then
+                        break
+                    end
+                end
+            end
+            return sectiondata
+        end
+    end
+end
+
+function sections.fullnumber(depth,what,raw)
+    local sectiondata = sections.findnumber(depth,what)
+    if sectiondata then
+        sections.typesetnumber(sectiondata,'section',sectiondata)
+    end
+end
+
+function sections.getnumber(depth,what) -- redefined here
+    local sectiondata = sections.findnumber(depth,what)
+    texwrite((sectiondata and sectiondata.numbers[depth]) or 0)
+end
