@@ -11,12 +11,16 @@ if not modules then modules = { } end modules ['strc-lst'] = {
 -- section, metadata cache (internal then has to move up one level) or a
 -- shared cache [we can use a fast and stupid serializer]
 
-local format, tonumber = string.format, tonumber
+local format, gmatch, gsub = string.format, string.gmatch, string.gsub
+local tonumber = tonumber
 local texsprint, texprint, texwrite, texcount = tex.sprint, tex.print, tex.write, tex.count
 local concat, insert, remove = table.concat, table.insert, table.remove
+local lpegmatch = lpeg.match
 
 local trace_lists = false  trackers.register("structure.lists", function(v) trace_lists = v end)
 
+local report_lists = logs.new("lists")
+
 local ctxcatcodes = tex.ctxcatcodes
 
 structure.lists     = structure.lists     or { }
@@ -36,6 +40,9 @@ lists.tobesaved = lists.tobesaved or { }
 lists.enhancers = lists.enhancers or { }
 lists.internals = lists.internals or { }
 lists.ordered   = lists.ordered   or { }
+lists.cached    = lists.cached    or { }
+
+local cached, pushed = lists.cached, { }
 
 local variables = interfaces.variables
 local matching_till_depth, number_at_depth = sections.matching_till_depth, sections.number_at_depth
@@ -79,8 +86,6 @@ if job then
     job.register('structure.lists.collected', structure.lists.tobesaved, initializer)
 end
 
-local cached, pushed = { }, { }
-
 function lists.push(t)
     local r = t.references
     local i = (r and r.internal) or 0 -- brrr
@@ -115,6 +120,27 @@ function lists.enhance(n)
         if enhancer then
             enhancer(l)
         end
+        return l
+    end
+end
+
+function lists.enforce(n)
+ -- todo: symbolic names for counters
+    local l = cached[n]
+    if l then
+        --
+        l.directives = nil -- might change
+        -- save in the right order (happens at shipout)
+        lists.tobesaved[#lists.tobesaved+1] = l
+        -- default enhancer (cross referencing)
+        l.references.realpage = texcount.realpageno
+        -- specific enhancer (kind of obsolete)
+        local kind = l.metadata.kind
+        local enhancer = kind and lists.enhancers[kind]
+        if enhancer then
+            enhancer(l)
+        end
+        return l
     end
 end
 
@@ -135,18 +161,19 @@ function lists.popnesting()
     lists.result = old.result
 end
 
+-- will be split
+
 local function filter_collected(names, criterium, number, collected, nested)
     local numbers, depth = documents.data.numbers, documents.data.depth
     local hash, result, all, detail = { }, { }, not names or names == "" or names == variables.all, nil
+    names, criterium = gsub(names," ",""), gsub(criterium," ","")
+    if trace_lists then
+        report_lists("filtering names: %s, criterium: %s, number: %s",names,criterium,number or "-")
+    end
     if not all then
-        for s in names:gmatch("[^, ]+") do
-            if trace_lists then
-                logs.report("lists","filtering: %s",s)
-            end
+        for s in gmatch(names,"[^, ]+") do -- sort of settings to hash
             hash[s] = true
         end
-    elseif trace_lists then
-        logs.report("lists","filtering all")
     end
     if criterium == variables.intro then
         -- special case, no structure yet
@@ -203,7 +230,7 @@ local function filter_collected(names, criterium, number, collected, nested)
             end
         end
     elseif criterium == variables.here then
-        -- this is qquite dirty ... as cnumbers is not sparse we can misuse #cnumbers
+        -- this is quite dirty ... as cnumbers is not sparse we can misuse #cnumbers
         if depth == 0 then
             return filter_collected(names,variables.intro,number,collected)
         else
@@ -305,9 +332,9 @@ local function filter_collected(names, criterium, number, collected, nested)
     end
     if trace_lists then
         if detail then
-            logs.report("lists","criterium: %s, %s, found: %s",criterium,detail,#result)
+            report_lists("criterium: %s, %s, found: %s",criterium,detail,#result)
         else
-            logs.report("lists","criterium: %s, found: %s",criterium,#result)
+            report_lists("criterium: %s, found: %s",criterium,#result)
         end
     end
     return result
@@ -327,6 +354,7 @@ function lists.process(...)
         local r = lists.result[i]
         local m = r.metadata
         texsprint(ctxcatcodes,format("\\processlistofstructure{%s}{%s}{%i}",m.name,m.kind,i))
+--~         context.processlistofstructure(m.name,m.kind,i)
     end
 end
 
@@ -334,12 +362,14 @@ function lists.analyze(...)
     lists.result = lists.filter(...)
 end
 
-function lists.userdata(name,r,tag) -- to tex
-    local str = lists.result[r]
-    str = str and str.userdata
-    str = str and str[tag]
-    if str then
-        texsprint(ctxcatcodes,str)
+function lists.userdata(name,r,tag) -- to tex (todo: xml)
+    local result = lists.result[r]
+    if result then
+        local userdata, metadata = result.userdata, result.metadata
+        local str = userdata and userdata[tag]
+        if str then
+            texsprint(metadata and metadata.catcodes or ctxcatcodes,str)
+        end
     end
 end
 
@@ -354,7 +384,7 @@ function lists.size()
     texprint(#lists.result)
 end
 
-function lists.location(name,n)
+function lists.location(n)
     local l = lists.result[n]
     texsprint(l.references.internal or n)
 end
@@ -451,7 +481,30 @@ function lists.prefixednumber(name,n,prefixspec,numberspec)
         helpers.prefix(data,prefixspec)
         local numberdata = data.numberdata
         if numberdata then
-            sections.typesetnumber(numberdata,"number",spec or false,numberdata or false)
+--~     print(table.serialize(numberspec))
+            sections.typesetnumber(numberdata,"number",numberspec or false,numberdata or false)
+        end
+    end
+end
+
+-- todo, do this in references namespace ordered instead (this is an experiment)
+--
+-- also see lpdf-ano (maybe move this there)
+
+local splitter = lpeg.splitat(":")
+
+function jobreferences.specials.order(var,actions) -- jobreferences.specials !
+    local operation = var.operation
+    if operation then
+        local kind, name, n = lpegmatch(splitter,operation)
+        local order = lists.ordered[kind]
+        order = order and order[name]
+        local v = order[tonumber(n)]
+        local r = v and v.references.realpage
+        if r then
+            actions.realpage = r
+            var.operation = r -- brrr, but test anyway
+            return jobreferences.specials.page(var,actions)
         end
     end
 end
