@@ -7,10 +7,13 @@ if not modules then modules = { } end modules ['strc-ref'] = {
 }
 
 local format, find, gmatch, match, concat = string.format, string.find, string.gmatch, string.match, table.concat
-local texsprint, texwrite, texcount = tex.sprint, tex.write, tex.count
+local lpegmatch = lpeg.match
+local texsprint, texwrite, texcount, texsetcount = tex.sprint, tex.write, tex.count, tex.setcount
 
 local trace_referencing = false  trackers.register("structure.referencing", function(v) trace_referencing = v end)
 
+local report_references = logs.new("references")
+
 local ctxcatcodes = tex.ctxcatcodes
 local variables   = interfaces.variables
 local constants   = interfaces.constants
@@ -69,6 +72,33 @@ end
 
 -- todo: delay split till later as in destinations we split anyway
 
+local orders, lastorder = { }, 0
+
+local function setnextorder(kind,name)
+    lastorder = 0
+    if kind and name then
+        local ok = orders[kind]
+        if not ok then
+            ok = { }
+            orders[kind] = ok
+        end
+        lastorder = (ok[name] or 0) + 1
+        ok[name] = lastorder
+    end
+    texsetcount("global","locationorder",lastorder)
+end
+
+jobreferences.setnextorder = setnextorder
+
+function jobreferences.setnextinternal(kind,name)
+    setnextorder(kind,name) -- always incremented with internal
+    texsetcount("global","locationcount",texcount.locationcount + 1)
+end
+
+function jobreferences.currentorder(kind,name)
+    texwrite((orders[kind] and orders[kind][name]) or lastorder)
+end
+
 function jobreferences.set(kind,prefix,tag,data)
     for ref in gmatch(tag,"[^,]+") do
         local p, r = match(ref,"^(%-):(.-)$")
@@ -133,11 +163,11 @@ local special_reference  = special * lparent * (operation * optional_arguments +
 local scanner = (reset * outer_reference * (special_reference + inner_reference)^-1 * -1) / function() return result end
 
 --~ function jobreferences.analyse(str) -- overloaded
---~     return scanner:match(str)
+--~     return lpegmatch(scanner,str)
 --~ end
 
 function jobreferences.split(str)
-    return scanner:match(str or "")
+    return lpegmatch(scanner,str or "")
 end
 
 --~ print(table.serialize(jobreferences.analyse("")))
@@ -198,7 +228,7 @@ local function register_from_lists(collected,derived)
                     local t = { kind, i }
                     for s in gmatch(reference,"%s*([^,]+)") do
                         if trace_referencing then
-                            logs.report("referencing","list entry %s provides %s reference '%s' on realpage %s)",i,kind,s,realpage)
+                            report_references("list entry %s provides %s reference '%s' on realpage %s",i,kind,s,realpage)
                         end
                         d[s] = t -- share them
                     end
@@ -235,6 +265,10 @@ function jobreferences.urls.get(name,method,space) -- method: none, before, afte
     end
 end
 
+function commands.doifurldefinedelse(name)
+    commands.doifelse(urls[name])
+end
+
 -- files
 
 jobreferences.files      = jobreferences.files      or { }
@@ -255,6 +289,59 @@ function jobreferences.files.get(name,method,space) -- method: none, before, aft
     end
 end
 
+function commands.doiffiledefinedelse(name)
+    commands.doifelse(files[name])
+end
+
+-- helpers
+
+function jobreferences.checkedfile(whatever) -- return whatever if not resolved
+    if whatever then
+        local w = files[whatever]
+        if w then
+            return w[1]
+        else
+            return whatever
+        end
+    end
+end
+
+function jobreferences.checkedurl(whatever) -- return whatever if not resolved
+    if whatever then
+        local w = urls[whatever]
+        if w then
+            local u, f = w[1], w[2]
+            if f and f ~= "" then
+                return u .. "/" .. f
+            else
+                return u
+            end
+        else
+            return whatever
+        end
+    end
+end
+
+function jobreferences.checkedfileorurl(whatever,default) -- return nil, nil if not resolved
+    if whatever then
+        local w = files[whatever]
+        if w then
+            return w[1], nil
+        else
+            local w = urls[whatever]
+            if w then
+                local u, f = w[1], w[2]
+                if f and f ~= "" then
+                    return nil, u .. "/" .. f
+                else
+                    return nil, u
+                end
+            end
+        end
+    end
+    return default
+end
+
 -- programs
 
 jobreferences.programs      = jobreferences.programs      or { }
@@ -275,34 +362,69 @@ function jobreferences.programs.get(name)
     end
 end
 
+function jobreferences.checkedprogram(whatever) -- return whatever if not resolved
+    if whatever then
+        local w = programs[whatever]
+        if w then
+            return w[1]
+        else
+            return whatever
+        end
+    end
+end
+
 -- shared by urls and files
 
 function jobreferences.whatfrom(name)
     texsprint(ctxcatcodes,(urls[name] and variables.url) or (files[name] and variables.file) or variables.unknown)
 end
 
+--~ function jobreferences.from(name)
+--~     local u = urls[name]
+--~     if u then
+--~         local url, file, description = u[1], u[2], u[3]
+--~         if description ~= "" then
+--~             texsprint(ctxcatcodes,format("\\dofromurldescription{%s}",description))
+--~             -- ok
+--~         elseif file and file ~= "" then
+--~             texsprint(ctxcatcodes,format("\\dofromurlliteral{%s}",url .. "/" .. file))
+--~         else
+--~             texsprint(ctxcatcodes,format("\\dofromurlliteral{%s}",url))
+--~         end
+--~     else
+--~         local f = files[name]
+--~         if f then
+--~             local description, file = f[1], f[2]
+--~             if description ~= "" then
+--~                 texsprint(ctxcatcodes,format("\\dofromfiledescription{%s}",description))
+--~             else
+--~                 texsprint(ctxcatcodes,format("\\dofromfileliteral{%s}",file))
+--~             end
+--~         end
+--~     end
+--~ end
+
 function jobreferences.from(name)
     local u = urls[name]
     if u then
         local url, file, description = u[1], u[2], u[3]
         if description ~= "" then
+            context.dofromurldescription(description)
             -- ok
         elseif file and file ~= "" then
-            description = url .. "/" .. file
+            context.dofromurlliteral(url .. "/" .. file)
         else
-            description = url
+            context.dofromurlliteral(url)
         end
-        texsprint(ctxcatcodes,format("\\dofromurl{%s}",description))
     else
         local f = files[name]
         if f then
             local description, file = f[1], f[2]
             if description ~= "" then
-                --
+                context.dofromfiledescription(description)
             else
-                description = file
+                context.dofromfileliteral(file)
             end
-            texsprint(ctxcatcodes,format("\\dofromfile{%s}",description))
         end
     end
 end
@@ -494,7 +616,7 @@ local function resolve(prefix,reference,args,set) -- we start with prefix,refere
             if d then
                 resolve(prefix,d[2],nil,set)
             else
-                local var = scanner:match(ri)
+                local var = lpegmatch(scanner,ri)
                 if var then
                     var.reference = ri
                     if not var.outer and var.inner then
@@ -513,7 +635,7 @@ local function resolve(prefix,reference,args,set) -- we start with prefix,refere
                         set.has_tex = true
                     end
                 else
-                --  logs.report("references","funny pattern: %s",ri or "?")
+                --  report_references("funny pattern: %s",ri or "?")
                 end
             end
         end
@@ -780,7 +902,7 @@ local function identify(prefix,reference)
         set[i] = var
     end
     jobreferences.currentset = set
---  print(bug,table.serialize(set))
+--~ print(bug,table.serialize(set))
     return set, bug
 end
 
@@ -821,6 +943,12 @@ function jobreferences.setinternalreference(prefix,tag,internal,view)
     return destination
 end
 
+function jobreferences.getinternalreference(n) -- n points into list
+    local l = structure.lists.collected[n]
+    texsprint(ctxcatcodes,(l and l.references.internal) or n)
+end
+
+
 --
 
 function jobreferences.get_current_metadata(tag)
@@ -838,8 +966,9 @@ jobreferences.current_metadata = current_metadata
 
 function jobreferences.get_current_prefixspec(default) -- todo: message
     texsprint(ctxcatcodes,"\\getreferencestructureprefix{",
-        current_metadata("kind"), "}{", current_metadata("name") or "?", "}{", default, "}")
+        current_metadata("kind") or "?", "}{", current_metadata("name") or "?", "}{", default or "?", "}")
 end
+
 --~ function jobreferences.get_current_prefixspec(default) -- we can consider storing the data at the lua end
 --~     context.getreferencestructureprefix(current_metadata("kind"),current_metadata("name"),default)
 --~ end
@@ -860,15 +989,18 @@ function jobreferences.filter(name,...) -- number page title ...
             local filter = filters[kind] or filters.generic
             filter = filter and (filter[name] or filter.unknown or filters.generic[name] or filters.generic.unknown)
             if filter then
+                if trace_referencing then
+                    report_references("name '%s', kind '%s', using dedicated filter",name,kind)
+                end
                 filter(data,name,...)
             elseif trace_referencing then
-                logs.report("referencing","no (generic) filter.name for '%s'",name)
+                report_references("name '%s', kind '%s', using generic filter",name,kind)
             end
         elseif trace_referencing then
-            logs.report("referencing","no metadata.kind for '%s'",name)
+            report_references("name '%s', unknown kind",name)
         end
     elseif trace_referencing then
-        logs.report("referencing","no current reference for '%s'",name)
+        report_references("name '%s', no reference",name)
     end
 end
 
@@ -892,12 +1024,13 @@ function filters.generic.text(data)
     end
 end
 
-function filters.generic.number(data,what,...) -- todo: spec and then no stopper
+function filters.generic.number(data,what,prefixspec) -- todo: spec and then no stopper
     if data then
         local numberdata = data.numberdata
         if numberdata then
-            helpers.prefix(data,...)
-            sections.typesetnumber(numberdata,"number",numberdata or false)
+--~ print(table.serialize(prefixspec))
+            helpers.prefix(data,prefixspec)
+            sections.typesetnumber(numberdata,"number",numberdata)
         else
             local useddata = data.useddata
             if useddata and useddsta.number then
@@ -964,11 +1097,11 @@ end
 
 filters.section = { }
 
-function filters.section.number(data) -- todo: spec and then no stopper
+function filters.section.number(data,what,prefixspec)
     if data then
         local numberdata = data.numberdata
         if numberdata then
-            sections.typesetnumber(numberdata,"number",numberdata or false)
+            sections.typesetnumber(numberdata,"number",prefixspec,numberdata)
         else
             local useddata = data.useddata
             if useddata and useddata.number then
@@ -1049,7 +1182,6 @@ function jobreferences.analyse(actions)
     return actions
 end
 
-
 function jobreferences.realpage() -- special case, we always want result
     local cs = jobreferences.analyse()
     texwrite(cs.realpage or 0)
@@ -1102,7 +1234,7 @@ end
 
 specials.i = specials.internal
 
-function specials.page(var,actions)
+function specials.page(var,actions) -- is this ok?
     local p = pages[var.operation]
     if type(p) == "function" then
         p = p()
