@@ -10,17 +10,27 @@ local next, type = next, type
 local texwrite, texsprint, texcount = tex.write, tex.sprint, tex.count
 local format, gmatch, concat = string.format, string.gmatch, table.concat
 local utfchar = utf.char
+local lpegmatch = lpeg.match
 
 local trace_registers = false  trackers.register("structure.registers", function(v) trace_registers = v end)
 
+local report_registers = logs.new("registers")
+
 local ctxcatcodes = tex.ctxcatcodes
 
 local variables = interfaces.variables
 
-local helpers   = structure.helpers
-local sections  = structure.sections
-local documents = structure.documents
-local pages     = structure.pages
+local helpers    = structure.helpers
+local sections   = structure.sections
+local documents  = structure.documents
+local pages      = structure.pages
+local processors = structure.processors
+
+local mappings     = sorters.mappings
+local entries      = sorters.entries
+local replacements = sorters.replacements
+
+local processor_split = processors.split
 
 local matching_till_depth, number_at_depth = sections.matching_till_depth, sections.number_at_depth
 
@@ -149,9 +159,9 @@ local function filter_collected(names,criterium,number,collected,prevmode)
     end
     if trace_registers then
         if detail then
-            logs.report("registers","criterium: %s, %s, found: %s",criterium,detail,#result)
+            report_registers("criterium: %s, %s, found: %s",criterium,detail,#result)
         else
-            logs.report("registers","criterium: %s, found: %s",criterium,#result)
+            report_registers("criterium: %s, found: %s",criterium,#result)
         end
     end
     return result
@@ -215,20 +225,34 @@ local function preprocessentries(rawdata)
     local entries = rawdata.entries
     if entries then
         local e, k = entries[1] or "", entries[2] or ""
-        local et = (type(e) == "table" and e) or entrysplitter:match(e)
-        local kt = (type(k) == "table" and k) or entrysplitter:match(k)
+        local et, kt, entryproc, pageproc
+        if type(e) == "table" then
+            et = e
+        else
+            entryproc, e = processor_split(e)
+            et = lpegmatch(entrysplitter,e)
+        end
+        if type(k) == "table" then
+            kt = e
+        else
+            pageproc, k = processor_split(k)
+            kt = lpegmatch(entrysplitter,k)
+        end
         entries = { }
         for k=1,#et do
             entries[k] = { et[k] or "", kt[k] or "" }
         end
-for k=#et,1,-1 do
-    if entries[k][1] ~= "" then
-        break
-    else
-        entries[k] = nil
-    end
-end
+        for k=#et,1,-1 do
+            if entries[k][1] ~= "" then
+                break
+            else
+                entries[k] = nil
+            end
+        end
         rawdata.list = entries
+        if pageproc or entryproc then
+            rawdata.processors = { entryproc, pageproc }
+        end
         rawdata.entries = nil
     else
         rawdata.list = { { "", "" } } -- br
@@ -265,12 +289,12 @@ function jobregisters.extend(name,tag,rawdata) -- maybe do lastsection internall
             rr.lastsection = structure.sections.currentid()
             if rawdata then
                 preprocessentries(rawdata)
-                for k,v in pairs(rawdata) do
+                for k,v in next, rawdata do
                     if not r[k] then
                         r[k] = v
                     else
                         local rk = r[k]
-                        for kk,vv in pairs(v) do
+                        for kk,vv in next, v do
                             if vv ~= "" then
                                 rk[kk] = vv
                             end
@@ -284,39 +308,23 @@ end
 
 -- sorting and rendering
 
+local compare = sorters.comparers.basic
+
 function jobregisters.compare(a,b)
-    local result = 0
-    local compare = sorters.comparers.basic
-    local ea, eb = a.split, b.split
-    local na, nb = #ea, #eb
-    local max = na
-    if nb < max then max = nb end
-    for i=1,max do
-        if result == 0 then
-            result = compare(ea[i],eb[i])
-        else
-            return result
-        end
-    end
+    local result = compare(a,b)
     if result ~= 0 then
         return result
-    elseif na > nb then
-        return 1
-    elseif nb > na then
-        return -1
     elseif a.metadata.kind == 'entry' then -- e/f/t
         local page_a, page_b = a.references.realpage, b.references.realpage
         if not page_a or not page_b then
---~ print(table.serialize(a),table.serialize(b))
             return 0
         elseif page_a < page_b then
             return -1
         elseif page_a > page_b then
             return  1
         end
-    else
-        return 0
     end
+    return 0
 end
 
 function jobregisters.filter(data,options)
@@ -353,7 +361,9 @@ end
 
 function jobregisters.unique(data,options)
     local result, prev, equal = { }, nil, table.are_equal
-    for _,v in ipairs(data.result) do
+    local dataresult = data.result
+    for k=1,#dataresult do
+        local v = dataresult[k]
         if not prev then
             result[#result+1], prev = v, v
         else
@@ -382,17 +392,21 @@ end
 function jobregisters.finalize(data,options)
     local result = data.result
     data.metadata.nofsorted = #result
-    local split = { }
+    local split, lasttag, s, d = { }, nil, nil, nil
     -- maps character to index (order)
     for k=1,#result do
         local v = result[k]
-        local entry, tag = sorters.firstofsplit(v.split)
-        local s = split[tag] -- keeps track of change
-        if not s then
-            s = { tag = tag, data = { } }
-            split[tag] = s
+        local entry, tag = sorters.firstofsplit(v)
+        if tag ~= lasttag then
+            if trace_registers then
+                report_registers("splitting at %s",tag)
+            end
+            d = { }
+            s = { tag = tag, data = d }
+            split[#split+1] = s
+            lasttag = tag
         end
-        s.data[#s.data+1] = v
+        d[#d+1] = v
     end
     data.result = split
 end
@@ -401,7 +415,7 @@ function jobregisters.analysed(class,options)
     local data = collected[class]
     if data and data.entries then
         options = options or { }
-        sorters.language = options.language or sorters.defaultlanguage
+        sorters.setlanguage(options.language)
         jobregisters.filter(data,options)   -- filter entries into results (criteria)
         jobregisters.prepare(data,options)  -- adds split table parallel to list table
         jobregisters.sort(data,options)     -- sorts results
@@ -416,6 +430,16 @@ end
 
 -- todo take conversion from index
 
+function jobregisters.userdata(index,name)
+    local data = jobreferences.internals[tonumber(index)]
+    data = data and data.userdata and data.userdata[name]
+    if data then
+        texsprint(ctxcatcodes,data)
+    end
+end
+
+-- proc can be wrapped
+
 function jobregisters.flush(data,options,prefixspec,pagespec)
     local equal = table.are_equal
     texsprint(ctxcatcodes,"\\startregisteroutput")
@@ -426,25 +450,51 @@ function jobregisters.flush(data,options,prefixspec,pagespec)
     local function pagenumber(entry)
         local er = entry.references
         texsprint(ctxcatcodes,format("\\registeronepage{%s}{%s}{",er.internal or 0,er.realpage or 0)) -- internal realpage content
-        helpers.prefixpage(entry,prefixspec,pagespec)
+        local proc = entry.processors and entry.processors[2]
+        if proc then
+            texsprint(ctxcatcodes,"\\applyprocessor{",proc,"}{")
+            helpers.prefixpage(entry,prefixspec,pagespec)
+            texsprint(ctxcatcodes,"}")
+        else
+            helpers.prefixpage(entry,prefixspec,pagespec)
+        end
         texsprint(ctxcatcodes,"}")
     end
     local function pagerange(f_entry,t_entry,is_last)
         local er = f_entry.references
         texsprint(ctxcatcodes,format("\\registerpagerange{%s}{%s}{",er.internal or 0,er.realpage or 0))
-        helpers.prefixpage(f_entry,prefixspec,pagespec)
-        local er = t_entry.references
-        texsprint(ctxcatcodes,format("}{%s}{%s}{",er.internal or 0,er.realpage or 0))
-        if is_last then
-            helpers.prefixlastpage(t_entry,prefixspec,pagespec) -- swaps page and realpage keys
+        local proc = f_entry.processors and f_entry.processors[2]
+        if proc then
+            texsprint(ctxcatcodes,"\\applyprocessor{",proc,"}{")
+            helpers.prefixpage(f_entry,prefixspec,pagespec)
+            texsprint(ctxcatcodes,"}")
         else
-            helpers.prefixpage(t_entry,prefixspec,pagespec)
+            helpers.prefixpage(f_entry,prefixspec,pagespec)
+        end
+        local er = t_entry.references
+        texsprint(ctxcatcodes,format("}{%s}{%s}{",er.internal or 0,er.lastrealpage or er.realpage or 0))
+        if is_last then
+            if proc then
+                texsprint(ctxcatcodes,"\\applyprocessor{",proc,"}{")
+                helpers.prefixlastpage(t_entry,prefixspec,pagespec) -- swaps page and realpage keys
+                texsprint(ctxcatcodes,"}")
+            else
+                helpers.prefixlastpage(t_entry,prefixspec,pagespec) -- swaps page and realpage keys
+            end
+        else
+            if proc then
+                texsprint(ctxcatcodes,"\\applyprocessor{",proc,"}{")
+                helpers.prefixpage(t_entry,prefixspec,pagespec)
+                texsprint(ctxcatcodes,"}")
+            else
+                helpers.prefixpage(t_entry,prefixspec,pagespec)
+            end
         end
         texsprint(ctxcatcodes,"}")
     end
     -- ranges need checking !
-    for k, letter in ipairs(table.sortedkeys(result)) do
-        local sublist = result[letter]
+    for i=1,#result do
+        local sublist = result[i]
         local done = { false, false, false, false }
         local data = sublist.data
         local d, n = 0, 0
@@ -473,13 +523,27 @@ function jobregisters.flush(data,options,prefixspec,pagespec)
                                 texsprint(ctxcatcodes,format("\\startregisterentries{%s}",n))
                             end
                         end
-if metadata then
-    texsprint(ctxcatcodes,"\\registerentry{")
-    helpers.title(e[i],metadata)
-    texsprint(ctxcatcodes,"}")
-else
-                        texsprint(ctxcatcodes,format("\\registerentry{%s}",e[i]))
-end
+                        if metadata then
+                            texsprint(ctxcatcodes,"\\registerentry{")
+                            local proc = entry.processors and entry.processors[1]
+                            if proc then
+                                texsprint(ctxcatcodes,"\\applyprocessor{",proc,"}{")
+                                helpers.title(e[i],metadata)
+                                texsprint(ctxcatcodes,"}")
+                            else
+                                helpers.title(e[i],metadata)
+                            end
+                            texsprint(ctxcatcodes,"}")
+                        else
+                            local proc = entry.processors and entry.processors[1]
+                            if proc then
+                                texsprint(ctxcatcodes,"\\applyprocessor{",proc,"}{")
+                                texsprint(ctxcatcodes,format("\\registerentry{%s}",e[i]))
+                                texsprint(ctxcatcodes,"}")
+                            else
+                                texsprint(ctxcatcodes,format("\\registerentry{%s}",e[i]))
+                            end
+                        end
                     else
                         done[i] = false
                     end
@@ -488,6 +552,7 @@ end
             local kind = entry.metadata.kind
             if kind == 'entry' then
                 texsprint(ctxcatcodes,"\\startregisterpages")
+            --~ collapse_ranges = true
                 if collapse_singles or collapse_ranges then
                     -- we collapse ranges and keep existing ranges as they are
                     -- so we get prebuilt as well as built ranges
@@ -638,7 +703,14 @@ end
             elseif kind == 'see' then
                 -- maybe some day more words
                 texsprint(ctxcatcodes,"\\startregisterseewords")
-                texsprint(ctxcatcodes,format("\\registeroneword{0}{0}{%s}",entry.seeword.text)) -- todo: internal
+                local proc = entry.processors and entry.processors[1]
+                if proc then
+                    texsprint(ctxcatcodes,"\\applyprocessor{",proc,"}{")
+                    texsprint(ctxcatcodes,format("\\registeroneword{0}{0}{%s}",entry.seeword.text)) -- todo: internal
+                    texsprint(ctxcatcodes,"}")
+                else
+                    texsprint(ctxcatcodes,format("\\registeroneword{0}{0}{%s}",entry.seeword.text)) -- todo: internal
+                end
                 texsprint(ctxcatcodes,"\\stopregisterseewords")
             end
         end
