@@ -9,15 +9,23 @@ if not modules then modules = { } end modules ['supp-ran'] = {
 -- We cannot ask for the current seed, so we need some messy hack
 -- here.
 
+local report_system = logs.new("system")
+
 commands = commands or { }
 
-local texwrite, random, randomseed, round, seed, last = tex.write, math.random, math.randomseed, math.round, false, 1
+local random, randomseed, round, seed, last = math.random, math.randomseed, math.round, false, 1
+local texwrite = tex.write
 
-function math.setrandomseedi(n)
+function math.setrandomseedi(n,comment)
     if n <= 1 then
         n = n*1073741823 -- maxcount
     end
-    randomseed(round(n))
+    n = round(n)
+    if false then
+        report_system("setting random seed to %s (%s)",n,comment or "normal")
+    end
+    randomseed(n)
+    last = random(0,1073741823) -- we need an initial value
 end
 
 function commands.getrandomcounta(min,max)
@@ -39,9 +47,13 @@ function commands.getrandomseed(n)
     texwrite(last)
 end
 
+-- maybe stack
+
 function commands.freezerandomseed(n)
-    if seed == false then
+ -- print("<<<<",seed,last)
+    if seed == false or seed == nil then
         seed = last
+        math.setrandomseedi(seed,"freeze",seed)
     end
     if n then
         randomseed(n)
@@ -49,8 +61,9 @@ function commands.freezerandomseed(n)
 end
 
 function commands.defrostrandomseed()
+ -- print(">>>>",seed,last)
     if seed ~= false then
-        math.setrandomseedi(last)
+        math.setrandomseedi(seed,"defrost",seed) -- was last (bug)
         seed = false
     end
 end
