@@ -9,13 +9,15 @@ if not modules then modules = { } end modules ['typo-mir'] = {
 local utf = unicode.utf8
 
 local next, type = next, type
-local format, insert = string.format, table.insert
+local format, insert, sub, find, match = string.format, table.insert, string.sub, string.find, string.match
 local utfchar = utf.char
 
 -- vertical space handler
 
 local trace_mirroring = false  trackers.register("nodes.mirroring", function(v) trace_mirroring = v end)
 
+local report_bidi = logs.new("bidi")
+
 local has_attribute      = node.has_attribute
 local unset_attribute    = node.unset_attribute
 local set_attribute      = node.set_attribute
@@ -29,6 +31,7 @@ local whatsit = node.id("whatsit")
 local mthnode = node.id('math')
 
 local fontdata = fonts.ids
+local fontchar = fonts.chr
 local chardata = characters.data
 
 --~ Analysis by Idris:
@@ -56,13 +59,12 @@ local chardata = characters.data
 --  elseif d == "ws"  then -- Whitespace
 --  elseif d == "on"  then -- Other Neutrals
 
-mirror         = mirror or { }
-mirror.enabled = false
-mirror.strip   = false
-mirror.attribute = attributes.private("mirror")
+mirroring           = mirroring or { }
+mirroring.strip     = false
+mirroring.attribute = attributes.private("mirroring")
 
-local state   = attributes.private('state')
-local mirrora = attributes.private('mirror')
+local a_state     = attributes.private('state')
+local a_mirroring = attributes.private('mirroring')
 
 local directions = characters.directions -- maybe make a special mirror table
 
@@ -82,7 +84,7 @@ local function finish_auto_before()
     head, inserted = insert_node_before(head,current,nodes.textdir("-"..finish))
     finished, finidir = inserted, finish
     if trace_mirroring then
-        insert(list,#list,format("finish %s",finish))
+        insert(list,#list,format("auto finish inserted before: %s",finish))
         finipos = #list-1
     end
     finish, autodir, done = nil, 0, true
@@ -92,7 +94,7 @@ local function finish_auto_after()
     head, current = insert_node_after(head,current,nodes.textdir("-"..finish))
     finished, finidir = current, finish
     if trace_mirroring then
-        list[#list+1] = format("finish %s",finish)
+        list[#list+1] = format("auto finish inserted after: %s",finish)
         finipos = #list
     end
     finish, autodir, done = nil, 0, true
@@ -108,15 +110,15 @@ local function force_auto_left_before()
         finish, autodir, done = "TRT", -1, true
     end
     if finidir == finish then
-        remove_node(head,finished,true)
+        head = remove_node(head,finished,true)
         if trace_mirroring then
-            list[finipos] = list[finipos].." (deleted)"
-            insert(list,#list,format("start %s (deleted)",finish))
+            list[finipos] = list[finipos] .. " (deleted afterwards)"
+            insert(list,#list,format("start text dir %s (embedded: %s)",finish,embedded))
         end
     else
         head, inserted = insert_node_before(head,current,nodes.textdir("+"..finish))
         if trace_mirroring then
-            insert(list,#list,format("start %s",finish))
+            insert(list,#list,format("start text dir %s (embedded: %s)",finish,embedded))
         end
     end
 end
@@ -131,20 +133,22 @@ local function force_auto_right_before()
         finish, autodir, done = "TLT", 1, true
     end
     if finidir == finish then
-        remove_node(head,finished,true)
+        head = remove_node(head,finished,true)
         if trace_mirroring then
-            list[finipos] = list[finipos].." (deleted)"
-            insert(list,#list,format("start %s (deleted)",finish))
+            list[finipos] = list[finipos] .. " (deleted afterwards)"
+            insert(list,#list,format("start text dir %s (embedded: %s)",finish,embedded))
         end
     else
         head, inserted = insert_node_before(head,current,nodes.textdir("+"..finish))
         if trace_mirroring then
-            insert(list,#list,format("start %s",finish))
+            insert(list,#list,format("start text dir %s (embedded: %s)",finish,embedded))
         end
     end
 end
 
-function mirror.process(namespace,attribute,start) -- todo: make faster
+-- todo: use new dir functions
+
+function mirroring.process(namespace,attribute,start) -- todo: make faster
     if not start.next then
         return start, false
     end
@@ -156,6 +160,7 @@ function mirror.process(namespace,attribute,start) -- todo: make faster
     local lro, rlo, prevattr, inmath = false, false, 0, false
     while current do
         local id = current.id
+--~ print(id,attribute,has_attribute(current,attribute))
         if skipmath and id == mthnode then
             local subtype = current.subtype
             if subtype == 0 then
@@ -172,8 +177,7 @@ function mirror.process(namespace,attribute,start) -- todo: make faster
         else
             local attr = has_attribute(current,attribute)
             if attr and attr > 0 then
-                unset_attribute(current,attribute) -- slow, needed?
---~             set_attribute(current,attribute,0) -- might be faster
+             -- unset_attribute(current,attribute) -- slow, needed?
                 if attr == 1 then
                     -- bidi parsing mode
                 elseif attr ~= prevattr then
@@ -217,7 +221,7 @@ function mirror.process(namespace,attribute,start) -- todo: make faster
                         end
                     elseif lro or override < 0 then
                         if d == "r" or d == "al" then
-                            set_attribute(current,state,4) -- maybe better have a special bidi attr value -> override (9) -> todo
+                            set_attribute(current,a_state,4) -- maybe better have a special bidi attr value -> override (9) -> todo
                             if trace_mirroring then
                                 list[#list+1] = format("char %s (%s / U+%04X) of class %s overidden to l (bidi=%s) (state=isol)",utfchar(char),char,char,d,attr)
                             end
@@ -238,7 +242,7 @@ function mirror.process(namespace,attribute,start) -- todo: make faster
                     end
                     if d == "on" then
                         local mirror = chardata[char].mirror -- maybe make a special mirror table
-                        if mirror and fontdata[current.font].characters[mirror] then
+                        if mirror and fontchar[current.font][mirror] then
                             -- todo: set attribute
                             if autodir < 0 then
                                 current.char = mirror
@@ -294,7 +298,7 @@ function mirror.process(namespace,attribute,start) -- todo: make faster
                         end
                         top = top + 1
                         stack[top] = { override, embedded }
-                        embedded = 1
+                        embedded = -1 -- was 1
                         obsolete[#obsolete+1] = current
                     elseif d == "pdf" then -- Pop Directional Format
                     --  override = 0
@@ -312,12 +316,10 @@ function mirror.process(namespace,attribute,start) -- todo: make faster
                         end
                         obsolete[#obsolete+1] = current
                     end
-                else
-                    if trace_mirroring then
-                        local char = current.char
-                        local d = directions[char]
-                        list[#list+1] = format("char %s (%s / U+%04X) of class %s (no bidi)",utfchar(char),char,char,d or "?")
-                    end
+                elseif trace_mirroring then
+                    local char = current.char
+                    local d = directions[char]
+                    list[#list+1] = format("char %s (%s / U+%04X) of class %s (no bidi)",utfchar(char),char,char,d or "?")
                 end
             elseif id == whatsit then
                 if finish then
@@ -326,20 +328,21 @@ function mirror.process(namespace,attribute,start) -- todo: make faster
                 local subtype = current.subtype
                 if subtype == 6 then
                     local dir = current.dir
-                    local d = dir:sub(2,2) -- -- -- -- -- why is this not used
-                    if dir:find(".R.") then -- -- d == "R" or just dir == "TRT"
+                    local d = sub(dir,2,2)
+                    if d == 'R' then -- find(dir,".R.") / dir == "TRT"
                         autodir = -1
                     else
                         autodir = 1
                     end
-                    embeddded = autodir
+                 -- embedded = autodir
                     if trace_mirroring then
                         list[#list+1] = format("pardir %s",dir)
                     end
                 elseif subtype == 7 then
                     local dir = current.dir
-                    local sign = dir:sub(1,1)
-                    local dire = dir:sub(3,3)
+                 -- local sign = sub(dir,1,1)
+                 -- local dire = sub(dir,3,3)
+                    local sign, dire = match(dir,"^(.).(.)")
                     if dire == "R" then
                         if sign == "+" then
                             finish, autodir = "TRT", -1
@@ -375,19 +378,19 @@ function mirror.process(namespace,attribute,start) -- todo: make faster
         end
     end
     if trace_mirroring and glyphs then
-        logs.report("bidi","start log")
+        report_bidi("start log")
         for i=1,#list do
-            logs.report("bidi","%02i: %s",i,list[i])
+            report_bidi("%02i: %s",i,list[i])
         end
-        logs.report("bidi","stop log")
+        report_bidi("stop log")
     end
-    if done and mirror.strip then
+    if done and mirroring.strip then
         local n = #obsolete
         if n > 0 then
             for i=1,n do
                 remove_node(head,obsolete[i],true)
             end
-            logs.report("bidi","%s character nodes removed",n)
+            report_bidi("%s character nodes removed",n)
         end
     end
     return head, done
@@ -410,7 +413,11 @@ end
 --~         end
 
 chars.handle_mirroring = nodes.install_attribute_handler {
-    name = "mirror",
-    namespace = mirror,
-    processor = mirror.process,
+    name = "mirroring",
+    namespace = mirroring,
+    processor = mirroring.process,
 }
+
+function mirroring.enable()
+    tasks.enableaction("processors","chars.handle_mirroring")
+end
